@@ -1,6 +1,6 @@
 // MIT License
 //
-// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -25,6 +25,11 @@
 
 #include "benchmark_utils.hpp"
 
+#include "../common/utils_device_ptr.hpp"
+#ifndef BENCHMARK_CONFIG_TUNING
+    #include "../common/utils_custom_type.hpp"
+#endif
+
 // Google Benchmark
 #include <benchmark/benchmark.h>
 
@@ -32,21 +37,37 @@
 #include <hip/hip_runtime.h>
 
 // rocPRIM
+#include <rocprim/config.hpp>
+#include <rocprim/device/config_types.hpp>
 #include <rocprim/device/detail/device_config_helper.hpp>
 #include <rocprim/device/device_scan.hpp>
-
-#include <string>
-#include <vector>
+#include <rocprim/functional.hpp>
+#ifdef BENCHMARK_CONFIG_TUNING
+    #include <rocprim/block/block_load.hpp>
+    #include <rocprim/block/block_scan.hpp>
+    #include <rocprim/block/block_store.hpp>
+#else
+    #include <rocprim/functional.hpp>
+    #include <rocprim/types.hpp>
+#endif
 
 #include <cstddef>
+#include <string>
+#include <type_traits>
+#include <vector>
+#ifdef BENCHMARK_CONFIG_TUNING
+    #include <memory>
+#else
+    #include <stdint.h>
+#endif
 
 template<typename Config>
 std::string config_name()
 {
     const rocprim::detail::scan_config_params config = Config();
     return "{bs:" + std::to_string(config.kernel_config.block_size)
-           + ",ipt:" + std::to_string(config.kernel_config.items_per_thread)
-           + ",method:" + std::string(get_block_scan_method_name(config.block_scan_method)) + "}";
+           + ",ipt:" + std::to_string(config.kernel_config.items_per_thread) + ",method:"
+           + std::string(get_block_scan_algorithm_name(config.block_scan_method)) + "}";
 }
 
 template<>
@@ -55,12 +76,12 @@ inline std::string config_name<rocprim::default_config>()
     return "default_config";
 }
 
-template<bool Exclusive     = false,
-         class T            = int,
-         class ScanOp       = rocprim::plus<T>,
-         bool Deterministic = false,
-         class Config       = rocprim::default_config>
-struct device_scan_benchmark : public config_autotune_interface
+template<bool Exclusive,
+         typename T,
+         typename ScanOp,
+         bool Deterministic,
+         typename Config = rocprim::default_config>
+struct device_scan_benchmark : public benchmark_utils::autotune_interface
 {
     std::string name() const override
     {
@@ -82,7 +103,7 @@ struct device_scan_benchmark : public config_autotune_interface
                          const bool        debug = false) const ->
         typename std::enable_if<excl, hipError_t>::type
     {
-        if ROCPRIM_IF_CONSTEXPR(!Deterministic)
+        if constexpr(!Deterministic)
         {
             return rocprim::exclusive_scan<Config>(temporary_storage,
                                                    storage_size,
@@ -121,7 +142,7 @@ struct device_scan_benchmark : public config_autotune_interface
         typename std::enable_if<!excl, hipError_t>::type
     {
         (void)initial_value;
-        if ROCPRIM_IF_CONSTEXPR(!Deterministic)
+        if constexpr(!Deterministic)
         {
             return rocprim::inclusive_scan<Config>(temporary_storage,
                                                    storage_size,
@@ -145,94 +166,52 @@ struct device_scan_benchmark : public config_autotune_interface
         }
     }
 
-    void run(benchmark::State&   state,
-             size_t              size,
-             const managed_seed& seed,
-             hipStream_t         stream) const override
+    void run(benchmark_utils::state&& state) override
     {
+        const auto& stream = state.stream;
+        const auto& bytes  = state.bytes;
+        const auto& seed   = state.seed;
+
+        // Calculate the number of elements
+        size_t size = bytes / sizeof(T);
+
         ScanOp         scan_op{};
         const auto     random_range = limit_random_range<T>(0, 1000);
         std::vector<T> input
             = get_random_data<T>(size, random_range.first, random_range.second, seed.get_0());
-        T              initial_value = T(123);
-        T*             d_input;
-        T*             d_output;
-        HIP_CHECK(hipMalloc(&d_input, size * sizeof(T)));
-        HIP_CHECK(hipMalloc(&d_output, size * sizeof(T)));
-        HIP_CHECK(hipMemcpy(d_input, input.data(), size * sizeof(T), hipMemcpyHostToDevice));
+        T                     initial_value = T(123);
+        common::device_ptr<T> d_input(input);
+        common::device_ptr<T> d_output(size);
         HIP_CHECK(hipDeviceSynchronize());
 
         // Allocate temporary storage memory
         size_t temp_storage_size_bytes;
-        void*  d_temp_storage = nullptr;
         // Get size of d_temp_storage
-        HIP_CHECK((run_device_scan(d_temp_storage,
+        HIP_CHECK((run_device_scan(nullptr,
                                    temp_storage_size_bytes,
-                                   d_input,
-                                   d_output,
+                                   d_input.get(),
+                                   d_output.get(),
                                    initial_value,
                                    size,
                                    scan_op,
                                    stream)));
-        HIP_CHECK(hipMalloc(&d_temp_storage, temp_storage_size_bytes));
+        common::device_ptr<void> d_temp_storage(temp_storage_size_bytes);
         HIP_CHECK(hipDeviceSynchronize());
 
-        // Warm-up
-        for(size_t i = 0; i < 5; i++)
-        {
-            HIP_CHECK((run_device_scan(d_temp_storage,
-                                       temp_storage_size_bytes,
-                                       d_input,
-                                       d_output,
-                                       initial_value,
-                                       size,
-                                       scan_op,
-                                       stream)));
-        }
-        HIP_CHECK(hipDeviceSynchronize());
-
-        // HIP events creation
-        hipEvent_t start, stop;
-        HIP_CHECK(hipEventCreate(&start));
-        HIP_CHECK(hipEventCreate(&stop));
-
-        const unsigned int batch_size = 10;
-        for(auto _ : state)
-        {
-            // Record start event
-            HIP_CHECK(hipEventRecord(start, stream));
-
-            for(size_t i = 0; i < batch_size; i++)
+        state.run(
+            [&]
             {
-                HIP_CHECK((run_device_scan(d_temp_storage,
+                HIP_CHECK((run_device_scan(d_temp_storage.get(),
                                            temp_storage_size_bytes,
-                                           d_input,
-                                           d_output,
+                                           d_input.get(),
+                                           d_output.get(),
                                            initial_value,
                                            size,
                                            scan_op,
                                            stream)));
-            }
+            });
 
-            // Record stop event and wait until it completes
-            HIP_CHECK(hipEventRecord(stop, stream));
-            HIP_CHECK(hipEventSynchronize(stop));
-
-            float elapsed_mseconds;
-            HIP_CHECK(hipEventElapsedTime(&elapsed_mseconds, start, stop));
-            state.SetIterationTime(elapsed_mseconds / 1000);
-        }
-
-        // Destroy HIP events
-        HIP_CHECK(hipEventDestroy(start));
-        HIP_CHECK(hipEventDestroy(stop));
-
-        state.SetBytesProcessed(state.iterations() * batch_size * size * sizeof(T));
-        state.SetItemsProcessed(state.iterations() * batch_size * size);
-
-        HIP_CHECK(hipFree(d_input));
-        HIP_CHECK(hipFree(d_output));
-        HIP_CHECK(hipFree(d_temp_storage));
+        state.set_throughput(size, sizeof(T));
     }
 };
 
@@ -250,7 +229,8 @@ struct device_scan_benchmark_generator
             template<unsigned int ItemsPerThread>
             struct create_ipt
             {
-                void operator()(std::vector<std::unique_ptr<config_autotune_interface>>& storage)
+                void operator()(
+                    std::vector<std::unique_ptr<benchmark_utils::autotune_interface>>& storage)
                 {
                     storage.emplace_back(
                         std::make_unique<device_scan_benchmark<
@@ -266,11 +246,12 @@ struct device_scan_benchmark_generator
                 }
             };
 
-            void operator()(std::vector<std::unique_ptr<config_autotune_interface>>& storage)
+            void operator()(
+                std::vector<std::unique_ptr<benchmark_utils::autotune_interface>>& storage)
             {
                 // Limit items per thread to not over-use shared memory
                 static constexpr unsigned int max_items_per_thread
-                    = ::rocprim::min<size_t>(65536 / (block_size * sizeof(T)), 24);
+                    = ::rocprim::min<size_t>(65536 / (block_size * sizeof(T)) - 1, 24);
                 static_for_each<make_index_range<unsigned int, 1, max_items_per_thread>,
                                 create_ipt>(storage);
             }
@@ -278,19 +259,50 @@ struct device_scan_benchmark_generator
             static constexpr unsigned int block_size = 1u << BlockSizeExponent;
         };
 
-        static void create(std::vector<std::unique_ptr<config_autotune_interface>>& storage)
+        static void
+            create(std::vector<std::unique_ptr<benchmark_utils::autotune_interface>>& storage)
         {
             static_for_each<index_range, create_block_size>(storage);
         }
     };
 
-    static void create(std::vector<std::unique_ptr<config_autotune_interface>>& storage)
+    static void create(std::vector<std::unique_ptr<benchmark_utils::autotune_interface>>& storage)
     {
         // Block sizes 64, 128, 256
         create_block_scan_algorithm<make_index_range<unsigned int, 6, 8>>::create(storage);
     }
 };
 
+#else
+
+    #define CREATE_EXCL_INCL_BENCHMARK(EXCL, T, SCAN_OP) \
+        executor.queue_instance(device_scan_benchmark<EXCL, T, SCAN_OP, Deterministic>());
+
+    #define CREATE_BENCHMARK(T, SCAN_OP)              \
+        CREATE_EXCL_INCL_BENCHMARK(false, T, SCAN_OP) \
+        CREATE_EXCL_INCL_BENCHMARK(true, T, SCAN_OP)
+
+template<bool Deterministic>
+void add_benchmarks(benchmark_utils::executor& executor)
+{
+    using custom_float2  = common::custom_type<float, float>;
+    using custom_double2 = common::custom_type<double, double>;
+
+    CREATE_BENCHMARK(int, rocprim::plus<int>)
+    CREATE_BENCHMARK(float, rocprim::plus<float>)
+    CREATE_BENCHMARK(double, rocprim::plus<double>)
+    CREATE_BENCHMARK(long long, rocprim::plus<long long>)
+    CREATE_BENCHMARK(float2, rocprim::plus<float2>)
+    CREATE_BENCHMARK(custom_float2, rocprim::plus<custom_float2>)
+    CREATE_BENCHMARK(double2, rocprim::plus<double2>)
+    CREATE_BENCHMARK(custom_double2, rocprim::plus<custom_double2>)
+    CREATE_BENCHMARK(int8_t, rocprim::plus<int8_t>)
+    CREATE_BENCHMARK(uint8_t, rocprim::plus<uint8_t>)
+    CREATE_BENCHMARK(rocprim::half, rocprim::plus<rocprim::half>)
+    CREATE_BENCHMARK(rocprim::int128_t, rocprim::plus<rocprim::int128_t>)
+    CREATE_BENCHMARK(rocprim::uint128_t, rocprim::plus<rocprim::uint128_t>)
+}
+
 #endif // BENCHMARK_CONFIG_TUNING
 
 #endif // ROCPRIM_BENCHMARK_DEVICE_SCAN_PARALLEL_HPP_
