@@ -1,5 +1,5 @@
 #!/usr/bin/python3
-""" Copyright (c) 2021-2023 Advanced Micro Devices, Inc. All rights reserved.
+""" Copyright (c) 2021-2025 Advanced Micro Devices, Inc. All rights reserved.
 Manage build and installation"""
 
 import re
@@ -75,7 +75,6 @@ def delete_dir(dir_path) :
         run_cmd( "RMDIR" , f"/S /Q {dir_path}")
     else:
         linux_path = pathlib.Path(dir_path).absolute()
-        #print( linux_path )
         run_cmd( "rm" , f"-rf {linux_path}")
 
 def cmake_path(os_path):
@@ -146,7 +145,12 @@ def config_cmd():
         deps_dir = os.path.abspath(os.path.join(build_dir, 'deps')).replace('\\','/')
     else:
         deps_dir = args.deps_dir
-    cmake_base_options = f"-DROCM_PATH={rocm_path} -DCMAKE_PREFIX_PATH:PATH={rocm_path[:-1]};{rocm_cmake_path[1:]}" # -DCMAKE_INSTALL_PREFIX=rocmath-install" #-DCMAKE_INSTALL_LIBDIR=
+
+    if (OS_info["ID"] == 'windows'):
+        cmake_base_options = f"-DROCM_PATH={rocm_path} -DCMAKE_PREFIX_PATH:PATH={rocm_path[:-1]};{rocm_cmake_path[1:]}" # -DCMAKE_INSTALL_PREFIX=rocmath-install" #-DCMAKE_INSTALL_LIBDIR=
+    else:
+        cmake_base_options = f"-DROCM_PATH={rocm_path} -DCMAKE_PREFIX_PATH:PATH={rocm_path[:-1]},{rocm_cmake_path[1:-1]}" # -DCMAKE_INSTALL_PREFIX=rocmath-install" #-DCMAKE_INSTALL_LIBDIR=
+    
     cmake_options.append( cmake_base_options )
 
     print( cmake_options )
@@ -211,7 +215,7 @@ def make_cmd():
           make_options.append( "--target package --target install" )
     else:
         nproc = OS_info["NUM_PROC"]
-        make_executable = f"make -j{nproc}"
+        make_executable = f"make -j {nproc}"
         if args.verbose:
           make_options.append( "VERBOSE=1" )
         if args.install:
