@@ -1,4 +1,4 @@
-// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -24,10 +24,15 @@
 #include "../config.hpp"
 #include "../detail/various.hpp"
 
-#include "../intrinsics.hpp"
 #include "../functional.hpp"
+#include "../intrinsics.hpp"
+#include "../intrinsics/arch.hpp"
 #include "../types.hpp"
 
+#include "config.hpp"
+
+#include <cstddef>
+
 /// \addtogroup blockmodule
 /// @{
 
@@ -36,9 +41,10 @@ BEGIN_ROCPRIM_NAMESPACE
 /// \brief The \p block_exchange class is a block level parallel primitive which provides
 /// methods for rearranging items partitioned across threads in a block.
 ///
-/// \tparam T - the input type.
-/// \tparam BlockSize - the number of threads in a block.
-/// \tparam ItemsPerThread - the number of items contributed by each thread.
+/// \tparam T the input type.
+/// \tparam BlockSize the number of threads in a block.
+/// \tparam ItemsPerThread the number of items contributed by each thread.
+/// \tparam PaddingHint a hint that decides when to use padding. May not always be applicable.
 ///
 /// \par Overview
 /// * The \p block_exchange class supports the following rearrangement methods:
@@ -71,33 +77,58 @@ BEGIN_ROCPRIM_NAMESPACE
 /// }
 /// \endcode
 /// \endparblock
-template<
-    class T,
-    unsigned int BlockSizeX,
-    unsigned int ItemsPerThread,
-    unsigned int BlockSizeY = 1,
-    unsigned int BlockSizeZ = 1
->
+template<class T,
+         unsigned int                       BlockSizeX,
+         unsigned int                       ItemsPerThread,
+         unsigned int                       BlockSizeY  = 1,
+         unsigned int                       BlockSizeZ  = 1,
+         block_padding_hint                 PaddingHint = block_padding_hint::avoid_conflicts,
+         ::rocprim::arch::wavefront::target TargetWaveSize
+         = ::rocprim::arch::wavefront::get_target()>
 class block_exchange
 {
     static constexpr unsigned int BlockSize = BlockSizeX * BlockSizeY * BlockSizeZ;
     // Select warp size
-    static constexpr unsigned int warp_size =
-        detail::get_min_warp_size(BlockSize, ::rocprim::device_warp_size());
+    static constexpr unsigned int warp_size = ::rocprim::detail::get_min_warp_size(
+        BlockSize, ::rocprim::arch::wavefront::size_from_target<TargetWaveSize>());
     // Number of warps in block
-    static constexpr unsigned int warps_no = (BlockSize + warp_size - 1) / warp_size;
-
-    // Minimize LDS bank conflicts for power-of-two strides, i.e. when items accessed
-    // using `thread_id * ItemsPerThread` pattern where ItemsPerThread is power of two
-    // (all exchanges from/to blocked).
-    static constexpr bool has_bank_conflicts =
-        ItemsPerThread >= 2 && ::rocprim::detail::is_power_of_two(ItemsPerThread);
+    static constexpr unsigned int warps_no = ::rocprim::detail::ceiling_div(BlockSize, warp_size);
     static constexpr unsigned int banks_no = ::rocprim::detail::get_lds_banks_no();
-    static constexpr unsigned int bank_conflicts_padding =
-        has_bank_conflicts ? (BlockSize * ItemsPerThread / banks_no) : 0;
+    static constexpr unsigned int buffer_size
+        = static_cast<unsigned int>(rocprim::max(size_t{1}, size_t{4} / sizeof(T)));
 
-    static constexpr unsigned int storage_count
-        = BlockSize * ItemsPerThread + bank_conflicts_padding;
+    struct unpadded_config
+    {
+        static constexpr bool         has_bank_conflicts = false;
+        static constexpr unsigned int padding            = 0;
+    };
+
+    struct padded_config
+    {
+        // Minimize LDS bank conflicts for power-of-two strides, i.e. when items accessed
+        // using `thread_id * ItemsPerThread` pattern where ItemsPerThread is power of two
+        // (all exchanges from/to blocked).
+        static constexpr bool has_bank_conflicts
+            = ItemsPerThread >= 2 && ::rocprim::detail::is_power_of_two(ItemsPerThread);
+        static constexpr unsigned int padding
+            = has_bank_conflicts ? (BlockSize * ItemsPerThread / banks_no) : 0;
+    };
+
+    template<typename Config>
+    struct build_config : Config
+    {
+        static constexpr unsigned int storage_count = BlockSize * ItemsPerThread + Config::padding;
+        static constexpr unsigned int storage_size  = sizeof(T) * storage_count;
+        static constexpr unsigned int occupancy     = detail::get_min_lds_size() / storage_size;
+    };
+
+    using config = detail::select_block_padding_config<PaddingHint,
+                                                       build_config<padded_config>,
+                                                       build_config<unpadded_config>>;
+
+    static constexpr bool         has_bank_conflicts     = config::has_bank_conflicts;
+    static constexpr unsigned int bank_conflicts_padding = config::padding;
+    static constexpr unsigned int storage_count          = config::storage_count;
 
     struct storage_type_
     {
@@ -118,10 +149,10 @@ public:
     /// \brief Transposes a blocked arrangement of items to a striped arrangement
     /// across the thread block.
     ///
-    /// \tparam U - [inferred] the output type.
+    /// \tparam U [inferred] the output type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
     template<class U>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE
     void blocked_to_striped(const T (&input)[ItemsPerThread],
@@ -134,11 +165,11 @@ public:
     /// \brief Transposes a blocked arrangement of items to a striped arrangement
     /// across the thread block, using temporary storage.
     ///
-    /// \tparam U - [inferred] the output type.
+    /// \tparam U [inferred] the output type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
     ///
     /// \par Storage reusage
     /// Synchronization barrier should be placed before \p storage is reused
@@ -185,10 +216,10 @@ public:
     /// \brief Transposes a striped arrangement of items to a blocked arrangement
     /// across the thread block.
     ///
-    /// \tparam U - [inferred] the output type.
+    /// \tparam U [inferred] the output type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
     template<class U>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE
     void striped_to_blocked(const T (&input)[ItemsPerThread],
@@ -201,11 +232,11 @@ public:
     /// \brief Transposes a striped arrangement of items to a blocked arrangement
     /// across the thread block, using temporary storage.
     ///
-    /// \tparam U - [inferred] the output type.
+    /// \tparam U [inferred] the output type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
     ///
     /// \par Storage reusage
     /// Synchronization barrier should be placed before \p storage is reused
@@ -252,10 +283,10 @@ public:
     /// \brief Transposes a blocked arrangement of items to a warp-striped arrangement
     /// across the thread block.
     ///
-    /// \tparam U - [inferred] the output type.
+    /// \tparam U [inferred] the output type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
     template<class U>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE
     void blocked_to_warp_striped(const T (&input)[ItemsPerThread],
@@ -268,11 +299,11 @@ public:
     /// \brief Transposes a blocked arrangement of items to a warp-striped arrangement
     /// across the thread block, using temporary storage.
     ///
-    /// \tparam U - [inferred] the output type.
+    /// \tparam U [inferred] the output type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
     ///
     /// \par Storage reusage
     /// Synchronization barrier should be placed before \p storage is reused
@@ -323,10 +354,10 @@ public:
     /// \brief Transposes a warp-striped arrangement of items to a blocked arrangement
     /// across the thread block.
     ///
-    /// \tparam U - [inferred] the output type.
+    /// \tparam U [inferred] the output type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
     template<class U>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE
     void warp_striped_to_blocked(const T (&input)[ItemsPerThread],
@@ -339,11 +370,11 @@ public:
     /// \brief Transposes a warp-striped arrangement of items to a blocked arrangement
     /// across the thread block, using temporary storage.
     ///
-    /// \tparam U - [inferred] the output type.
+    /// \tparam U [inferred] the output type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
     ///
     /// \par Storage reusage
     /// Synchronization barrier should be placed before \p storage is reused
@@ -394,12 +425,12 @@ public:
     /// \brief Scatters items to a blocked arrangement based on their ranks
     /// across the thread block.
     ///
-    /// \tparam U - [inferred] the output type.
-    /// \tparam Offset - [inferred] the rank type.
+    /// \tparam U [inferred] the output type.
+    /// \tparam Offset [inferred] the rank type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
-    /// \param [out] ranks - array that has rank of data.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [out] ranks array that has rank of data.
     template<class U, class Offset>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE
     void scatter_to_blocked(const T (&input)[ItemsPerThread],
@@ -413,12 +444,12 @@ public:
     /// \brief Gathers items from a striped arrangement based on their ranks
     /// across the thread block.
     ///
-    /// \tparam U - [inferred] the output type.
-    /// \tparam Offset - [inferred] the rank type.
+    /// \tparam U [inferred] the output type.
+    /// \tparam Offset [inferred] the rank type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
-    /// \param [out] ranks - array that has rank of data.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [out] ranks array that has rank of data.
     template<class U, class Offset>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE
     void gather_from_striped(const T (&input)[ItemsPerThread],
@@ -432,13 +463,13 @@ public:
     /// \brief Scatters items to a blocked arrangement based on their ranks
     /// across the thread block, using temporary storage.
     ///
-    /// \tparam U - [inferred] the output type.
-    /// \tparam Offset - [inferred] the rank type.
+    /// \tparam U [inferred] the output type.
+    /// \tparam Offset [inferred] the rank type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
-    /// \param [out] ranks - array that has rank of data.
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [out] ranks array that has rank of data.
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
     ///
     /// \par Storage reusage
     /// Synchronization barrier should be placed before \p storage is reused
@@ -488,13 +519,13 @@ public:
     /// \brief Gathers items from a striped arrangement based on their ranks
     /// across the thread block, using temporary storage.
     ///
-    /// \tparam U - [inferred] the output type.
-    /// \tparam Offset - [inferred] the rank type.
+    /// \tparam U [inferred] the output type.
+    /// \tparam Offset [inferred] the rank type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
-    /// \param [out] ranks - array that has rank of data.
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [out] ranks array that has rank of data.
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
     template <class U, class Offset>
     ROCPRIM_DEVICE ROCPRIM_INLINE
     void gather_from_striped(const T (&input)[ItemsPerThread],
@@ -522,12 +553,12 @@ public:
     /// \brief Scatters items to a striped arrangement based on their ranks
     /// across the thread block.
     ///
-    /// \tparam U - [inferred] the output type.
-    /// \tparam Offset - [inferred] the rank type.
+    /// \tparam U [inferred] the output type.
+    /// \tparam Offset [inferred] the rank type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
-    /// \param [out] ranks - array that has rank of data.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [out] ranks array that has rank of data.
     template<class U, class Offset>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE
     void scatter_to_striped(const T (&input)[ItemsPerThread],
@@ -541,13 +572,13 @@ public:
     /// \brief Scatters items to a striped arrangement based on their ranks
     /// across the thread block, using temporary storage.
     ///
-    /// \tparam U - [inferred] the output type.
-    /// \tparam Offset - [inferred] the rank type.
+    /// \tparam U [inferred] the output type.
+    /// \tparam Offset [inferred] the rank type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
-    /// \param [out] ranks - array that has rank of data.
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [out] ranks array that has rank of data.
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
     ///
     /// \par Storage reusage
     /// Synchronization barrier should be placed before \p storage is reused
@@ -594,18 +625,90 @@ public:
         }
     }
 
+    /// \brief Scatters items to a *warp* striped arrangement based on their ranks
+    /// across the thread block, using temporary storage.
+    ///
+    /// \tparam U [inferred] the output type.
+    /// \tparam Offset [inferred] the rank type.
+    ///
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [out] ranks array that has rank of data.
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
+    ///
+    /// \par Storage reusage
+    /// Synchronization barrier should be placed before \p storage is reused
+    /// or repurposed: \p __syncthreads() or \p rocprim::syncthreads().
+    ///
+    /// \par Example.
+    /// \code{.cpp}
+    /// __global__ void example_kernel(...)
+    /// {
+    ///     // specialize block_exchange for int, block of 128 threads and 8 items per thread
+    ///     using block_exchange_int = rocprim::block_exchange<int, 128, 8>;
+    ///     // allocate storage in shared memory
+    ///     __shared__ block_exchange_int::storage_type storage;
+    ///
+    ///     int items[8];
+    ///     int ranks[8];
+    ///     ...
+    ///     block_exchange_int b_exchange;
+    ///     b_exchange.scatter_to_warp_striped(items, items, ranks, storage);
+    ///     ...
+    /// }
+    /// \endcode
+    template<unsigned int VirtualWaveSize = arch::wavefront::size_from_target<TargetWaveSize>(),
+             class U,
+             class Offset>
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    void scatter_to_warp_striped(const T (&input)[ItemsPerThread],
+                                 U (&output)[ItemsPerThread],
+                                 const Offset (&ranks)[ItemsPerThread],
+                                 storage_type& storage)
+    {
+        static_assert(detail::is_power_of_two(VirtualWaveSize)
+                          && VirtualWaveSize <= arch::wavefront::max_size(),
+                      "VirtualWaveSize must be a power of two and equal or less"
+                      "than the size of hardware warp.");
+        assert(VirtualWaveSize <= arch::wavefront::size());
+
+        const unsigned int flat_id
+            = ::rocprim::flat_block_thread_id<BlockSizeX, BlockSizeY, BlockSizeZ>();
+        const unsigned int thread_id     = detail::logical_lane_id<VirtualWaveSize>();
+        const unsigned int warp_id       = flat_id / VirtualWaveSize;
+        const unsigned int warp_offset   = warp_id * VirtualWaveSize * ItemsPerThread;
+        const unsigned int thread_offset = thread_id + warp_offset;
+
+        ROCPRIM_UNROLL
+        for(unsigned int i = 0; i < ItemsPerThread; i++)
+        {
+            const Offset rank = ranks[i];
+            storage.buffer.emplace(index(rank), input[i]);
+        }
+
+        ::rocprim::syncthreads();
+
+        const auto& storage_buffer = storage.buffer.get_unsafe_array();
+
+        ROCPRIM_UNROLL
+        for(unsigned int i = 0; i < ItemsPerThread; i++)
+        {
+            output[i] = storage_buffer[index(thread_offset + i * VirtualWaveSize)];
+        }
+    }
+
     /// \brief Scatters items to a striped arrangement based on their ranks
     /// across the thread block, guarded by rank.
     ///
     /// \par Overview
     /// * Items with rank -1 are not scattered.
     ///
-    /// \tparam U - [inferred] the output type.
-    /// \tparam Offset - [inferred] the rank type.
+    /// \tparam U [inferred] the output type.
+    /// \tparam Offset [inferred] the rank type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
-    /// \param [in] ranks - array that has rank of data.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [in] ranks array that has rank of data.
     template<class U, class Offset>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE
     void scatter_to_striped_guarded(const T (&input)[ItemsPerThread],
@@ -622,13 +725,13 @@ public:
     /// \par Overview
     /// * Items with rank -1 are not scattered.
     ///
-    /// \tparam U - [inferred] the output type.
-    /// \tparam Offset - [inferred] the rank type.
+    /// \tparam U [inferred] the output type.
+    /// \tparam Offset [inferred] the rank type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
-    /// \param [in] ranks - array that has rank of data.
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [in] ranks array that has rank of data.
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
     ///
     /// \par Storage reusage
     /// Synchronization barrier should be placed before \p storage is reused
@@ -681,14 +784,14 @@ public:
     /// \brief Scatters items to a striped arrangement based on their ranks
     /// across the thread block, with a flag to denote validity.
     ///
-    /// \tparam U - [inferred] the output type.
-    /// \tparam Offset - [inferred] the rank type.
-    /// \tparam ValidFlag - [inferred] the validity flag type.
+    /// \tparam U [inferred] the output type.
+    /// \tparam Offset [inferred] the rank type.
+    /// \tparam ValidFlag [inferred] the validity flag type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
-    /// \param [in] ranks - array that has rank of data.
-    /// \param [in] is_valid - array that has flags to denote validity.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [in] ranks array that has rank of data.
+    /// \param [in] is_valid array that has flags to denote validity.
     template<class U, class Offset, class ValidFlag>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE
     void scatter_to_striped_flagged(const T (&input)[ItemsPerThread],
@@ -704,15 +807,15 @@ public:
     /// across the thread block, with a flag to denote validity, using temporary
     /// storage.
     ///
-    /// \tparam U - [inferred] the output type.
-    /// \tparam Offset - [inferred] the rank type.
-    /// \tparam ValidFlag - [inferred] the validity flag type.
+    /// \tparam U [inferred] the output type.
+    /// \tparam Offset [inferred] the rank type.
+    /// \tparam ValidFlag [inferred] the validity flag type.
     ///
-    /// \param [in] input - array that data is loaded from.
-    /// \param [out] output - array that data is loaded to.
-    /// \param [in] ranks - array that has rank of data.
-    /// \param [in] is_valid - array that has flags to denote validity.
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
+    /// \param [in] input array that data is loaded from.
+    /// \param [out] output array that data is loaded to.
+    /// \param [in] ranks array that has rank of data.
+    /// \param [in] is_valid array that has flags to denote validity.
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
     ///
     /// \par Storage reusage
     /// Synchronization barrier should be placed before \p storage is reused
@@ -780,10 +883,125 @@ private:
     unsigned int index(unsigned int n)
     {
         // Move every 32-bank wide "row" (32 banks * 4 bytes) by one item
-        return has_bank_conflicts ? (n + n / banks_no) : n;
+        return has_bank_conflicts ? (n + (n / (banks_no * buffer_size)) * buffer_size) : n;
     }
 };
 
+#ifndef DOXYGEN_SHOULD_SKIP_THIS
+
+template<typename T,
+         unsigned int       BlockSizeX,
+         unsigned int       ItemsPerThread,
+         unsigned int       BlockSizeY,
+         unsigned int       BlockSizeZ,
+         block_padding_hint PaddingHint>
+class block_exchange<T,
+                     BlockSizeX,
+                     ItemsPerThread,
+                     BlockSizeY,
+                     BlockSizeZ,
+                     PaddingHint,
+                     ::rocprim::arch::wavefront::target::dynamic>
+{
+private:
+    using block_exchange_wave32 = block_exchange<T,
+                                                 BlockSizeX,
+                                                 ItemsPerThread,
+                                                 BlockSizeY,
+                                                 BlockSizeZ,
+                                                 PaddingHint,
+                                                 ::rocprim::arch::wavefront::target::size32>;
+    using block_exchange_wave64 = block_exchange<T,
+                                                 BlockSizeX,
+                                                 ItemsPerThread,
+                                                 BlockSizeY,
+                                                 BlockSizeZ,
+                                                 PaddingHint,
+                                                 ::rocprim::arch::wavefront::target::size64>;
+    using dispatch
+        = ::rocprim::detail::dispatch_wave_size<block_exchange_wave32, block_exchange_wave64>;
+
+public:
+    using storage_type = typename dispatch::storage_type;
+
+    template<typename... Args>
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    auto blocked_to_striped(Args&&... args)
+    {
+        dispatch{}([](auto impl, auto&&... args) { impl.blocked_to_striped(args...); }, args...);
+    }
+
+    template<typename... Args>
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    auto striped_to_blocked(Args&&... args)
+    {
+        dispatch{}([](auto impl, auto&&... args) { impl.striped_to_blocked(args...); }, args...);
+    }
+
+    template<typename... Args>
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    auto blocked_to_warp_striped(Args&&... args)
+    {
+        dispatch{}([](auto impl, auto&&... args) { impl.blocked_to_warp_striped(args...); },
+                   args...);
+    }
+
+    template<typename... Args>
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    auto warp_striped_to_blocked(Args&&... args)
+    {
+        dispatch{}([](auto impl, auto&&... args) { impl.warp_striped_to_blocked(args...); },
+                   args...);
+    }
+
+    template<typename... Args>
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    auto scatter_to_blocked(Args&&... args)
+    {
+        dispatch{}([](auto impl, auto&&... args) { impl.scatter_to_blocked(args...); }, args...);
+    }
+
+    template<typename... Args>
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    auto gather_from_striped(Args&&... args)
+    {
+        dispatch{}([](auto impl, auto&&... args) { impl.gather_from_striped(args...); }, args...);
+    }
+
+    template<typename... Args>
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    auto scatter_to_striped(Args&&... args)
+    {
+        dispatch{}([](auto impl, auto&&... args) { impl.scatter_to_striped(args...); }, args...);
+    }
+
+    template<typename... Args>
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    auto scatter_to_warp_striped(Args&&... args)
+    {
+        dispatch{}([](auto impl, auto&&... args) { impl.scatter_to_warp_striped(args...); },
+                   args...);
+    }
+
+    template<typename... Args>
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    auto scatter_to_striped_guarded(Args&&... args)
+    {
+        dispatch{}([](auto impl, auto&&... args) { impl.scatter_to_striped_guarded(args...); },
+                   args...);
+    }
+
+    template<typename... Args>
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    auto scatter_to_striped_flagged(Args&&... args)
+    {
+        dispatch{}([](auto impl, auto&&... args) { impl.scatter_to_striped_flagged(args...); },
+                   args...);
+    }
+};
+
+#endif
+
 END_ROCPRIM_NAMESPACE
 
 /// @}
