@@ -1,7 +1,7 @@
 /******************************************************************************
  * Copyright (c) 2011, Duane Merrill.  All rights reserved.
  * Copyright (c) 2011-2018, NVIDIA CORPORATION.  All rights reserved.
- * Modifications Copyright (c) 2021-2023, Advanced Micro Devices, Inc.  All rights reserved.
+ * Modifications Copyright (c) 2021-2025, Advanced Micro Devices, Inc.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
@@ -37,6 +37,7 @@
 
 #include "../functional.hpp"
 #include "../intrinsics.hpp"
+#include "../types/uninitialized_array.hpp"
 
 #include "detail/block_reduce_raking_reduce.hpp"
 #include "detail/block_reduce_warp_reduce.hpp"
@@ -49,10 +50,10 @@ BEGIN_ROCPRIM_NAMESPACE
 /// \brief The block_shuffle class is a block level parallel primitive which provides methods
 /// for shuffling data partitioned across a block
 ///
-/// \tparam T - the input/output type.
-/// \tparam BlockSizeX - the number of threads in a block's x dimension, it has no defaults value.
-/// \tparam BlockSizeY - the number of threads in a block's y dimension, defaults to 1.
-/// \tparam BlockSizeZ - the number of threads in a block's z dimension, defaults to 1.
+/// \tparam T the input/output type.
+/// \tparam BlockSizeX the number of threads in a block's x dimension, it has no defaults value.
+/// \tparam BlockSizeY the number of threads in a block's y dimension, defaults to 1.
+/// \tparam BlockSizeZ the number of threads in a block's z dimension, defaults to 1.
 ///
 /// \par Overview
 /// It is commonplace for blocks of threads to rearrange data items between
@@ -92,35 +93,30 @@ class block_shuffle
 {
     static constexpr unsigned int BlockSize = BlockSizeX * BlockSizeY * BlockSizeZ;
 
-    // Struct used for creating a raw_storage object for this primitive's temporary storage.
     struct storage_type_
     {
-        T buffer[BlockSize];
+        uninitialized_array<T, BlockSize> buffer;
     };
 
 public:
-/// \brief Struct used to allocate a temporary memory that is required for thread
-/// communication during operations provided by related parallel primitive.
-///
-/// Depending on the implemention the operations exposed by parallel primitive may
-/// require a temporary storage for thread communication. The storage should be allocated
-/// using keywords <tt>__shared__</tt>. It can be aliased to
-/// an externally allocated memory, or be a part of a union type with other storage types
-/// to increase shared memory reusability.
-#ifndef DOXYGEN_SHOULD_SKIP_THIS // hides storage_type implementation for Doxygen
-    using storage_type = detail::raw_storage<storage_type_>;
-#else
+    /// \brief Struct used to allocate a temporary memory that is required for thread
+    /// communication during operations provided by related parallel primitive.
+    ///
+    /// Depending on the implementation the operations exposed by parallel primitive may
+    /// require a temporary storage for thread communication. The storage should be allocated
+    /// using keywords <tt>__shared__</tt>. It can be aliased to
+    /// an externally allocated memory, or be a part of a union type with other storage types
+    /// to increase shared memory reusability.
     using storage_type = storage_type_; // only for Doxygen
-#endif
 
     /// \brief Shuffles data across threads in a block, offseted by the distance value.
     ///
     /// \par A thread with threadIdx i receives data from a thread with threadIdx (i - distance), where distance may be a negative value.
     /// \par Any shuffle operation with invalid input or output threadIds are not carried out, i.e. threadId < 0 || threadId >= BlockSize.
     ///
-    /// \param [in] input - input data to be shuffled to another thread.
-    /// \param [out] output - reference to a output value, that receives data from another thread
-    /// \param [in] distance - The input threadId + distance = output threadId.
+    /// \param [in] input input data to be shuffled to another thread.
+    /// \param [out] output reference to a output value, that receives data from another thread
+    /// \param [in] distance The input threadId + distance = output threadId.
     ///
     /// \par Example.
     /// \code{.cpp}
@@ -151,10 +147,10 @@ public:
     /// \par A thread with threadIdx i receives data from a thread with threadIdx (i - distance), where distance may be a negative value.
     /// \par Any shuffle operation with invalid input or output threadIds are not carried out, i.e. threadId < 0 || threadId >= BlockSize.
     ///
-    /// \param [in] flat_id - flat thread ID obtained from rocprim::flat_block_thread_id
-    /// \param [in] input - input data to be shuffled to another thread.
-    /// \param [out] output - reference to a output value, that receives data from another thread
-    /// \param [in] distance - The input threadId + distance = output threadId.
+    /// \param [in] flat_id flat thread ID obtained from rocprim::flat_block_thread_id
+    /// \param [in] input input data to be shuffled to another thread.
+    /// \param [out] output reference to a output value, that receives data from another thread
+    /// \param [in] distance The input threadId + distance = output threadId.
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE void
         offset(const size_t& flat_id, T input, T& output, int distance)
     {
@@ -167,24 +163,24 @@ public:
     /// \par A thread with threadIdx i receives data from a thread with threadIdx (i - distance), where distance may be a negative value.
     /// \par Any shuffle operation with invalid input or output threadIds are not carried out, i.e. threadId < 0 || threadId >= BlockSize.
     ///
-    /// \param [in] flat_id - flat thread ID obtained from rocprim::flat_block_thread_id
-    /// \param [in] input - input data to be shuffled to another thread.
-    /// \param [out] output - reference to a output value, that receives data from another thread
-    /// \param [in] distance - The input threadId + distance = output threadId.
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
-    ROCPRIM_DEVICE ROCPRIM_INLINE void
-        offset(const size_t& flat_id, T input, T& output, int distance, storage_type& storage)
+    /// \param [in] flat_id flat thread ID obtained from rocprim::flat_block_thread_id
+    /// \param [in] input input data to be shuffled to another thread.
+    /// \param [out] output reference to a output value, that receives data from another thread
+    /// \param [in] distance The input threadId + distance = output threadId.
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    void offset(const size_t& flat_id, T input, T& output, int distance, storage_type& storage)
     {
-        storage_type_& storage_  = storage.get();
-        storage_.buffer[flat_id] = input;
+        storage.buffer.emplace(flat_id, input);
 
         const int offset_tid = static_cast<int>(flat_id) + distance;
 
         ::rocprim::syncthreads();
+        const auto& storage_buffer = storage.buffer.get_unsafe_array();
 
         if((offset_tid >= 0) && (offset_tid < (int)BlockSize))
         {
-            output = storage_.buffer[static_cast<size_t>(offset_tid)];
+            output = storage_buffer[static_cast<size_t>(offset_tid)];
         }
     }
 
@@ -193,9 +189,9 @@ public:
     /// \par A thread with threadIdx i receives data from a thread with threadIdx (i - distance) % BlockSize, where distance may be a negative value.
     /// \par Data is rotated around the block, using (input_threadId + distance) modulous BlockSize to ensure valid threadIds.
     ///
-    /// \param [in] input - input data to be shuffled to another thread.
-    /// \param [out] output - reference to a output value, that receives data from another thread
-    /// \param [in] distance - The input threadId + distance = output threadId. Distance magnitude should be <= BlockSize.
+    /// \param [in] input input data to be shuffled to another thread.
+    /// \param [out] output reference to a output value, that receives data from another thread
+    /// \param [in] distance The input threadId + distance = output threadId. Distance magnitude should be <= BlockSize.
     ///
     /// \par Example.
     /// \code{.cpp}
@@ -226,10 +222,10 @@ public:
     /// \par A thread with threadIdx i receives data from a thread with threadIdx (i - distance) % BlockSize, where distance may be a negative value.
     /// \par Data is rotated around the block, using (input_threadId + distance) modulous BlockSize to ensure valid threadIds.
     ///
-    /// \param [in] flat_id - flat thread ID obtained from rocprim::flat_block_thread_id
-    /// \param [in] input - input data to be shuffled to another thread.
-    /// \param [out] output - reference to a output value, that receives data from another thread
-    /// \param [in] distance - The input threadId + distance = output threadId.
+    /// \param [in] flat_id flat thread ID obtained from rocprim::flat_block_thread_id
+    /// \param [in] input input data to be shuffled to another thread.
+    /// \param [out] output reference to a output value, that receives data from another thread
+    /// \param [in] distance The input threadId + distance = output threadId.
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE void
         rotate(const size_t& flat_id, T input, T& output, int distance)
     {
@@ -242,16 +238,15 @@ public:
     /// \par A thread with threadIdx i receives data from a thread with threadIdx (i - distance) % BlockSize, where distance may be a negative value.
     /// \par Data is rotated around the block, using (input_threadId + distance) modulous BlockSize to ensure valid threadIds.
     ///
-    /// \param [in] flat_id - flat thread ID obtained from rocprim::flat_block_thread_id
-    /// \param [in] input - input data to be shuffled to another thread.
-    /// \param [out] output - reference to a output value, that receives data from another thread
-    /// \param [in] distance - The input threadId + distance = output threadId.
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
-    ROCPRIM_DEVICE ROCPRIM_INLINE void
-        rotate(const size_t& flat_id, T input, T& output, int distance, storage_type& storage)
+    /// \param [in] flat_id flat thread ID obtained from rocprim::flat_block_thread_id
+    /// \param [in] input input data to be shuffled to another thread.
+    /// \param [out] output reference to a output value, that receives data from another thread
+    /// \param [in] distance The input threadId + distance = output threadId.
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    void rotate(const size_t& flat_id, T input, T& output, int distance, storage_type& storage)
     {
-        storage_type_& storage_  = storage.get();
-        storage_.buffer[flat_id] = input;
+        storage.buffer.emplace(flat_id, input);
 
         int offset = static_cast<int>(flat_id) + distance;
         if(offset >= (int)BlockSize)
@@ -264,15 +259,16 @@ public:
         }
 
         ::rocprim::syncthreads();
+        const auto& storage_buffer = storage.buffer.get_unsafe_array();
 
-        output = storage_.buffer[offset];
+        output = storage_buffer[offset];
     }
 
     /// \brief The thread block rotates a blocked arrange of input items,
     /// shifting it up by one item
     ///
-    /// \param [in]  input -  The calling thread's input items
-    /// \param [out] prev  -  The corresponding predecessor items (may be aliased to \p input).
+    /// \param [in]  input  The calling thread's input items
+    /// \param [out] prev   The corresponding predecessor items (may be aliased to \p input).
     /// The item \p prev[0] is not updated for <em>thread</em><sub>0</sub>.
     ///
     /// \par Example.
@@ -303,9 +299,9 @@ public:
     /// \brief The thread block rotates a blocked arrange of input items,
     /// shifting it up by one item
     ///
-    /// \param [in] flat_id - flat thread ID obtained from rocprim::flat_block_thread_id
-    /// \param [in]  input -  The calling thread's input items
-    /// \param [out] prev  -  The corresponding predecessor items (may be aliased to \p input).
+    /// \param [in] flat_id flat thread ID obtained from rocprim::flat_block_thread_id
+    /// \param [in]  input  The calling thread's input items
+    /// \param [out] prev   The corresponding predecessor items (may be aliased to \p input).
     /// The item \p prev[0] is not updated for <em>thread</em><sub>0</sub>.
     template<unsigned int ItemsPerThread>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE void
@@ -318,19 +314,19 @@ public:
     /// \brief The thread block rotates a blocked arrange of input items,
     /// shifting it up by one item, using temporary storage.
     ///
-    /// \param [in] flat_id - flat thread ID obtained from rocprim::flat_block_thread_id
-    /// \param [in]  input -  The calling thread's input items
-    /// \param [out] prev  -  The corresponding predecessor items (may be aliased to \p input).
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
+    /// \param [in] flat_id flat thread ID obtained from rocprim::flat_block_thread_id
+    /// \param [in]  input  The calling thread's input items
+    /// \param [out] prev   The corresponding predecessor items (may be aliased to \p input).
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
     /// The item \p prev[0] is not updated for <em>thread</em><sub>0</sub>.
     template<unsigned int ItemsPerThread>
-    ROCPRIM_DEVICE ROCPRIM_INLINE void up(const size_t& flat_id,
-                                          T (&input)[ItemsPerThread],
-                                          T (&prev)[ItemsPerThread],
-                                          storage_type& storage)
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    void up(const size_t& flat_id,
+            T (&input)[ItemsPerThread],
+            T (&prev)[ItemsPerThread],
+            storage_type& storage)
     {
-        storage_type_& storage_  = storage.get();
-        storage_.buffer[flat_id] = input[ItemsPerThread - 1];
+        storage.buffer.emplace(flat_id, input[ItemsPerThread - 1]);
 
         ROCPRIM_UNROLL
         for(unsigned int i = ItemsPerThread - 1; i > 0; --i)
@@ -339,20 +335,21 @@ public:
         }
 
         ::rocprim::syncthreads();
+        const auto& storage_buffer = storage.buffer.get_unsafe_array();
 
         if(flat_id > 0)
         {
-            prev[0] = storage_.buffer[flat_id - 1];
+            prev[0] = storage_buffer[flat_id - 1];
         }
     }
 
     /// \brief The thread block rotates a blocked arrange of input items,
     /// shifting it up by one item
     ///
-    /// \param [in]  input - The calling thread's input items
-    /// \param [out] prev  - The corresponding predecessor items (may be aliased to \p input).
+    /// \param [in]  input The calling thread's input items
+    /// \param [out] prev  The corresponding predecessor items (may be aliased to \p input).
     /// The item \p prev[0] is not updated for <em>thread</em><sub>0</sub>.
-    /// \param [out] block_suffix - The item \p input[ItemsPerThread-1] from
+    /// \param [out] block_suffix The item \p input[ItemsPerThread-1] from
     /// <em>thread</em><sub><tt>BlockSize-1</tt></sub>, provided to all threads
     template<unsigned int ItemsPerThread>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE void
@@ -367,11 +364,11 @@ public:
     /// \brief The thread block rotates a blocked arrange of input items,
     /// shifting it up by one item
     ///
-    /// \param [in] flat_id - flat thread ID obtained from rocprim::flat_block_thread_id
-    /// \param [in]  input - The calling thread's input items
-    /// \param [out] prev  - The corresponding predecessor items (may be aliased to \p input).
+    /// \param [in] flat_id flat thread ID obtained from rocprim::flat_block_thread_id
+    /// \param [in]  input The calling thread's input items
+    /// \param [out] prev  The corresponding predecessor items (may be aliased to \p input).
     /// The item \p prev[0] is not updated for <em>thread</em><sub>0</sub>.
-    /// \param [out] block_suffix - The item \p input[ItemsPerThread-1] from
+    /// \param [out] block_suffix The item \p input[ItemsPerThread-1] from
     /// <em>thread</em><sub><tt>BlockSize-1</tt></sub>, provided to all threads
     template<unsigned int ItemsPerThread>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE void up(const size_t& flat_id,
@@ -386,31 +383,32 @@ public:
     /// \brief The thread block rotates a blocked arrange of input items,
     /// shifting it up by one item, using temporary storage.
     ///
-    /// \param [in] flat_id - flat thread ID obtained from rocprim::flat_block_thread_id
-    /// \param [in]  input - The calling thread's input items
-    /// \param [out] prev  - The corresponding predecessor items (may be aliased to \p input).
+    /// \param [in] flat_id flat thread ID obtained from rocprim::flat_block_thread_id
+    /// \param [in]  input The calling thread's input items
+    /// \param [out] prev  The corresponding predecessor items (may be aliased to \p input).
     /// The item \p prev[0] is not updated for <em>thread</em><sub>0</sub>.
-    /// \param [out] block_suffix - The item \p input[ItemsPerThread-1] from
+    /// \param [out] block_suffix The item \p input[ItemsPerThread-1] from
     /// <em>thread</em><sub><tt>BlockSize-1</tt></sub>, provided to all threads
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
     template<int ItemsPerThread>
-    ROCPRIM_DEVICE ROCPRIM_INLINE void up(const size_t& flat_id,
-                                          T (&input)[ItemsPerThread],
-                                          T (&prev)[ItemsPerThread],
-                                          T&            block_suffix,
-                                          storage_type& storage)
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    void up(const size_t& flat_id,
+            T (&input)[ItemsPerThread],
+            T (&prev)[ItemsPerThread],
+            T&            block_suffix,
+            storage_type& storage)
     {
         up(flat_id, input, prev, storage);
 
         // Update block prefix
-        block_suffix = storage->buffer[BlockSize - 1];
+        block_suffix = storage.buffer.get_unsafe_array()[BlockSize - 1];
     }
 
     /// \brief The thread block rotates a blocked arrange of input items,
     /// shifting it down by one item
     ///
-    /// \param [in]  input -  The calling thread's input items
-    /// \param [out] next  -  The corresponding successor items (may be aliased to \p input).
+    /// \param [in]  input  The calling thread's input items
+    /// \param [out] next   The corresponding successor items (may be aliased to \p input).
     /// The item \p prev[0] is not updated for <em>thread</em><sub>BlockSize - 1</sub>.
     ///
     /// \par Example.
@@ -441,9 +439,9 @@ public:
     /// \brief The thread block rotates a blocked arrange of input items,
     /// shifting it down by one item
     ///
-    /// \param [in] flat_id - flat thread ID obtained from rocprim::flat_block_thread_id
-    /// \param [in]  input -  The calling thread's input items
-    /// \param [out] next  -  The corresponding successor items (may be aliased to \p input).
+    /// \param [in] flat_id flat thread ID obtained from rocprim::flat_block_thread_id
+    /// \param [in]  input  The calling thread's input items
+    /// \param [out] next   The corresponding successor items (may be aliased to \p input).
     /// The item \p prev[0] is not updated for <em>thread</em><sub>BlockSize - 1</sub>.
     template<unsigned int ItemsPerThread>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE void
@@ -456,19 +454,19 @@ public:
     /// \brief The thread block rotates a blocked arrange of input items,
     /// shifting it down by one item, using temporary storage.
     ///
-    /// \param [in] flat_id - flat thread ID obtained from rocprim::flat_block_thread_id
-    /// \param [in]  input -  The calling thread's input items
-    /// \param [out] next  -  The corresponding successor items (may be aliased to \p input).
+    /// \param [in] flat_id flat thread ID obtained from rocprim::flat_block_thread_id
+    /// \param [in]  input  The calling thread's input items
+    /// \param [out] next   The corresponding successor items (may be aliased to \p input).
     /// The item \p prev[0] is not updated for <em>thread</em><sub>BlockSize - 1</sub>.
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
     template<unsigned int ItemsPerThread>
-    ROCPRIM_DEVICE ROCPRIM_INLINE void down(const size_t& flat_id,
-                                            T (&input)[ItemsPerThread],
-                                            T (&next)[ItemsPerThread],
-                                            storage_type& storage)
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    void down(const size_t& flat_id,
+              T (&input)[ItemsPerThread],
+              T (&next)[ItemsPerThread],
+              storage_type& storage)
     {
-        storage_type_& storage_  = storage.get();
-        storage_.buffer[flat_id] = input[0];
+        storage.buffer.emplace(flat_id, input[0]);
 
         ROCPRIM_UNROLL
         for(unsigned int i = 0; i < (ItemsPerThread - 1); ++i)
@@ -477,20 +475,21 @@ public:
         }
 
         ::rocprim::syncthreads();
+        const auto& storage_buffer = storage.buffer.get_unsafe_array();
 
         if(flat_id < (BlockSize - 1))
         {
-            next[ItemsPerThread - 1] = storage_.buffer[flat_id + 1];
+            next[ItemsPerThread - 1] = storage_buffer[flat_id + 1];
         }
     }
 
     /// \brief The thread block rotates a blocked arrange of input items,
     /// shifting it down by one item
     ///
-    /// \param [in]  input -  The calling thread's input items
-    /// \param [out] next  -  The corresponding successor items (may be aliased to \p input).
+    /// \param [in]  input  The calling thread's input items
+    /// \param [out] next   The corresponding successor items (may be aliased to \p input).
     /// The item \p prev[0] is not updated for <em>thread</em><sub>BlockSize - 1</sub>.
-    /// \param [out] block_prefix -  The item \p input[0] from <em>thread</em><sub><tt>0</tt></sub>, provided to all threads
+    /// \param [out] block_prefix  The item \p input[0] from <em>thread</em><sub><tt>0</tt></sub>, provided to all threads
     template<unsigned int ItemsPerThread>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE void
         down(T (&input)[ItemsPerThread], T (&next)[ItemsPerThread], T& block_prefix)
@@ -504,11 +503,11 @@ public:
     /// \brief The thread block rotates a blocked arrange of input items,
     /// shifting it down by one item
     ///
-    /// \param [in] flat_id - flat thread ID obtained from rocprim::flat_block_thread_id
-    /// \param [in]  input -  The calling thread's input items
-    /// \param [out] next  -  The corresponding successor items (may be aliased to \p input).
+    /// \param [in] flat_id flat thread ID obtained from rocprim::flat_block_thread_id
+    /// \param [in]  input  The calling thread's input items
+    /// \param [out] next   The corresponding successor items (may be aliased to \p input).
     /// The item \p prev[0] is not updated for <em>thread</em><sub>BlockSize - 1</sub>.
-    /// \param [out] block_prefix -  The item \p input[0] from <em>thread</em><sub><tt>0</tt></sub>, provided to all threads
+    /// \param [out] block_prefix  The item \p input[0] from <em>thread</em><sub><tt>0</tt></sub>, provided to all threads
     template<unsigned int ItemsPerThread>
     ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE void down(const size_t& flat_id,
                                                   T (&input)[ItemsPerThread],
@@ -522,23 +521,24 @@ public:
     /// \brief The thread block rotates a blocked arrange of input items,
     /// shifting it down by one item, using temporary storage.
     ///
-    /// \param [in] flat_id - flat thread ID obtained from rocprim::flat_block_thread_id
-    /// \param [in]  input -  The calling thread's input items
-    /// \param [out] next  -  The corresponding successor items (may be aliased to \p input).
+    /// \param [in] flat_id flat thread ID obtained from rocprim::flat_block_thread_id
+    /// \param [in]  input  The calling thread's input items
+    /// \param [out] next   The corresponding successor items (may be aliased to \p input).
     /// The item \p prev[0] is not updated for <em>thread</em><sub>BlockSize - 1</sub>.
-    /// \param [out] block_prefix -  The item \p input[0] from <em>thread</em><sub><tt>0</tt></sub>, provided to all threads
-    /// \param [in] storage - reference to a temporary storage object of type storage_type.
+    /// \param [out] block_prefix  The item \p input[0] from <em>thread</em><sub><tt>0</tt></sub>, provided to all threads
+    /// \param [in] storage reference to a temporary storage object of type storage_type.
     template<unsigned int ItemsPerThread>
-    ROCPRIM_DEVICE ROCPRIM_INLINE void down(const size_t& flat_id,
-                                            T (&input)[ItemsPerThread],
-                                            T (&next)[ItemsPerThread],
-                                            T&            block_prefix,
-                                            storage_type& storage)
+    ROCPRIM_DEVICE ROCPRIM_INLINE
+    void down(const size_t& flat_id,
+              T (&input)[ItemsPerThread],
+              T (&next)[ItemsPerThread],
+              T&            block_prefix,
+              storage_type& storage)
     {
         this->down(flat_id, input, next, storage);
 
         // Update block prefixstorage_->
-        block_prefix = storage->next[0];
+        block_prefix = storage.buffer.get_unsafe_array()[0];
     }
 };
 
@@ -547,4 +547,4 @@ END_ROCPRIM_NAMESPACE
 /// @}
 // end of group blockmodule
 
-#endif // ROCPRIM_BLOCK_BLOCK_SHUFFLE_HPP_
+#endif // ROCPRIM_BLOCK_BLOCK_SHUFFLE_HPP_
\ No newline at end of file
