@@ -1,4 +1,4 @@
-// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -23,7 +23,6 @@
 
 #include "device_config_helper.hpp"
 #include "lookback_scan_state.hpp"
-#include "ordered_block_id.hpp"
 
 #include "../../block/block_discontinuity.hpp"
 #include "../../block/block_load.hpp"
@@ -34,6 +33,7 @@
 #include "../../thread/thread_operators.hpp"
 
 #include "../../config.hpp"
+#include "../../type_traits.hpp"
 
 #include <iterator>
 #include <type_traits>
@@ -47,12 +47,9 @@ namespace detail
 namespace reduce_by_key
 {
 
-template<typename Iterator>
-using value_type_t = typename std::iterator_traits<Iterator>::value_type;
-
 template<typename ValueIterator, typename BinaryOp>
-using accumulator_type_t =
-    typename invoke_result_binary_op<reduce_by_key::value_type_t<ValueIterator>, BinaryOp>::type;
+using accumulator_type_t
+    = ::rocprim::accumulator_t<BinaryOp, ::rocprim::detail::value_type_t<ValueIterator>>;
 
 template<typename AccumulatorType>
 using wrapped_type_t = rocprim::tuple<unsigned int, AccumulatorType>;
@@ -61,36 +58,6 @@ template<typename AccumulatorType, bool UseSleep = false>
 using lookback_scan_state_t
     = detail::lookback_scan_state<wrapped_type_t<AccumulatorType>, UseSleep>;
 
-template<typename EqualityOp>
-struct guarded_inequality_wrapper
-{
-    /// Wrapped equality operator
-    EqualityOp op;
-
-    /// Out-of-bounds limit
-    size_t guard;
-
-    /// Constructor
-    ROCPRIM_HOST_DEVICE inline guarded_inequality_wrapper(EqualityOp op, size_t guard)
-        : op(op), guard(guard)
-    {}
-
-    /// \brief Guarded boolean inequality operator.
-    ///
-    /// \tparam T Type of the operands compared by the equality operator
-    /// \param a Left hand-side operand
-    /// \param b Right hand-side operand
-    /// \param idx Index of the thread calling to this operator. This is used to determine which
-    /// operations are out-of-bounds
-    /// \returns <tt>!op(a, b)</tt> for a certain equality operator \p op when in-bounds.
-    template<typename T>
-    ROCPRIM_HOST_DEVICE inline bool operator()(const T& a, const T& b, size_t idx) const
-    {
-        // In-bounds return operation result, out-of-bounds return false.
-        return (idx < guard) ? !op(a, b) : 0;
-    }
-};
-
 template<typename KeyType,
          typename AccumulatorType,
          unsigned int      BlockSize,
@@ -102,6 +69,14 @@ struct load_helper
     using block_load_keys = block_load<KeyType, BlockSize, ItemsPerThread, load_keys_method>;
     using block_load_values
         = block_load<AccumulatorType, BlockSize, ItemsPerThread, load_values_method>;
+
+    /// We only need to sync between loading keys & values if BOTH the key and value
+    /// loading method require shared memory.
+    constexpr static bool requires_inner_sync
+        = !std::is_same<typename block_load_keys::storage_type, detail::empty_storage_type>::value
+          && !std::is_same<typename block_load_values::storage_type,
+                           detail::empty_storage_type>::value;
+
     union storage_type
     {
         typename block_load_keys::storage_type   keys;
@@ -109,24 +84,32 @@ struct load_helper
     };
 
     template<typename KeyIterator, typename ValueIterator>
-    ROCPRIM_DEVICE void load_keys_values(KeyIterator        tile_keys,
-                                         ValueIterator      tile_values,
-                                         const bool         is_global_last_tile,
-                                         const unsigned int valid_in_global_last_tile,
-                                         KeyType (&keys)[ItemsPerThread],
-                                         AccumulatorType (&values)[ItemsPerThread],
-                                         storage_type& storage)
+    ROCPRIM_DEVICE
+    void load_keys_values(KeyIterator        tile_keys,
+                          ValueIterator      tile_values,
+                          const bool         is_global_last_tile,
+                          const unsigned int valid_in_global_last_tile,
+                          KeyType (&keys)[ItemsPerThread],
+                          AccumulatorType (&values)[ItemsPerThread],
+                          storage_type& storage)
     {
+
         if(!is_global_last_tile)
         {
             block_load_keys{}.load(tile_keys, keys, storage.keys);
-            ::rocprim::syncthreads();
+            if constexpr(requires_inner_sync)
+            {
+                ::rocprim::syncthreads();
+            }
             block_load_values{}.load(tile_values, values, storage.values);
         }
         else
         {
             block_load_keys{}.load(tile_keys, keys, valid_in_global_last_tile, storage.keys);
-            ::rocprim::syncthreads();
+            if constexpr(requires_inner_sync)
+            {
+                ::rocprim::syncthreads();
+            }
             block_load_values{}.load(tile_values,
                                      values,
                                      valid_in_global_last_tile,
@@ -142,20 +125,22 @@ struct discontinuity_helper
     using storage_type             = typename block_discontinuity_type::storage_type;
 
     template<typename KeyIterator, typename CompareFunction, unsigned int ItemsPerThread>
-    ROCPRIM_DEVICE void flag_heads(KeyIterator tile_keys,
-                                   const KeyType (&keys)[ItemsPerThread],
-                                   CompareFunction compare,
-                                   unsigned int (&head_flags)[ItemsPerThread],
-                                   const bool    is_global_first_tile,
-                                   const bool    is_global_last_tile,
-                                   const size_t  remaining,
-                                   storage_type& storage)
+    ROCPRIM_DEVICE
+    void flag_heads(KeyIterator tile_keys,
+                    const KeyType (&keys)[ItemsPerThread],
+                    CompareFunction compare,
+                    unsigned int (&head_flags)[ItemsPerThread],
+                    const bool    is_global_first_tile,
+                    const bool    is_global_last_tile,
+                    const size_t  remaining,
+                    storage_type& storage)
     {
         if(is_global_last_tile)
         {
             // If it's the last tile globally, the out-of-bound items should not be flagged.
             auto guarded_not_equal
-                = guarded_inequality_wrapper<CompareFunction>(compare, remaining);
+                = ::rocprim::detail::guarded_inequality_wrapper<CompareFunction>(compare,
+                                                                                 remaining);
 
             if(!is_global_first_tile)
             {
@@ -200,24 +185,60 @@ struct scatter_helper
     ROCPRIM_DETAIL_SUPPRESS_DEPRECATION_POP
 
     template<typename ValueIterator, typename Flag, typename ValueFunction, typename IndexFunction>
-    ROCPRIM_DEVICE void scatter(ValueIterator   tile_values,
-                                ValueFunction&& values,
-                                const Flag (&is_selected)[ItemsPerThread],
-                                IndexFunction&&    block_indices,
-                                const unsigned int selected_in_tile,
-                                const unsigned int flat_thread_id,
-                                storage_type&      storage)
+    ROCPRIM_DEVICE
+    void scatter(ValueIterator   tile_values,
+                 ValueFunction&& values,
+                 const Flag (&is_selected)[ItemsPerThread],
+                 IndexFunction&&    block_indices,
+                 const unsigned int selected_in_tile,
+                 const unsigned int flat_thread_id,
+                 storage_type&      storage)
     {
+        // Check if all threads in the warp are selecting the same location (selected or rejected)
+        uint8_t all_check = 3; // [true, true]
+        for(unsigned int i = 0; i < ItemsPerThread; ++i)
+        {
+            all_check &= is_selected[i] ? 1 /* [false, true] */ : 2 /* [true, false] */;
+        }
+        // all_check:
+        //   0: mixed selection
+        //   1: all selected
+        //   2: all rejected
+        //   3: impossible
+
+        // If selected_in_tile >= BlockSize, use shared memory for storing scattered data
         if(selected_in_tile >= BlockSize)
         {
             auto& scatter_storage = storage.get();
-            for(unsigned int i = 0; i < ItemsPerThread; ++i)
+
+            // If all threads in the warp are storing to the same location, we can skip the conditional
+            if(all_check == 1)
             {
-                if(is_selected[i])
+                // Perform the scatter directly for selected items
+                ROCPRIM_UNROLL
+                for(unsigned int i = 0; i < ItemsPerThread; ++i)
                 {
                     scatter_storage[block_indices(i)] = values(i);
                 }
             }
+            else if(all_check == 2)
+            {
+                // If all threads reject, we can skip doing any work for the selected items
+                // (nothing to store to the scatter storage)
+            }
+            else
+            {
+                // Otherwise, process each item individually
+                ROCPRIM_UNROLL
+                for(unsigned int i = 0; i < ItemsPerThread; ++i)
+                {
+                    if(is_selected[i])
+                    {
+                        scatter_storage[block_indices(i)] = values(i);
+                    }
+                }
+            }
+
             ::rocprim::syncthreads();
 
             // Coalesced write from shared memory to global memory
@@ -228,13 +249,32 @@ struct scatter_helper
         }
         else
         {
-            for(unsigned int i = 0; i < ItemsPerThread; ++i)
+            // When selected_in_tile < BlockSize, store directly to global memory
+            if(all_check == 1)
             {
-                if(is_selected[i])
+                // Directly store all values since they are all selected
+                ROCPRIM_UNROLL
+                for(unsigned int i = 0; i < ItemsPerThread; ++i)
                 {
                     tile_values[block_indices(i)] = values(i);
                 }
             }
+            else if(all_check == 2)
+            {
+                // If all are rejected, do nothing
+            }
+            else
+            {
+                // Process selected items
+                ROCPRIM_UNROLL
+                for(unsigned int i = 0; i < ItemsPerThread; ++i)
+                {
+                    if(is_selected[i])
+                    {
+                        tile_values[block_indices(i)] = values(i);
+                    }
+                }
+            }
         }
     }
 };
@@ -494,13 +534,11 @@ ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE auto kernel_impl(KeyIterator,
                                                      const BinaryOp,
                                                      const CompareFunction,
                                                      const LookbackScanState,
-                                                     ordered_block_id<unsigned int>,
                                                      const std::size_t,
                                                      const std::size_t,
                                                      const std::size_t,
                                                      const std::size_t* const,
-                                                     const AccumulatorType* const,
-                                                     const std::size_t)
+                                                     const AccumulatorType* const)
     -> std::enable_if_t<!is_lookback_kernel_runnable<LookbackScanState>()>
 {
     // No need to build the kernel with sleep on a device that does not require it
@@ -526,26 +564,23 @@ ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE auto
                 const BinaryOp                 reduce_op,
                 const CompareFunction          compare,
                 const LookbackScanState        scan_state,
-                ordered_block_id<unsigned int> ordered_tile_id,
-                const std::size_t              starting_tile,
-                const std::size_t              total_number_of_tiles,
+                const std::size_t              starting_block,
+                const std::size_t              total_number_of_blocks,
                 const std::size_t              size,
                 const std::size_t* const       global_head_count,
-                const AccumulatorType* const   previous_accumulated,
-                const std::size_t              number_of_tiles_launch)
+                const AccumulatorType* const   previous_accumulated)
         -> std::enable_if_t<is_lookback_kernel_runnable<LookbackScanState>()>
 {
     static constexpr reduce_by_key_config_params params = device_params<Config>();
 
     static constexpr unsigned int         block_size       = params.kernel_config.block_size;
     static constexpr unsigned int         items_per_thread = params.kernel_config.items_per_thread;
-    static constexpr unsigned int         tiles_per_block  = params.tiles_per_block;
     static constexpr block_load_method    load_keys_method = params.load_keys_method;
     static constexpr block_load_method    load_values_method = params.load_values_method;
     static constexpr block_scan_algorithm scan_algorithm     = params.scan_algorithm;
-    static constexpr unsigned int         items_per_tile     = block_size * items_per_thread;
+    static constexpr unsigned int         items_per_block    = block_size * items_per_thread;
 
-    using key_type = reduce_by_key::value_type_t<KeyIterator>;
+    using key_type = ::rocprim::detail::value_type_t<KeyIterator>;
 
     using tile_processor = tile_helper<key_type,
                                        AccumulatorType,
@@ -558,40 +593,29 @@ ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE auto
 
     ROCPRIM_SHARED_MEMORY union
     {
-        typename decltype(ordered_tile_id)::storage_type tile_id;
-        typename tile_processor::storage_type            tile;
+        typename tile_processor::storage_type tile;
     } storage;
 
-    for(unsigned int i = 0; i < tiles_per_block; ++i)
-    {
-        rocprim::syncthreads();
-        const std::size_t tile_id = ordered_tile_id.get(threadIdx.x, storage.tile_id);
-        if(tile_id >= number_of_tiles_launch)
-        {
-            return;
-        }
+    const unsigned int  block_id     = rocprim::flat_block_id<block_size, 1, 1>();
+    const unsigned int  block_offset = block_id * items_per_block;
+    const KeyIterator   block_keys   = keys_input + block_offset;
+    const ValueIterator block_values = values_input + block_offset;
 
-        const std::size_t   tile_offset = tile_id * items_per_tile;
-        const KeyIterator   tile_keys   = keys_input + tile_offset;
-        const ValueIterator tile_values = values_input + tile_offset;
-
-        rocprim::syncthreads();
-        tile_processor{}.process_tile(tile_keys,
-                                      tile_values,
-                                      unique_keys,
-                                      reductions,
-                                      unique_count,
-                                      reduce_op,
-                                      compare,
-                                      scan_state,
-                                      tile_id,
-                                      starting_tile,
-                                      total_number_of_tiles,
-                                      size,
-                                      storage.tile,
-                                      global_head_count,
-                                      previous_accumulated);
-    }
+    tile_processor{}.process_tile(block_keys,
+                                  block_values,
+                                  unique_keys,
+                                  reductions,
+                                  unique_count,
+                                  reduce_op,
+                                  compare,
+                                  scan_state,
+                                  block_id,
+                                  starting_block,
+                                  total_number_of_blocks,
+                                  size,
+                                  storage.tile,
+                                  global_head_count,
+                                  previous_accumulated);
 }
 
 } // namespace reduce_by_key
