@@ -1,4 +1,4 @@
-// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -27,6 +27,7 @@
 #include <type_traits>
 
 #include "../config.hpp"
+#include "../common.hpp"
 #include "../detail/various.hpp"
 #include "../functional.hpp"
 
@@ -42,11 +43,9 @@ namespace detail
 {
 
 template<class Config, unsigned int ActiveChannels, class Counter>
-ROCPRIM_KERNEL __launch_bounds__(
-    device_params<Config>()
-        .histogram_config
-        .block_size) void init_histogram_kernel(fixed_array<Counter*, ActiveChannels>     histogram,
-                                                fixed_array<unsigned int, ActiveChannels> bins)
+ROCPRIM_KERNEL ROCPRIM_LAUNCH_BOUNDS(device_params<Config>().histogram_config.block_size) void
+    init_histogram_kernel(fixed_array<Counter*, ActiveChannels>           histogram,
+                          const fixed_array<unsigned int, ActiveChannels> bins)
 {
     static constexpr histogram_config_params params = device_params<Config>();
 
@@ -59,23 +58,25 @@ template<class Config,
          class SampleIterator,
          class Counter,
          class SampleToBinOp>
-ROCPRIM_KERNEL __launch_bounds__(
-    device_params<Config>()
-        .histogram_config
-        .block_size) void histogram_shared_kernel(SampleIterator samples,
-                                                  unsigned int   columns,
-                                                  unsigned int   rows,
-                                                  unsigned int   row_stride,
-                                                  unsigned int   rows_per_block,
-                                                  unsigned int   shared_histograms,
-                                                  fixed_array<Counter*, ActiveChannels> histogram,
-                                                  fixed_array<SampleToBinOp, ActiveChannels>
-                                                      sample_to_bin_op,
-                                                  fixed_array<unsigned int, ActiveChannels> bins)
+ROCPRIM_KERNEL ROCPRIM_LAUNCH_BOUNDS(device_params<Config>().histogram_config.block_size) void
+    histogram_shared_kernel(SampleIterator                                   samples,
+                            unsigned int                                     columns,
+                            unsigned int                                     rows,
+                            unsigned int                                     row_stride,
+                            unsigned int                                     rows_per_block,
+                            unsigned int                                     shared_histograms,
+                            fixed_array<Counter*, ActiveChannels>            histogram,
+                            const fixed_array<SampleToBinOp, ActiveChannels> sample_to_bin_op,
+                            const fixed_array<unsigned int, ActiveChannels>  bins)
 {
     static constexpr histogram_config_params params = device_params<Config>();
 
+// Temporary fix: issue with dynamic shared memory on windows.
+#ifndef _WIN32
     HIP_DYNAMIC_SHARED(unsigned int, block_histogram);
+#else
+    __shared__ unsigned int block_histogram[params.shared_impl_max_bins];
+#endif
 
     histogram_shared<params.histogram_config.block_size,
                      params.histogram_config.items_per_thread,
@@ -98,17 +99,13 @@ template<class Config,
          class SampleIterator,
          class Counter,
          class SampleToBinOp>
-ROCPRIM_KERNEL __launch_bounds__(
-    device_params<Config>()
-        .histogram_config
-        .block_size) void histogram_global_kernel(SampleIterator                        samples,
-                                                  unsigned int                          columns,
-                                                  unsigned int                          row_stride,
-                                                  fixed_array<Counter*, ActiveChannels> histogram,
-                                                  fixed_array<SampleToBinOp, ActiveChannels>
-                                                      sample_to_bin_op,
-                                                  fixed_array<unsigned int, ActiveChannels>
-                                                      bins_bits)
+ROCPRIM_KERNEL ROCPRIM_LAUNCH_BOUNDS(device_params<Config>().histogram_config.block_size) void
+    histogram_global_kernel(SampleIterator                                   samples,
+                            unsigned int                                     columns,
+                            unsigned int                                     row_stride,
+                            fixed_array<Counter*, ActiveChannels>            histogram,
+                            const fixed_array<SampleToBinOp, ActiveChannels> sample_to_bin_op,
+                            const fixed_array<unsigned int, ActiveChannels>  bins_bits)
 {
     static constexpr histogram_config_params params = device_params<Config>();
 
@@ -123,23 +120,6 @@ ROCPRIM_KERNEL __launch_bounds__(
                                      bins_bits);
 }
 
-#define ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR(name, size, start)                           \
-    {                                                                                            \
-        auto _error = hipGetLastError();                                                         \
-        if(_error != hipSuccess)                                                                 \
-            return _error;                                                                       \
-        if(debug_synchronous)                                                                    \
-        {                                                                                        \
-            std::cout << name << "(" << size << ")";                                             \
-            auto __error = hipStreamSynchronize(stream);                                         \
-            if(__error != hipSuccess)                                                            \
-                return __error;                                                                  \
-            auto _end = std::chrono::high_resolution_clock::now();                               \
-            auto _d   = std::chrono::duration_cast<std::chrono::duration<double>>(_end - start); \
-            std::cout << " " << _d.count() * 1000 << " ms" << '\n';                              \
-        }                                                                                        \
-    }
-
 template<unsigned int Channels,
          unsigned int ActiveChannels,
          class Config,
@@ -206,22 +186,24 @@ inline hipError_t histogram_impl(void*          temporary_storage,
 
     unsigned int bins[ActiveChannels];
     unsigned int bins_bits[ActiveChannels];
-    unsigned int total_bins = 0;
-    unsigned int max_bins   = 0;
+    unsigned int total_shared_bins = 0;
+    unsigned int max_bins          = 0;
     for(unsigned int channel = 0; channel < ActiveChannels; channel++)
     {
         bins[channel] = levels[channel] - 1;
         bins_bits[channel]
             = static_cast<unsigned int>(std::log2(detail::next_power_of_two(bins[channel])));
-        total_bins += bins[channel];
+        const unsigned int size = bins[channel];
+        // Prevent LDS bank conflicts
+        total_shared_bins += rocprim::detail::is_power_of_two(size) ? size + 1 : size;
         max_bins = std::max(max_bins, bins[channel]);
     }
 
-    std::chrono::high_resolution_clock::time_point start;
+    std::chrono::steady_clock::time_point start;
 
     if(debug_synchronous)
     {
-        start = std::chrono::high_resolution_clock::now();
+        start = std::chrono::steady_clock::now();
     }
     hipLaunchKernelGGL(HIP_KERNEL_NAME(init_histogram_kernel<config, ActiveChannels>),
                        dim3(::rocprim::detail::ceiling_div(max_bins, block_size)),
@@ -237,11 +219,11 @@ inline hipError_t histogram_impl(void*          temporary_storage,
         return hipSuccess;
     }
 
-    if(total_bins <= shared_impl_max_bins)
+    if(total_shared_bins <= shared_impl_max_bins)
     {
         if(debug_synchronous)
         {
-            start = std::chrono::high_resolution_clock::now();
+            start = std::chrono::steady_clock::now();
         }
         auto kernel = HIP_KERNEL_NAME(histogram_shared_kernel<config,
                                                               Channels,
@@ -250,7 +232,7 @@ inline hipError_t histogram_impl(void*          temporary_storage,
                                                               Counter,
                                                               SampleToBinOp>);
 
-        const size_t block_histogram_bytes = total_bins * sizeof(unsigned int);
+        const size_t block_histogram_bytes = total_shared_bins * sizeof(unsigned int);
 
         // Use up to shared_impl_histograms histograms in shared memory to reduce atomic conflicts
         // for the case of samples concentrated in one bin
@@ -318,7 +300,7 @@ inline hipError_t histogram_impl(void*          temporary_storage,
     {
         if(debug_synchronous)
         {
-            start = std::chrono::high_resolution_clock::now();
+            start = std::chrono::steady_clock::now();
         }
         hipLaunchKernelGGL(
             HIP_KERNEL_NAME(histogram_global_kernel<config, Channels, ActiveChannels>),
@@ -436,7 +418,7 @@ inline hipError_t histogram_range_impl(void*          temporary_storage,
                                                             debug_synchronous);
 }
 
-#undef ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR
+
 
 } // namespace detail
 
@@ -449,24 +431,24 @@ inline hipError_t histogram_range_impl(void*          temporary_storage,
 /// * Returns the required size of \p temporary_storage in \p storage_size
 /// if \p temporary_storage in a null pointer.
 ///
-/// \tparam Config - [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
-/// \tparam SampleIterator - random-access iterator type of the input range. Must meet the
+/// \tparam Config [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
+/// \tparam SampleIterator random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam Counter - integer type for histogram bin counters.
-/// \tparam Level - type of histogram boundaries (levels)
+/// \tparam Counter integer type for histogram bin counters.
+/// \tparam Level type of histogram boundaries (levels)
 ///
-/// \param [in] temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param [in] temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// \p storage_size and function returns without performing the reduction operation.
-/// \param [in,out] storage_size - reference to a size (in bytes) of \p temporary_storage.
-/// \param [in] samples - iterator to the first element in the range of input samples.
-/// \param [in] size - number of elements in the samples range.
-/// \param [out] histogram - pointer to the first element in the histogram range.
-/// \param [in] levels - number of boundaries (levels) for histogram bins.
-/// \param [in] lower_level - lower sample value bound (inclusive) for the first histogram bin.
-/// \param [in] upper_level - upper sample value bound (exclusive) for the last histogram bin.
-/// \param [in] stream - [optional] HIP stream object. Default is \p 0 (default stream).
-/// \param [in] debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param [in,out] storage_size reference to a size (in bytes) of \p temporary_storage.
+/// \param [in] samples iterator to the first element in the range of input samples.
+/// \param [in] size number of elements in the samples range.
+/// \param [out] histogram pointer to the first element in the histogram range.
+/// \param [in] levels number of boundaries (levels) for histogram bins.
+/// \param [in] lower_level lower sample value bound (inclusive) for the first histogram bin.
+/// \param [in] upper_level upper sample value bound (exclusive) for the last histogram bin.
+/// \param [in] stream [optional] HIP stream object. Default is \p 0 (default stream).
+/// \param [in] debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors. Default value is \p false.
 ///
 /// \returns \p hipSuccess (\p 0) after successful histogram operation; otherwise a HIP runtime error of
@@ -552,26 +534,26 @@ inline hipError_t histogram_even(void*          temporary_storage,
 /// * Returns the required size of \p temporary_storage in \p storage_size
 /// if \p temporary_storage in a null pointer.
 ///
-/// \tparam Config - [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
-/// \tparam SampleIterator - random-access iterator type of the input range. Must meet the
+/// \tparam Config [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
+/// \tparam SampleIterator random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam Counter - integer type for histogram bin counters.
-/// \tparam Level - type of histogram boundaries (levels)
+/// \tparam Counter integer type for histogram bin counters.
+/// \tparam Level type of histogram boundaries (levels)
 ///
-/// \param [in] temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param [in] temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// \p storage_size and function returns without performing the reduction operation.
-/// \param [in,out] storage_size - reference to a size (in bytes) of \p temporary_storage.
-/// \param [in] samples - iterator to the first element in the range of input samples.
-/// \param [in] columns - number of elements in each row of the region.
-/// \param [in] rows - number of rows of the region.
-/// \param [in] row_stride_bytes - number of bytes between starts of consecutive rows of the region.
-/// \param [out] histogram - pointer to the first element in the histogram range.
-/// \param [in] levels - number of boundaries (levels) for histogram bins.
-/// \param [in] lower_level - lower sample value bound (inclusive) for the first histogram bin.
-/// \param [in] upper_level - upper sample value bound (exclusive) for the last histogram bin.
-/// \param [in] stream - [optional] HIP stream object. Default is \p 0 (default stream).
-/// \param [in] debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param [in,out] storage_size reference to a size (in bytes) of \p temporary_storage.
+/// \param [in] samples iterator to the first element in the range of input samples.
+/// \param [in] columns number of elements in each row of the region.
+/// \param [in] rows number of rows of the region.
+/// \param [in] row_stride_bytes number of bytes between starts of consecutive rows of the region.
+/// \param [out] histogram pointer to the first element in the histogram range.
+/// \param [in] levels number of boundaries (levels) for histogram bins.
+/// \param [in] lower_level lower sample value bound (inclusive) for the first histogram bin.
+/// \param [in] upper_level upper sample value bound (exclusive) for the last histogram bin.
+/// \param [in] stream [optional] HIP stream object. Default is \p 0 (default stream).
+/// \param [in] debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors. Default value is \p false.
 ///
 /// \returns \p hipSuccess (\p 0) after successful histogram operation; otherwise a HIP runtime error of
@@ -661,31 +643,38 @@ inline hipError_t histogram_even(void*          temporary_storage,
 /// * Returns the required size of \p temporary_storage in \p storage_size
 /// if \p temporary_storage in a null pointer.
 ///
-/// \tparam Channels - number of channels interleaved in the input samples.
-/// \tparam ActiveChannels - number of channels being used for computing histograms.
-/// \tparam Config - [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
-/// \tparam SampleIterator - random-access iterator type of the input range. Must meet the
+/// \tparam Channels number of channels interleaved in the input samples.
+/// \tparam ActiveChannels number of channels being used for computing histograms.
+/// \tparam Config [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
+/// \tparam SampleIterator random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam Counter - integer type for histogram bin counters.
-/// \tparam Level - type of histogram boundaries (levels)
+/// \tparam Counter integer type for histogram bin counters.
+/// \tparam Level type of histogram boundaries (levels)
 ///
-/// \param [in] temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param [in] temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// \p storage_size and function returns without performing the reduction operation.
-/// \param [in,out] storage_size - reference to a size (in bytes) of \p temporary_storage.
-/// \param [in] samples - iterator to the first element in the range of input samples.
-/// \param [in] size - number of pixels in the samples range.
-/// \param [out] histogram - pointers to the first element in the histogram range, one for each active channel.
-/// \param [in] levels - number of boundaries (levels) for histogram bins in each active channel.
-/// \param [in] lower_level - lower sample value bound (inclusive) for the first histogram bin in each active channel.
-/// \param [in] upper_level - upper sample value bound (exclusive) for the last histogram bin in each active channel.
-/// \param [in] stream - [optional] HIP stream object. Default is \p 0 (default stream).
-/// \param [in] debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param [in,out] storage_size reference to a size (in bytes) of \p temporary_storage.
+/// \param [in] samples iterator to the first element in the range of input samples.
+/// \param [in] size number of pixels in the samples range.
+/// \param [out] histogram pointers to the first element in the histogram range, one for each active channel.
+/// \param [in] levels number of boundaries (levels) for histogram bins in each active channel.
+/// \param [in] lower_level lower sample value bound (inclusive) for the first histogram bin in each active channel.
+/// \param [in] upper_level upper sample value bound (exclusive) for the last histogram bin in each active channel.
+/// \param [in] stream [optional] HIP stream object. Default is \p 0 (default stream).
+/// \param [in] debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors. Default value is \p false.
 ///
 /// \returns \p hipSuccess (\p 0) after successful histogram operation; otherwise a HIP runtime error of
 /// type \p hipError_t.
 ///
+/// \par Notes
+/// * Currently the \p Channels template parameter has no strict restriction on its value. However,
+///   internally a vector type of elements of type \p SampleIterator and length \p Channels is used
+///   to represent the input items, so the amount of local memory available will limit the range of
+///   possible values for this template parameter.
+/// * \p ActiveChannels must be less or equal than \p Channels.
+///
 /// \par Example
 /// \parblock
 /// In this example histograms for 3 channels (RGB) are computed on an array of 8-bit RGBA samples.
@@ -773,33 +762,40 @@ inline hipError_t multi_histogram_even(void*          temporary_storage,
 /// * Returns the required size of \p temporary_storage in \p storage_size
 /// if \p temporary_storage in a null pointer.
 ///
-/// \tparam Channels - number of channels interleaved in the input samples.
-/// \tparam ActiveChannels - number of channels being used for computing histograms.
-/// \tparam Config - [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
-/// \tparam SampleIterator - random-access iterator type of the input range. Must meet the
+/// \tparam Channels number of channels interleaved in the input samples.
+/// \tparam ActiveChannels number of channels being used for computing histograms.
+/// \tparam Config [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
+/// \tparam SampleIterator random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam Counter - integer type for histogram bin counters.
-/// \tparam Level - type of histogram boundaries (levels)
+/// \tparam Counter integer type for histogram bin counters.
+/// \tparam Level type of histogram boundaries (levels)
 ///
-/// \param [in] temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param [in] temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// \p storage_size and function returns without performing the reduction operation.
-/// \param [in,out] storage_size - reference to a size (in bytes) of \p temporary_storage.
-/// \param [in] samples - iterator to the first element in the range of input samples.
-/// \param [in] columns - number of elements in each row of the region.
-/// \param [in] rows - number of rows of the region.
-/// \param [in] row_stride_bytes - number of bytes between starts of consecutive rows of the region.
-/// \param [out] histogram - pointers to the first element in the histogram range, one for each active channel.
-/// \param [in] levels - number of boundaries (levels) for histogram bins in each active channel.
-/// \param [in] lower_level - lower sample value bound (inclusive) for the first histogram bin in each active channel.
-/// \param [in] upper_level - upper sample value bound (exclusive) for the last histogram bin in each active channel.
-/// \param [in] stream - [optional] HIP stream object. Default is \p 0 (default stream).
-/// \param [in] debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param [in,out] storage_size reference to a size (in bytes) of \p temporary_storage.
+/// \param [in] samples iterator to the first element in the range of input samples.
+/// \param [in] columns number of elements in each row of the region.
+/// \param [in] rows number of rows of the region.
+/// \param [in] row_stride_bytes number of bytes between starts of consecutive rows of the region.
+/// \param [out] histogram pointers to the first element in the histogram range, one for each active channel.
+/// \param [in] levels number of boundaries (levels) for histogram bins in each active channel.
+/// \param [in] lower_level lower sample value bound (inclusive) for the first histogram bin in each active channel.
+/// \param [in] upper_level upper sample value bound (exclusive) for the last histogram bin in each active channel.
+/// \param [in] stream [optional] HIP stream object. Default is \p 0 (default stream).
+/// \param [in] debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors. Default value is \p false.
 ///
 /// \returns \p hipSuccess (\p 0) after successful histogram operation; otherwise a HIP runtime error of
 /// type \p hipError_t.
 ///
+/// \par Notes
+/// * Currently the \p Channels template parameter has no strict restriction on its value. However,
+///   internally a vector type of elements of type \p SampleIterator and length \p Channels is used
+///   to represent the input items, so the amount of local memory available will limit the range of
+///   possible values for this template parameter.
+/// * \p ActiveChannels must be less or equal than \p Channels.
+///
 /// \par Example
 /// \parblock
 /// In this example histograms for 3 channels (RGB) are computed on an array of 8-bit RGBA samples.
@@ -882,23 +878,23 @@ inline hipError_t multi_histogram_even(void*          temporary_storage,
 /// * Returns the required size of \p temporary_storage in \p storage_size
 /// if \p temporary_storage in a null pointer.
 ///
-/// \tparam Config - [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
-/// \tparam SampleIterator - random-access iterator type of the input range. Must meet the
+/// \tparam Config [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
+/// \tparam SampleIterator random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam Counter - integer type for histogram bin counters.
-/// \tparam Level - type of histogram boundaries (levels)
+/// \tparam Counter integer type for histogram bin counters.
+/// \tparam Level type of histogram boundaries (levels)
 ///
-/// \param [in] temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param [in] temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// \p storage_size and function returns without performing the reduction operation.
-/// \param [in,out] storage_size - reference to a size (in bytes) of \p temporary_storage.
-/// \param [in] samples - iterator to the first element in the range of input samples.
-/// \param [in] size - number of elements in the samples range.
-/// \param [out] histogram - pointer to the first element in the histogram range.
-/// \param [in] levels - number of boundaries (levels) for histogram bins.
-/// \param [in] level_values - pointer to the array of bin boundaries.
-/// \param [in] stream - [optional] HIP stream object. Default is \p 0 (default stream).
-/// \param [in] debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param [in,out] storage_size reference to a size (in bytes) of \p temporary_storage.
+/// \param [in] samples iterator to the first element in the range of input samples.
+/// \param [in] size number of elements in the samples range.
+/// \param [out] histogram pointer to the first element in the histogram range.
+/// \param [in] levels number of boundaries (levels) for histogram bins.
+/// \param [in] level_values pointer to the array of bin boundaries.
+/// \param [in] stream [optional] HIP stream object. Default is \p 0 (default stream).
+/// \param [in] debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors. Default value is \p false.
 ///
 /// \returns \p hipSuccess (\p 0) after successful histogram operation; otherwise a HIP runtime error of
@@ -979,25 +975,25 @@ inline hipError_t histogram_range(void*          temporary_storage,
 /// * Returns the required size of \p temporary_storage in \p storage_size
 /// if \p temporary_storage in a null pointer.
 ///
-/// \tparam Config - [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
-/// \tparam SampleIterator - random-access iterator type of the input range. Must meet the
+/// \tparam Config [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
+/// \tparam SampleIterator random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam Counter - integer type for histogram bin counters.
-/// \tparam Level - type of histogram boundaries (levels)
+/// \tparam Counter integer type for histogram bin counters.
+/// \tparam Level type of histogram boundaries (levels)
 ///
-/// \param [in] temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param [in] temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// \p storage_size and function returns without performing the reduction operation.
-/// \param [in,out] storage_size - reference to a size (in bytes) of \p temporary_storage.
-/// \param [in] samples - iterator to the first element in the range of input samples.
-/// \param [in] columns - number of elements in each row of the region.
-/// \param [in] rows - number of rows of the region.
-/// \param [in] row_stride_bytes - number of bytes between starts of consecutive rows of the region.
-/// \param [out] histogram - pointer to the first element in the histogram range.
-/// \param [in] levels - number of boundaries (levels) for histogram bins.
-/// \param [in] level_values - pointer to the array of bin boundaries.
-/// \param [in] stream - [optional] HIP stream object. Default is \p 0 (default stream).
-/// \param [in] debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param [in,out] storage_size reference to a size (in bytes) of \p temporary_storage.
+/// \param [in] samples iterator to the first element in the range of input samples.
+/// \param [in] columns number of elements in each row of the region.
+/// \param [in] rows number of rows of the region.
+/// \param [in] row_stride_bytes number of bytes between starts of consecutive rows of the region.
+/// \param [out] histogram pointer to the first element in the histogram range.
+/// \param [in] levels number of boundaries (levels) for histogram bins.
+/// \param [in] level_values pointer to the array of bin boundaries.
+/// \param [in] stream [optional] HIP stream object. Default is \p 0 (default stream).
+/// \param [in] debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors. Default value is \p false.
 ///
 /// \returns \p hipSuccess (\p 0) after successful histogram operation; otherwise a HIP runtime error of
@@ -1083,30 +1079,37 @@ inline hipError_t histogram_range(void*          temporary_storage,
 /// * Returns the required size of \p temporary_storage in \p storage_size
 /// if \p temporary_storage in a null pointer.
 ///
-/// \tparam Channels - number of channels interleaved in the input samples.
-/// \tparam ActiveChannels - number of channels being used for computing histograms.
-/// \tparam Config - [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
-/// \tparam SampleIterator - random-access iterator type of the input range. Must meet the
+/// \tparam Channels number of channels interleaved in the input samples.
+/// \tparam ActiveChannels number of channels being used for computing histograms.
+/// \tparam Config [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
+/// \tparam SampleIterator random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam Counter - integer type for histogram bin counters.
-/// \tparam Level - type of histogram boundaries (levels)
+/// \tparam Counter integer type for histogram bin counters.
+/// \tparam Level type of histogram boundaries (levels)
 ///
-/// \param [in] temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param [in] temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// \p storage_size and function returns without performing the reduction operation.
-/// \param [in,out] storage_size - reference to a size (in bytes) of \p temporary_storage.
-/// \param [in] samples - iterator to the first element in the range of input samples.
-/// \param [in] size - number of pixels in the samples range.
-/// \param [out] histogram - pointers to the first element in the histogram range, one for each active channel.
-/// \param [in] levels - number of boundaries (levels) for histogram bins in each active channel.
-/// \param [in] level_values - pointer to the array of bin boundaries for each active channel.
-/// \param [in] stream - [optional] HIP stream object. Default is \p 0 (default stream).
-/// \param [in] debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param [in,out] storage_size reference to a size (in bytes) of \p temporary_storage.
+/// \param [in] samples iterator to the first element in the range of input samples.
+/// \param [in] size number of pixels in the samples range.
+/// \param [out] histogram pointers to the first element in the histogram range, one for each active channel.
+/// \param [in] levels number of boundaries (levels) for histogram bins in each active channel.
+/// \param [in] level_values pointer to the array of bin boundaries for each active channel.
+/// \param [in] stream [optional] HIP stream object. Default is \p 0 (default stream).
+/// \param [in] debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors. Default value is \p false.
 ///
 /// \returns \p hipSuccess (\p 0) after successful histogram operation; otherwise a HIP runtime error of
 /// type \p hipError_t.
 ///
+/// \par Notes
+/// * Currently the \p Channels template parameter has no strict restriction on its value. However,
+///   internally a vector type of elements of type \p SampleIterator and length \p Channels is used
+///   to represent the input items, so the amount of local memory available will limit the range of
+///   possible values for this template parameter.
+/// * \p ActiveChannels must be less or equal than \p Channels.
+///
 /// \par Example
 /// \parblock
 /// In this example histograms for 3 channels (RGB) are computed on an array of 8-bit RGBA samples.
@@ -1190,32 +1193,39 @@ inline hipError_t multi_histogram_range(void*          temporary_storage,
 /// * Returns the required size of \p temporary_storage in \p storage_size
 /// if \p temporary_storage in a null pointer.
 ///
-/// \tparam Channels - number of channels interleaved in the input samples.
-/// \tparam ActiveChannels - number of channels being used for computing histograms.
-/// \tparam Config - [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
-/// \tparam SampleIterator - random-access iterator type of the input range. Must meet the
+/// \tparam Channels number of channels interleaved in the input samples.
+/// \tparam ActiveChannels number of channels being used for computing histograms.
+/// \tparam Config [optional] Configuration of the primitive, must be `default_config` or `histogram_config`.
+/// \tparam SampleIterator random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam Counter - integer type for histogram bin counters.
-/// \tparam Level - type of histogram boundaries (levels)
+/// \tparam Counter integer type for histogram bin counters.
+/// \tparam Level type of histogram boundaries (levels)
 ///
-/// \param [in] temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param [in] temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// \p storage_size and function returns without performing the reduction operation.
-/// \param [in,out] storage_size - reference to a size (in bytes) of \p temporary_storage.
-/// \param [in] samples - iterator to the first element in the range of input samples.
-/// \param [in] columns - number of elements in each row of the region.
-/// \param [in] rows - number of rows of the region.
-/// \param [in] row_stride_bytes - number of bytes between starts of consecutive rows of the region.
-/// \param [out] histogram - pointers to the first element in the histogram range, one for each active channel.
-/// \param [in] levels - number of boundaries (levels) for histogram bins in each active channel.
-/// \param [in] level_values - pointer to the array of bin boundaries for each active channel.
-/// \param [in] stream - [optional] HIP stream object. Default is \p 0 (default stream).
-/// \param [in] debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param [in,out] storage_size reference to a size (in bytes) of \p temporary_storage.
+/// \param [in] samples iterator to the first element in the range of input samples.
+/// \param [in] columns number of elements in each row of the region.
+/// \param [in] rows number of rows of the region.
+/// \param [in] row_stride_bytes number of bytes between starts of consecutive rows of the region.
+/// \param [out] histogram pointers to the first element in the histogram range, one for each active channel.
+/// \param [in] levels number of boundaries (levels) for histogram bins in each active channel.
+/// \param [in] level_values pointer to the array of bin boundaries for each active channel.
+/// \param [in] stream [optional] HIP stream object. Default is \p 0 (default stream).
+/// \param [in] debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors. Default value is \p false.
 ///
 /// \returns \p hipSuccess (\p 0) after successful histogram operation; otherwise a HIP runtime error of
 /// type \p hipError_t.
 ///
+/// \par Notes
+/// * Currently the \p Channels template parameter has no strict restriction on its value. However,
+///   internally a vector type of elements of type \p SampleIterator and length \p Channels is used
+///   to represent the input items, so the amount of local memory available will limit the range of
+///   possible values for this template parameter.
+/// * \p ActiveChannels must be less or equal than \p Channels.
+///
 /// \par Example
 /// \parblock
 /// In this example histograms for 3 channels (RGB) are computed on an array of 8-bit RGBA samples.
