@@ -1,4 +1,4 @@
-// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -25,11 +25,12 @@
 #include <iterator>
 #include <type_traits>
 
+#include "../common.hpp"
 #include "../config.hpp"
 #include "../detail/temp_storage.hpp"
 #include "../detail/various.hpp"
+#include "../detail/virtual_shared_memory.hpp"
 
-#include "device_merge_config.hpp"
 #include "detail/device_merge.hpp"
 
 BEGIN_ROCPRIM_NAMESPACE
@@ -40,21 +41,19 @@ BEGIN_ROCPRIM_NAMESPACE
 namespace detail
 {
 
-template<
-    class IndexIterator,
-    class KeysInputIterator1,
-    class KeysInputIterator2,
-    class BinaryFunction
->
-ROCPRIM_KERNEL
-__launch_bounds__(ROCPRIM_DEFAULT_MAX_BLOCK_SIZE)
-void partition_kernel(IndexIterator index,
-                      KeysInputIterator1 keys_input1,
-                      KeysInputIterator2 keys_input2,
-                      const size_t input1_size,
-                      const size_t input2_size,
-                      const unsigned int spacing,
-                      BinaryFunction compare_function)
+template<class Config,
+         class IndexIterator,
+         class KeysInputIterator1,
+         class KeysInputIterator2,
+         class BinaryFunction>
+ROCPRIM_KERNEL ROCPRIM_LAUNCH_BOUNDS(device_params<Config>().kernel_config.block_size) void
+    partition_kernel(IndexIterator      index,
+                     KeysInputIterator1 keys_input1,
+                     KeysInputIterator2 keys_input2,
+                     const size_t       input1_size,
+                     const size_t       input2_size,
+                     const unsigned int spacing,
+                     BinaryFunction     compare_function)
 {
     partition_kernel_impl(
         index, keys_input1, keys_input2, input1_size, input2_size,
@@ -62,52 +61,60 @@ void partition_kernel(IndexIterator index,
     );
 }
 
-template<
-    unsigned int BlockSize,
-    unsigned int ItemsPerThread,
-    class IndexIterator,
-    class KeysInputIterator1,
-    class KeysInputIterator2,
-    class KeysOutputIterator,
-    class ValuesInputIterator1,
-    class ValuesInputIterator2,
-    class ValuesOutputIterator,
-    class BinaryFunction
->
-ROCPRIM_KERNEL
-__launch_bounds__(BlockSize)
-void merge_kernel(IndexIterator index,
-                  KeysInputIterator1 keys_input1,
-                  KeysInputIterator2 keys_input2,
-                  KeysOutputIterator keys_output,
-                  ValuesInputIterator1 values_input1,
-                  ValuesInputIterator2 values_input2,
-                  ValuesOutputIterator values_output,
-                  const size_t input1_size,
-                  const size_t input2_size,
-                  BinaryFunction compare_function)
+template<class Config,
+         class IndexIterator,
+         class KeysInputIterator1,
+         class KeysInputIterator2,
+         class KeysOutputIterator,
+         class ValuesInputIterator1,
+         class ValuesInputIterator2,
+         class ValuesOutputIterator,
+         class BinaryFunction>
+ROCPRIM_KERNEL ROCPRIM_LAUNCH_BOUNDS(device_params<Config>().kernel_config.block_size) void
+    merge_kernel(IndexIterator        index,
+                 KeysInputIterator1   keys_input1,
+                 KeysInputIterator2   keys_input2,
+                 KeysOutputIterator   keys_output,
+                 ValuesInputIterator1 values_input1,
+                 ValuesInputIterator2 values_input2,
+                 ValuesOutputIterator values_output,
+                 const size_t         input1_size,
+                 const size_t         input2_size,
+                 BinaryFunction       compare_function,
+                 detail::vsmem_t      vsmem)
 {
-    merge_kernel_impl<BlockSize, ItemsPerThread>(
-        index, keys_input1, keys_input2, keys_output,
-        values_input1, values_input2, values_output,
-        input1_size, input2_size, compare_function
-    );
+    using key_type   = typename std::iterator_traits<KeysInputIterator1>::value_type;
+    using value_type = typename std::iterator_traits<ValuesInputIterator1>::value_type;
+
+    using merge_kernel_impl_t = merge_kernel_impl_<Config, key_type, value_type>;
+
+    using VSmemHelperT = detail::vsmem_helper_impl<merge_kernel_impl_t>;
+    ROCPRIM_SHARED_MEMORY typename VSmemHelperT::static_temp_storage_t static_temp_storage;
+    // Get temporary storage
+    typename merge_kernel_impl_t::storage_type& storage
+        = VSmemHelperT::get_temp_storage(static_temp_storage, vsmem);
+
+    merge_kernel_impl_t().merge(index,
+                                keys_input1,
+                                keys_input2,
+                                keys_output,
+                                values_input1,
+                                values_input2,
+                                values_output,
+                                input1_size,
+                                input2_size,
+                                compare_function,
+                                storage);
 }
 
-#define ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR(name, size, start) \
-    { \
-        auto _error = hipGetLastError(); \
-        if(_error != hipSuccess) return _error; \
-        if(debug_synchronous) \
-        { \
-            std::cout << name << "(" << size << ")"; \
-            auto __error = hipStreamSynchronize(stream); \
-            if(__error != hipSuccess) return __error; \
-            auto _end = std::chrono::high_resolution_clock::now(); \
-            auto _d = std::chrono::duration_cast<std::chrono::duration<double>>(_end - start); \
-            std::cout << " " << _d.count() * 1000 << " ms" << '\n'; \
-        } \
-    }
+template<class Config, class Key, class Value>
+inline size_t get_merge_vsmem_size_per_block()
+{
+    using merge_kernel_impl_t = merge_kernel_impl_<Config, Key, Value>;
+    using MergeVSmemHelperT   = detail::vsmem_helper_impl<merge_kernel_impl_t>;
+
+    return MergeVSmemHelperT::vsmem_per_block;
+}
 
 template<
     class Config,
@@ -138,38 +145,53 @@ hipError_t merge_impl(void * temporary_storage,
     using key_type = typename std::iterator_traits<KeysInputIterator1>::value_type;
     using value_type = typename std::iterator_traits<ValuesInputIterator1>::value_type;
 
-    // Get default config if Config is default_config
-    using config = detail::default_or_custom_config<
-        Config,
-        detail::default_merge_config<ROCPRIM_TARGET_ARCH, key_type, value_type>
-    >;
+    using config = wrapped_merge_config<Config, key_type, value_type>;
 
-    static constexpr unsigned int block_size = config::block_size;
-    static constexpr unsigned int half_block = block_size / 2;
-    static constexpr unsigned int items_per_thread = config::items_per_thread;
-    static constexpr auto items_per_block = block_size * items_per_thread;
+    detail::target_arch target_arch;
+    hipError_t          result = detail::host_target_arch(stream, target_arch);
+    if(result != hipSuccess)
+    {
+        return result;
+    }
+    const merge_config_params params = detail::dispatch_target_arch<config>(target_arch);
 
-    const unsigned int partitions
+    const unsigned int block_size       = params.kernel_config.block_size;
+    const unsigned int half_block       = block_size / 2;
+    const unsigned int items_per_thread = params.kernel_config.items_per_thread;
+    const unsigned int items_per_block  = block_size * items_per_thread;
+
+    const unsigned int number_of_blocks
         = ((input1_size + input2_size) + items_per_block - 1) / items_per_block;
 
-    unsigned int* index;
+    size_t virtual_shared_memory_size
+        = get_merge_vsmem_size_per_block<config, key_type, value_type>() * number_of_blocks;
+
+    unsigned int* index = nullptr;
+    void*         vsmem = nullptr;
 
     const hipError_t partition_result = detail::temp_storage::partition(
         temporary_storage,
         storage_size,
-        detail::temp_storage::ptr_aligned_array(&index, partitions + 1));
+        detail::temp_storage::make_linear_partition(
+            detail::temp_storage::ptr_aligned_array(&index, number_of_blocks + 1),
+            // vsmem
+            detail::temp_storage::make_partition(&vsmem,
+                                                 virtual_shared_memory_size,
+                                                 cache_line_size)));
+
     if(partition_result != hipSuccess || temporary_storage == nullptr)
     {
         return partition_result;
     }
 
-    if( partitions == 0u )
+    if(number_of_blocks == 0u)
+    {
         return hipSuccess;
+    }
 
     // Start point for time measurements
-    std::chrono::high_resolution_clock::time_point start;
+    std::chrono::steady_clock::time_point start;
 
-    auto number_of_blocks = partitions;
     if(debug_synchronous)
     {
         std::cout << "block_size " << block_size << '\n';
@@ -177,31 +199,38 @@ hipError_t merge_impl(void * temporary_storage,
         std::cout << "items_per_block " << items_per_block << '\n';
     }
 
-    const unsigned partition_blocks = ((partitions + 1) + half_block - 1) / half_block;
+    const unsigned int partition_blocks = ((number_of_blocks + 1) + half_block - 1) / half_block;
 
-    if(debug_synchronous) start = std::chrono::high_resolution_clock::now();
-    hipLaunchKernelGGL(
-        HIP_KERNEL_NAME(detail::partition_kernel),
-        dim3(partition_blocks), dim3(half_block), 0, stream,
-        index, keys_input1, keys_input2, input1_size, input2_size,
-        items_per_block, compare_function
-    );
+    if(debug_synchronous) start = std::chrono::steady_clock::now();
+    detail::partition_kernel<config>
+        <<<dim3(partition_blocks), dim3(half_block), 0, stream>>>(index,
+                                                                  keys_input1,
+                                                                  keys_input2,
+                                                                  input1_size,
+                                                                  input2_size,
+                                                                  items_per_block,
+                                                                  compare_function);
     ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR("partition_kernel", input1_size, start);
 
-    if(debug_synchronous) start = std::chrono::high_resolution_clock::now();
-    hipLaunchKernelGGL(
-        HIP_KERNEL_NAME(detail::merge_kernel<block_size, items_per_thread>),
-        dim3(number_of_blocks), dim3(block_size), 0, stream,
-        index, keys_input1, keys_input2, keys_output,
-        values_input1, values_input2, values_output,
-        input1_size, input2_size, compare_function
-    );
+    if(debug_synchronous) start = std::chrono::steady_clock::now();
+    detail::merge_kernel<config>
+        <<<dim3(number_of_blocks), dim3(block_size), 0, stream>>>(index,
+                                                                  keys_input1,
+                                                                  keys_input2,
+                                                                  keys_output,
+                                                                  values_input1,
+                                                                  values_input2,
+                                                                  values_output,
+                                                                  input1_size,
+                                                                  input2_size,
+                                                                  compare_function,
+                                                                  detail::vsmem_t{vsmem});
     ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR("merge_kernel", input1_size, start);
 
     return hipSuccess;
 }
 
-#undef ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR
+
 
 } // end of detail namespace
 
@@ -216,30 +245,30 @@ hipError_t merge_impl(void * temporary_storage,
 /// if \p temporary_storage in a null pointer.
 /// * Accepts custom compare_functions for merging across the device.
 ///
-/// \tparam Config - [optional] Configuration of the primitive, must be `default_config` or `merge_config`.
-/// \tparam InputIterator1 - random-access iterator type of the first input range. Must meet the
+/// \tparam Config [optional] Configuration of the primitive, must be `default_config` or `merge_config`.
+/// \tparam InputIterator1 random-access iterator type of the first input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam InputIterator2 - random-access iterator type of the second input range. Must meet the
+/// \tparam InputIterator2 random-access iterator type of the second input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam OutputIterator - random-access iterator type of the output range. Must meet the
+/// \tparam OutputIterator random-access iterator type of the output range. Must meet the
 /// requirements of a C++ OutputIterator concept. It can be a simple pointer type.
 ///
-/// \param [in] temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param [in] temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// \p storage_size and function returns without performing the sort operation.
-/// \param [in,out] storage_size - reference to a size (in bytes) of \p temporary_storage.
-/// \param [in] input1 - iterator to the first element in the first range to merge.
-/// \param [in] input2 - iterator to the first element in the second range to merge.
-/// \param [out] output - iterator to the first element in the output range.
-/// \param [in] input1_size - number of element in the first input range.
-/// \param [in] input2_size - number of element in the second input range.
-/// \param [in] compare_function - binary operation function object that will be used for comparison.
+/// \param [in,out] storage_size reference to a size (in bytes) of \p temporary_storage.
+/// \param [in] input1 iterator to the first element in the first range to merge.
+/// \param [in] input2 iterator to the first element in the second range to merge.
+/// \param [out] output iterator to the first element in the output range.
+/// \param [in] input1_size number of element in the first input range.
+/// \param [in] input2_size number of element in the second input range.
+/// \param [in] compare_function binary operation function object that will be used for comparison.
 /// The signature of the function should be equivalent to the following:
 /// <tt>bool f(const T &a, const T &b);</tt>. The signature does not need to have
 /// <tt>const &</tt>, but function object must not modify the objects passed to it.
 /// The default value is \p BinaryFunction().
-/// \param [in] stream - [optional] HIP stream object. Default is \p 0 (default stream).
-/// \param [in] debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param [in] stream [optional] HIP stream object. Default is \p 0 (default stream).
+/// \param [in] debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors. Default value is \p false.
 ///
 /// \returns \p hipSuccess (\p 0) after successful sort; otherwise a HIP runtime error of
@@ -320,39 +349,39 @@ hipError_t merge(void * temporary_storage,
 /// if \p temporary_storage in a null pointer.
 /// * Accepts custom compare_functions for merging across the device.
 ///
-/// \tparam Config - [optional] Configuration of the primitive, must be `default_config` or `merge_config`.
-/// \tparam KeysInputIterator1 - random-access iterator type of the first keys input range. Must meet the
+/// \tparam Config [optional] Configuration of the primitive, must be `default_config` or `merge_config`.
+/// \tparam KeysInputIterator1 random-access iterator type of the first keys input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam KeysInputIterator2 - random-access iterator type of the second keys input range. Must meet the
+/// \tparam KeysInputIterator2 random-access iterator type of the second keys input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam KeysOutputIterator - random-access iterator type of the keys output range. Must meet the
+/// \tparam KeysOutputIterator random-access iterator type of the keys output range. Must meet the
 /// requirements of a C++ OutputIterator concept. It can be a simple pointer type.
-/// \tparam ValuesInputIterator1 - random-access iterator type of the first values input range. Must meet the
+/// \tparam ValuesInputIterator1 random-access iterator type of the first values input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam ValuesInputIterator2 - random-access iterator type of the second values input range. Must meet the
+/// \tparam ValuesInputIterator2 random-access iterator type of the second values input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam ValuesOutputIterator - random-access iterator type of the values output range. Must meet the
+/// \tparam ValuesOutputIterator random-access iterator type of the values output range. Must meet the
 /// requirements of a C++ OutputIterator concept. It can be a simple pointer type.
 ///
-/// \param [in] temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param [in] temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// \p storage_size and function returns without performing the sort operation.
-/// \param [in,out] storage_size - reference to a size (in bytes) of \p temporary_storage.
-/// \param [in] keys_input1 - iterator to the first key in the first range to merge.
-/// \param [in] keys_input2 - iterator to the first key in the second range to merge.
-/// \param [out] keys_output - iterator to the first key in the output range.
-/// \param [in] values_input1 - iterator to the first value in the first range to merge.
-/// \param [in] values_input2 - iterator to the first value in the second range to merge.
-/// \param [out] values_output - iterator to the first value in the output range.
-/// \param [in] input1_size - number of element in the first input range.
-/// \param [in] input2_size - number of element in the second input range.
-/// \param [in] compare_function - binary operation function object that will be used for key comparison.
+/// \param [in,out] storage_size reference to a size (in bytes) of \p temporary_storage.
+/// \param [in] keys_input1 iterator to the first key in the first range to merge.
+/// \param [in] keys_input2 iterator to the first key in the second range to merge.
+/// \param [out] keys_output iterator to the first key in the output range.
+/// \param [in] values_input1 iterator to the first value in the first range to merge.
+/// \param [in] values_input2 iterator to the first value in the second range to merge.
+/// \param [out] values_output iterator to the first value in the output range.
+/// \param [in] input1_size number of element in the first input range.
+/// \param [in] input2_size number of element in the second input range.
+/// \param [in] compare_function binary operation function object that will be used for key comparison.
 /// The signature of the function should be equivalent to the following:
 /// <tt>bool f(const T &a, const T &b);</tt>. The signature does not need to have
 /// <tt>const &</tt>, but function object must not modify the objects passed to it.
 /// The default value is \p BinaryFunction().
-/// \param [in] stream - [optional] HIP stream object. Default is \p 0 (default stream).
-/// \param [in] debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param [in] stream [optional] HIP stream object. Default is \p 0 (default stream).
+/// \param [in] debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors. Default value is \p false.
 ///
 /// \returns \p hipSuccess (\p 0) after successful sort; otherwise a HIP runtime error of
