@@ -1,4 +1,4 @@
-// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -30,12 +30,14 @@
 #include "detail/device_scan_common.hpp"
 #include "detail/lookback_scan_state.hpp"
 
+#include "../common.hpp"
 #include "../config.hpp"
 #include "../detail/temp_storage.hpp"
 #include "../detail/various.hpp"
 #include "../functional.hpp"
 #include "../intrinsics/thread.hpp"
 #include "../iterator/constant_iterator.hpp"
+#include "../type_traits.hpp"
 
 #include <chrono>
 #include <iostream>
@@ -51,13 +53,12 @@ namespace detail
 
 template<typename LookBackScanState, typename AccumulatorType>
 ROCPRIM_KERNEL ROCPRIM_LAUNCH_BOUNDS(ROCPRIM_DEFAULT_MAX_BLOCK_SIZE) void
-    reduce_by_key_init_kernel(LookBackScanState              lookback_scan_state,
-                              const unsigned int             number_of_tiles,
-                              ordered_block_id<unsigned int> ordered_bid,
-                              const bool                     is_first_launch,
-                              const unsigned int             tile_save_idx,
-                              std::size_t* const             global_head_count,
-                              AccumulatorType* const         previous_accumulated)
+    reduce_by_key_init_kernel(LookBackScanState      lookback_scan_state,
+                              const unsigned int     number_of_blocks,
+                              const bool             is_first_launch,
+                              const unsigned int     block_save_idx,
+                              std::size_t* const     global_head_count,
+                              AccumulatorType* const previous_accumulated)
 {
     const unsigned int block_id        = ::rocprim::detail::block_id<0>();
     const unsigned int block_size      = ::rocprim::detail::block_size<0>();
@@ -82,13 +83,13 @@ ROCPRIM_KERNEL ROCPRIM_LAUNCH_BOUNDS(ROCPRIM_DEFAULT_MAX_BLOCK_SIZE) void
             *previous_accumulated = ::rocprim::get<1>(value);
         };
         access_indexed_lookback_value(lookback_scan_state,
-                                      number_of_tiles,
-                                      tile_save_idx,
+                                      number_of_blocks,
+                                      block_save_idx,
                                       flat_thread_id,
                                       update_func);
     }
 
-    init_lookback_scan_state(lookback_scan_state, number_of_tiles, ordered_bid, flat_thread_id);
+    init_lookback_scan_state(lookback_scan_state, number_of_blocks, flat_thread_id);
 }
 
 template<lookback_scan_determinism Determinism,
@@ -102,23 +103,20 @@ template<lookback_scan_determinism Determinism,
          typename CompareFunction,
          typename BinaryOp,
          typename LookbackScanState>
-ROCPRIM_KERNEL
-    __launch_bounds__(device_params<Config>().kernel_config.block_size) void reduce_by_key_kernel(
-        const KeyIterator                    keys_input,
-        const ValueIterator                  values_input,
-        const UniqueIterator                 unique_keys,
-        const ReductionIterator              reductions,
-        const UniqueCountIterator            unique_count,
-        const BinaryOp                       reduce_op,
-        const CompareFunction                compare,
-        const LookbackScanState              scan_state,
-        const ordered_block_id<unsigned int> ordered_tile_id,
-        const std::size_t                    starting_tile,
-        const std::size_t                    total_number_of_tiles,
-        const std::size_t                    size,
-        const std::size_t* const             global_head_count,
-        const AccumulatorType* const         previous_accumulated,
-        const std::size_t                    number_of_tiles_launch)
+ROCPRIM_KERNEL ROCPRIM_LAUNCH_BOUNDS(device_params<Config>().kernel_config.block_size) void
+    reduce_by_key_kernel(const KeyIterator            keys_input,
+                         const ValueIterator          values_input,
+                         const UniqueIterator         unique_keys,
+                         const ReductionIterator      reductions,
+                         const UniqueCountIterator    unique_count,
+                         const BinaryOp               reduce_op,
+                         const CompareFunction        compare,
+                         const LookbackScanState      scan_state,
+                         const std::size_t            starting_block,
+                         const std::size_t            total_number_of_blocks,
+                         const std::size_t            size,
+                         const std::size_t* const     global_head_count,
+                         const AccumulatorType* const previous_accumulated)
 {
     reduce_by_key::kernel_impl<Determinism, Config>(keys_input,
                                                     values_input,
@@ -128,61 +126,36 @@ ROCPRIM_KERNEL
                                                     reduce_op,
                                                     compare,
                                                     scan_state,
-                                                    ordered_tile_id,
-                                                    starting_tile,
-                                                    total_number_of_tiles,
+                                                    starting_block,
+                                                    total_number_of_blocks,
                                                     size,
                                                     global_head_count,
-                                                    previous_accumulated,
-                                                    number_of_tiles_launch);
+                                                    previous_accumulated);
 }
 
-#define ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR(name, size, start)                           \
-    do                                                                                           \
-    {                                                                                            \
-        auto _error = hipGetLastError();                                                         \
-        if(_error != hipSuccess)                                                                 \
-            return _error;                                                                       \
-        if(debug_synchronous)                                                                    \
-        {                                                                                        \
-            std::cout << name << "(" << size << ")";                                             \
-            auto __error = hipStreamSynchronize(stream);                                         \
-            if(__error != hipSuccess)                                                            \
-                return __error;                                                                  \
-            auto _end = std::chrono::high_resolution_clock::now();                               \
-            auto _d   = std::chrono::duration_cast<std::chrono::duration<double>>(_end - start); \
-            std::cout << " " << _d.count() * 1000 << " ms" << '\n';                              \
-        }                                                                                        \
-    }                                                                                            \
-    while(false)
-
 template<lookback_scan_determinism Determinism,
-         class Config,
-         class KeysInputIterator,
-         class ValuesInputIterator,
-         class UniqueOutputIterator,
-         class AggregatesOutputIterator,
-         class UniqueCountOutputIterator,
-         class BinaryFunction,
-         class KeyCompareFunction>
-hipError_t reduce_by_key_impl(void*                     temporary_storage,
-                              size_t&                   storage_size,
-                              KeysInputIterator         keys_input,
-                              ValuesInputIterator       values_input,
-                              const size_t              size,
-                              UniqueOutputIterator      unique_output,
-                              AggregatesOutputIterator  aggregates_output,
-                              UniqueCountOutputIterator unique_count_output,
-                              BinaryFunction            reduce_op,
-                              KeyCompareFunction        key_compare_op,
-                              const hipStream_t         stream,
-                              const bool                debug_synchronous)
+         typename config,
+         typename KeysInputIterator,
+         typename ValuesInputIterator,
+         typename UniqueOutputIterator,
+         typename AggregatesOutputIterator,
+         typename UniqueCountOutputIterator,
+         typename BinaryFunction,
+         typename KeyCompareFunction>
+hipError_t reduce_by_key_impl_wrapped_config(void*                     temporary_storage,
+                                             size_t&                   storage_size,
+                                             KeysInputIterator         keys_input,
+                                             ValuesInputIterator       values_input,
+                                             const size_t              size,
+                                             UniqueOutputIterator      unique_output,
+                                             AggregatesOutputIterator  aggregates_output,
+                                             UniqueCountOutputIterator unique_count_output,
+                                             BinaryFunction            reduce_op,
+                                             KeyCompareFunction        key_compare_op,
+                                             const hipStream_t         stream,
+                                             const bool                debug_synchronous)
 {
-    using key_type         = reduce_by_key::value_type_t<KeysInputIterator>;
     using accumulator_type = reduce_by_key::accumulator_type_t<ValuesInputIterator, BinaryFunction>;
-
-    using config = wrapped_reduce_by_key_config<Config, key_type, accumulator_type, BinaryFunction>;
-
     detail::target_arch target_arch;
     hipError_t          result = host_target_arch(stream, target_arch);
     if(result != hipSuccess)
@@ -196,12 +169,8 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
     using scan_state_with_sleep_type
         = reduce_by_key::lookback_scan_state_t<accumulator_type, /*UseSleep=*/true>;
 
-    using ordered_tile_id_type = detail::ordered_block_id<unsigned int>;
-
     const unsigned int block_size      = params.kernel_config.block_size;
-    const unsigned int tiles_per_block = params.tiles_per_block;
-    const unsigned int items_per_tile  = block_size * params.kernel_config.items_per_thread;
-    const unsigned int items_per_block = items_per_tile * tiles_per_block;
+    const unsigned int items_per_block = block_size * params.kernel_config.items_per_thread;
 
     const size_t size_limit = params.kernel_config.size_limit;
     const size_t aligned_size_limit
@@ -210,20 +179,18 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
     const size_t limited_size     = std::min<size_t>(size, aligned_size_limit);
     const bool   use_limited_size = limited_size == aligned_size_limit;
 
-    // Number of tiles in a single launch (or the only launch if it fits)
-    const std::size_t number_of_tiles  = detail::ceiling_div(limited_size, items_per_tile);
-    const std::size_t number_of_blocks = detail::ceiling_div(number_of_tiles, tiles_per_block);
+    // Number of blocks in a single launch (or the only launch if it fits)
+    const std::size_t number_of_blocks = detail::ceiling_div(limited_size, items_per_block);
 
     // Calculate required temporary storage
-    void*                          scan_state_storage;
-    ordered_tile_id_type::id_type* ordered_bid_storage;
+    void* scan_state_storage;
     // The number of segment heads in previous launches.
     std::size_t* d_global_head_count = nullptr;
     // The running accumulation across the launch boundary.
     accumulator_type* d_previous_accumulated = nullptr;
 
     detail::temp_storage::layout layout{};
-    result = scan_state_type::get_temp_storage_layout(number_of_tiles, stream, layout);
+    result = scan_state_type::get_temp_storage_layout(number_of_blocks, stream, layout);
     if(result != hipSuccess)
     {
         return result;
@@ -235,8 +202,6 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
         detail::temp_storage::make_linear_partition(
             // This is valid even with scan_state_with_sleep_type
             detail::temp_storage::make_partition(&scan_state_storage, layout),
-            detail::temp_storage::make_partition(&ordered_bid_storage,
-                                                 ordered_tile_id_type::get_temp_storage_layout()),
             detail::temp_storage::ptr_aligned_array(&d_global_head_count, use_limited_size ? 1 : 0),
             detail::temp_storage::ptr_aligned_array(&d_previous_accumulated,
                                                     use_limited_size ? 1 : 0)));
@@ -252,7 +217,7 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
         return result;
     }
     scan_state_type scan_state{};
-    result = scan_state_type::create(scan_state, scan_state_storage, number_of_tiles, stream);
+    result = scan_state_type::create(scan_state, scan_state_storage, number_of_blocks, stream);
     if(result != hipSuccess)
     {
         return result;
@@ -260,7 +225,7 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
     scan_state_with_sleep_type scan_state_with_sleep{};
     result = scan_state_with_sleep_type::create(scan_state_with_sleep,
                                                 scan_state_storage,
-                                                number_of_tiles,
+                                                number_of_blocks,
                                                 stream);
     if(result != hipSuccess)
     {
@@ -280,8 +245,6 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
         }
     };
 
-    auto ordered_bid = ordered_tile_id_type::create(ordered_bid_storage);
-
     if(size == 0)
     {
         // Fill out unique_count_output with zero
@@ -293,9 +256,9 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
                                   debug_synchronous);
     }
 
-    // Total number of tiles in all launches
-    const std::size_t total_number_of_tiles = ceiling_div(size, items_per_tile);
-    const std::size_t number_of_launch      = ceiling_div(size, limited_size);
+    // Total number of blocks in all launches
+    const std::size_t total_number_of_blocks = ceiling_div(size, items_per_block);
+    const std::size_t number_of_launch       = ceiling_div(size, limited_size);
 
     if(debug_synchronous)
     {
@@ -304,29 +267,24 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
         std::cout << "use_limited_size:   " << std::boolalpha << use_limited_size << '\n';
         std::cout << "number_of_launch:   " << number_of_launch << '\n';
         std::cout << "block_size:         " << block_size << '\n';
-        std::cout << "tiles_per_block:    " << tiles_per_block << '\n';
-        std::cout << "number_of_tiles:    " << number_of_tiles << '\n';
         std::cout << "number_of_blocks:   " << number_of_blocks << '\n';
-        std::cout << "items_per_tile:     " << items_per_tile << '\n';
+        std::cout << "items_per_block:    " << items_per_block << '\n';
     }
 
-    for(size_t i = 0, offset = 0; i < number_of_launch; i++, offset += limited_size)
+    for(size_t i = 0, offset = 0; i < number_of_launch; ++i, offset += limited_size)
     {
         const std::size_t current_size = std::min<std::size_t>(size - offset, limited_size);
-        const std::size_t number_of_tiles_launch = ceiling_div(current_size, items_per_tile);
-        const std::size_t number_of_blocks_launch
-            = ceiling_div(number_of_tiles_launch, tiles_per_block);
+        const std::size_t number_of_blocks_launch = ceiling_div(current_size, items_per_block);
 
         // Start point for time measurements
-        std::chrono::high_resolution_clock::time_point start;
+        std::chrono::steady_clock::time_point start;
         if(debug_synchronous)
         {
             std::cout << "index:            " << i << '\n';
             std::cout << "current_size:     " << current_size << '\n';
-            std::cout << "number of tiles:  " << number_of_tiles_launch << '\n';
             std::cout << "number of blocks: " << number_of_blocks_launch << '\n';
 
-            start = std::chrono::high_resolution_clock::now();
+            start = std::chrono::steady_clock::now();
         }
 
         with_scan_state(
@@ -334,19 +292,18 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
             {
                 const unsigned int block_size = ROCPRIM_DEFAULT_MAX_BLOCK_SIZE;
                 const std::size_t  grid_size
-                    = detail::ceiling_div(number_of_tiles_launch, block_size);
+                    = detail::ceiling_div(number_of_blocks_launch, block_size);
 
                 reduce_by_key_init_kernel<<<dim3(grid_size), dim3(block_size), 0, stream>>>(
                     scan_state,
-                    number_of_tiles_launch,
-                    ordered_bid,
+                    number_of_blocks_launch,
                     i == 0,
-                    number_of_tiles - 1,
+                    number_of_blocks - 1,
                     d_global_head_count,
                     d_previous_accumulated);
             });
-        ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR("init_lookback_scan_state_kernel",
-                                                    number_of_tiles_launch,
+        ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR("reduce_by_key_init_kernel",
+                                                    number_of_blocks_launch,
                                                     start);
 
         with_scan_state(
@@ -362,13 +319,11 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
                         reduce_op,
                         key_compare_op,
                         scan_state,
-                        ordered_bid,
-                        i * number_of_tiles,
-                        total_number_of_tiles,
+                        i * number_of_blocks,
+                        total_number_of_blocks,
                         size,
                         i > 0 ? d_global_head_count : nullptr,
-                        i > 0 ? d_previous_accumulated : nullptr,
-                        number_of_tiles_launch);
+                        i > 0 ? d_previous_accumulated : nullptr);
             });
         ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR("reduce_by_key_kernel", current_size, start);
     }
@@ -376,7 +331,46 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
     return hipSuccess;
 }
 
-#undef ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR
+template<lookback_scan_determinism Determinism,
+         typename Config,
+         typename KeysInputIterator,
+         typename ValuesInputIterator,
+         typename UniqueOutputIterator,
+         typename AggregatesOutputIterator,
+         typename UniqueCountOutputIterator,
+         typename BinaryFunction,
+         typename KeyCompareFunction>
+hipError_t reduce_by_key_impl(void*                     temporary_storage,
+                              size_t&                   storage_size,
+                              KeysInputIterator         keys_input,
+                              ValuesInputIterator       values_input,
+                              const size_t              size,
+                              UniqueOutputIterator      unique_output,
+                              AggregatesOutputIterator  aggregates_output,
+                              UniqueCountOutputIterator unique_count_output,
+                              BinaryFunction            reduce_op,
+                              KeyCompareFunction        key_compare_op,
+                              const hipStream_t         stream,
+                              const bool                debug_synchronous)
+{
+    using key_type         = ::rocprim::detail::value_type_t<KeysInputIterator>;
+    using accumulator_type = reduce_by_key::accumulator_type_t<ValuesInputIterator, BinaryFunction>;
+
+    using config = wrapped_reduce_by_key_config<Config, key_type, accumulator_type, BinaryFunction>;
+
+    return detail::reduce_by_key_impl_wrapped_config<Determinism, config>(temporary_storage,
+                                                                          storage_size,
+                                                                          keys_input,
+                                                                          values_input,
+                                                                          size,
+                                                                          unique_output,
+                                                                          aggregates_output,
+                                                                          unique_count_output,
+                                                                          reduce_op,
+                                                                          key_compare_op,
+                                                                          stream,
+                                                                          debug_synchronous);
+}
 
 } // namespace detail
 
@@ -403,46 +397,46 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
 /// * Ranges specified by \p unique_output and \p aggregates_output must have at least
 /// <tt>*unique_count_output</tt> (i.e. the number of unique keys) elements.
 ///
-/// \tparam Config - [optional] Configuration of the primitive, must be `default_config` or `reduce_by_key_config`.
-/// \tparam KeysInputIterator - random-access iterator type of the input range. Must meet the
+/// \tparam Config [optional] Configuration of the primitive, must be `default_config` or `reduce_by_key_config`.
+/// \tparam KeysInputIterator random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam ValuesInputIterator - random-access iterator type of the input range. Must meet the
+/// \tparam ValuesInputIterator random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam UniqueOutputIterator - random-access iterator type of the output range. Must meet the
+/// \tparam UniqueOutputIterator random-access iterator type of the output range. Must meet the
 /// requirements of a C++ OutputIterator concept. It can be a simple pointer type.
-/// \tparam AggregatesOutputIterator - random-access iterator type of the output range. Must meet the
+/// \tparam AggregatesOutputIterator random-access iterator type of the output range. Must meet the
 /// requirements of a C++ OutputIterator concept. It can be a simple pointer type.
-/// \tparam UniqueCountOutputIterator - random-access iterator type of the output range. Must meet the
+/// \tparam UniqueCountOutputIterator random-access iterator type of the output range. Must meet the
 /// requirements of a C++ OutputIterator concept. It can be a simple pointer type.
-/// \tparam BinaryFunction - type of binary function used for reduction. Default type
+/// \tparam BinaryFunction type of binary function used for reduction. Default type
 /// is \p rocprim::plus<T>, where \p T is a \p value_type of \p ValuesInputIterator.
-/// \tparam KeyCompareFunction - type of binary function used to determine keys equality. Default type
+/// \tparam KeyCompareFunction type of binary function used to determine keys equality. Default type
 /// is \p rocprim::equal_to<T>, where \p T is a \p value_type of \p KeysInputIterator.
 ///
-/// \param [in] temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param [in] temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// \p storage_size and function returns without performing the reduction operation.
-/// \param [in,out] storage_size - reference to a size (in bytes) of \p temporary_storage.
-/// \param [in] keys_input - iterator to the first element in the range of keys.
-/// \param [in] values_input - iterator to the first element in the range of values to reduce.
-/// \param [in] size - number of element in the input range.
-/// \param [out] unique_output - iterator to the first element in the output range of unique keys.
-/// \param [out] aggregates_output - iterator to the first element in the output range of reductions.
-/// \param [out] unique_count_output - iterator to total number of groups.
-/// \param [in] reduce_op - binary operation function object that will be used for reduction.
+/// \param [in,out] storage_size reference to a size (in bytes) of \p temporary_storage.
+/// \param [in] keys_input iterator to the first element in the range of keys.
+/// \param [in] values_input iterator to the first element in the range of values to reduce.
+/// \param [in] size number of element in the input range.
+/// \param [out] unique_output iterator to the first element in the output range of unique keys.
+/// \param [out] aggregates_output iterator to the first element in the output range of reductions.
+/// \param [out] unique_count_output iterator to total number of groups.
+/// \param [in] reduce_op binary operation function object that will be used for reduction.
 /// The signature of the function should be equivalent to the following:
 /// <tt>T f(const T &a, const T &b);</tt>. The signature does not need to have
 /// <tt>const &</tt>, but function object must not modify the objects passed to it and must not have
 /// any side effects since the function may be called on uninitalized data.
 /// Default is BinaryFunction().
-/// \param [in] key_compare_op - binary operation function object that will be used to determine key equality.
+/// \param [in] key_compare_op binary operation function object that will be used to determine key equality.
 /// The signature of the function should be equivalent to the following:
 /// <tt>bool f(const T &a, const T &b);</tt>. The signature does not need to have
 /// <tt>const &</tt>, but function object must not modify the objects passed to it and must not have
 /// any side effects since the function may be called on uninitalized data.
 /// Default is KeyCompareFunction().
-/// \param [in] stream - [optional] HIP stream object. Default is \p 0 (default stream).
-/// \param [in] debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param [in] stream [optional] HIP stream object. Default is \p 0 (default stream).
+/// \param [in] debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors. Default value is \p false.
 ///
 /// \returns \p hipSuccess (\p 0) after successful reduction; otherwise a HIP runtime error of
@@ -487,15 +481,15 @@ hipError_t reduce_by_key_impl(void*                     temporary_storage,
 /// // unique_count_output: [4]
 /// \endcode
 /// \endparblock
-template<class Config = default_config,
-         class KeysInputIterator,
-         class ValuesInputIterator,
-         class UniqueOutputIterator,
-         class AggregatesOutputIterator,
-         class UniqueCountOutputIterator,
-         class BinaryFunction
+template<typename Config = default_config,
+         typename KeysInputIterator,
+         typename ValuesInputIterator,
+         typename UniqueOutputIterator,
+         typename AggregatesOutputIterator,
+         typename UniqueCountOutputIterator,
+         typename BinaryFunction
          = ::rocprim::plus<typename std::iterator_traits<ValuesInputIterator>::value_type>,
-         class KeyCompareFunction
+         typename KeyCompareFunction
          = ::rocprim::equal_to<typename std::iterator_traits<KeysInputIterator>::value_type>>
 inline hipError_t reduce_by_key(void*                     temporary_storage,
                                 size_t&                   storage_size,
@@ -532,15 +526,15 @@ inline hipError_t reduce_by_key(void*                     temporary_storage,
 /// non-associative scan operators like floating point arithmetic operations.
 /// Refer to the documentation for \link reduce_by_key() rocprim::reduce_by_key \endlink
 /// for a detailed description of this function.
-template<class Config = default_config,
-         class KeysInputIterator,
-         class ValuesInputIterator,
-         class UniqueOutputIterator,
-         class AggregatesOutputIterator,
-         class UniqueCountOutputIterator,
-         class BinaryFunction
+template<typename Config = default_config,
+         typename KeysInputIterator,
+         typename ValuesInputIterator,
+         typename UniqueOutputIterator,
+         typename AggregatesOutputIterator,
+         typename UniqueCountOutputIterator,
+         typename BinaryFunction
          = ::rocprim::plus<typename std::iterator_traits<ValuesInputIterator>::value_type>,
-         class KeyCompareFunction
+         typename KeyCompareFunction
          = ::rocprim::equal_to<typename std::iterator_traits<KeysInputIterator>::value_type>>
 inline hipError_t deterministic_reduce_by_key(void*                     temporary_storage,
                                               size_t&                   storage_size,
