@@ -1,4 +1,4 @@
-// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -25,6 +25,7 @@
 #include <iterator>
 #include <type_traits>
 
+#include "../common.hpp"
 #include "../config.hpp"
 #include "../detail/temp_storage.hpp"
 #include "../detail/various.hpp"
@@ -47,6 +48,7 @@ namespace detail
 
 // Single kernel scan (performs scan on one thread block only)
 template<bool Exclusive,
+         bool UseInitialValue,
          class Config,
          class InputIterator,
          class OutputIterator,
@@ -81,11 +83,11 @@ ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE void single_scan_kernel_impl(InputIterator
     block_load_type().load(input, values, input_size, *(input), storage.load);
     ::rocprim::syncthreads(); // sync threads to reuse shared memory
 
-    single_scan_block_scan<Exclusive, block_scan_type>(values, // input
-                                                       values, // output
-                                                       initial_value,
-                                                       storage.scan,
-                                                       scan_op);
+    single_scan_block_scan<Exclusive, UseInitialValue, block_scan_type>(values, // input
+                                                                        values, // output
+                                                                        initial_value,
+                                                                        storage.scan,
+                                                                        scan_op);
     ::rocprim::syncthreads(); // sync threads to reuse shared memory
 
     // Save values into output array
@@ -93,31 +95,33 @@ ROCPRIM_DEVICE ROCPRIM_FORCE_INLINE void single_scan_kernel_impl(InputIterator
 }
 
 template<bool Exclusive,
+         bool UseInitialValue,
          class Config,
          class InputIterator,
          class OutputIterator,
          class BinaryFunction,
          class InitValueType,
          class AccType>
-ROCPRIM_KERNEL
-    __launch_bounds__(device_params<Config>().kernel_config.block_size) void single_scan_kernel(
-        InputIterator       input,
-        const size_t        size,
-        const InitValueType initial_value,
-        OutputIterator      output,
-        BinaryFunction      scan_op)
+ROCPRIM_KERNEL ROCPRIM_LAUNCH_BOUNDS(device_params<Config>().kernel_config.block_size) void
+    single_scan_kernel(InputIterator       input,
+                       const size_t        size,
+                       const InitValueType initial_value,
+                       OutputIterator      output,
+                       BinaryFunction      scan_op)
 {
-    single_scan_kernel_impl<Exclusive, Config>(input,
-                                               size,
-                                               static_cast<AccType>(get_input_value(initial_value)),
-                                               output,
-                                               scan_op);
+    single_scan_kernel_impl<Exclusive, UseInitialValue, Config>(
+        input,
+        size,
+        static_cast<AccType>(get_input_value(initial_value)),
+        output,
+        scan_op);
 }
 
 // Single pass (look-back kernels)
 
 template<lookback_scan_determinism Determinism,
          bool                      Exclusive,
+         bool                      UseInitialValue,
          class Config,
          class InputIterator,
          class OutputIterator,
@@ -125,21 +129,20 @@ template<lookback_scan_determinism Determinism,
          class InitValueType,
          class AccType,
          class LookBackScanState>
-ROCPRIM_KERNEL
-    __launch_bounds__(device_params<Config>().kernel_config.block_size) void lookback_scan_kernel(
-        InputIterator       input,
-        OutputIterator      output,
-        const size_t        size,
-        const InitValueType initial_value,
-        BinaryFunction      scan_op,
-        LookBackScanState   lookback_scan_state,
-        const unsigned int  number_of_blocks,
-        AccType*            previous_last_element = nullptr,
-        AccType*            new_last_element      = nullptr,
-        bool                override_first_value  = false,
-        bool                save_last_value       = false)
+ROCPRIM_KERNEL ROCPRIM_LAUNCH_BOUNDS(device_params<Config>().kernel_config.block_size) void
+    lookback_scan_kernel(InputIterator      input,
+                         OutputIterator     output,
+                         const size_t       size,
+                         InitValueType      initial_value,
+                         BinaryFunction     scan_op,
+                         LookBackScanState  lookback_scan_state,
+                         const unsigned int number_of_blocks,
+                         AccType*           previous_last_element = nullptr,
+                         AccType*           new_last_element      = nullptr,
+                         bool               override_first_value  = false,
+                         bool               save_last_value       = false)
 {
-    lookback_scan_kernel_impl<Determinism, Exclusive, Config>(
+    lookback_scan_kernel_impl<Determinism, Exclusive, UseInitialValue, Config>(
         input,
         output,
         size,
@@ -153,34 +156,9 @@ ROCPRIM_KERNEL
         save_last_value);
 }
 
-#define ROCPRIM_DETAIL_HIP_SYNC(name, size, start) \
-    if(debug_synchronous) \
-    { \
-        std::cout << name << "(" << size << ")"; \
-        auto error = hipStreamSynchronize(stream); \
-        if(error != hipSuccess) return error; \
-        auto end = std::chrono::high_resolution_clock::now(); \
-        auto d = std::chrono::duration_cast<std::chrono::duration<double>>(end - start); \
-        std::cout << " " << d.count() * 1000 << " ms" << '\n'; \
-    }
-
-#define ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR(name, size, start) \
-    { \
-        auto _error = hipGetLastError(); \
-        if(_error != hipSuccess) return _error; \
-        if(debug_synchronous) \
-        { \
-            std::cout << name << "(" << size << ")"; \
-            auto __error = hipStreamSynchronize(stream); \
-            if(__error != hipSuccess) return __error; \
-            auto _end = std::chrono::high_resolution_clock::now(); \
-            auto _d = std::chrono::duration_cast<std::chrono::duration<double>>(_end - start); \
-            std::cout << " " << _d.count() * 1000 << " ms" << '\n'; \
-        } \
-    }
-
 template<lookback_scan_determinism Determinism,
          bool                      Exclusive,
+         bool                      UseInitialValue,
          class Config,
          class InputIterator,
          class OutputIterator,
@@ -217,10 +195,10 @@ inline auto scan_impl(void*               temporary_storage,
     const size_t size_limit = params.kernel_config.size_limit;
     const size_t aligned_size_limit
         = ::rocprim::max<size_t>(size_limit - size_limit % items_per_block, items_per_block);
-    size_t limited_size = std::min<size_t>(size, aligned_size_limit);
+    size_t     limited_size     = std::min<size_t>(size, aligned_size_limit);
     const bool use_limited_size = limited_size == aligned_size_limit;
 
-    unsigned int number_of_blocks = (limited_size + items_per_block - 1)/items_per_block;
+    unsigned int number_of_blocks = (limited_size + items_per_block - 1) / items_per_block;
 
     // Pointer to array with block_prefixes
     void*    scan_state_storage;
@@ -250,9 +228,9 @@ inline auto scan_impl(void*               temporary_storage,
     }
 
     // Start point for time measurements
-    std::chrono::high_resolution_clock::time_point start;
+    std::chrono::steady_clock::time_point start;
 
-    if( number_of_blocks == 0u )
+    if(number_of_blocks == 0u)
         return hipSuccess;
 
     if(number_of_blocks > 1 || use_limited_size)
@@ -292,14 +270,15 @@ inline auto scan_impl(void*               temporary_storage,
             }
         };
 
-        if(debug_synchronous) start = std::chrono::high_resolution_clock::now();
+        if(debug_synchronous)
+            start = std::chrono::steady_clock::now();
 
-        size_t number_of_launch = (size + limited_size - 1)/limited_size;
-        for (size_t i = 0, offset = 0; i < number_of_launch; i++, offset+=limited_size )
+        size_t number_of_launch = (size + limited_size - 1) / limited_size;
+        for(size_t i = 0, offset = 0; i < number_of_launch; i++, offset += limited_size)
         {
             size_t current_size = std::min<size_t>(size - offset, limited_size);
-            number_of_blocks = (current_size + items_per_block - 1)/items_per_block;
-            auto grid_size = (number_of_blocks + block_size - 1)/block_size;
+            number_of_blocks    = (current_size + items_per_block - 1) / items_per_block;
+            auto grid_size      = (number_of_blocks + block_size - 1) / block_size;
 
             if(debug_synchronous)
             {
@@ -323,9 +302,10 @@ inline auto scan_impl(void*               temporary_storage,
                 });
             ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR("init_lookback_scan_state_kernel",
                                                         number_of_blocks,
-                                                        start)
+                                                        start);
 
-            if(debug_synchronous) start = std::chrono::high_resolution_clock::now();
+            if(debug_synchronous)
+                start = std::chrono::steady_clock::now();
             grid_size = number_of_blocks;
 
             if(debug_synchronous)
@@ -343,6 +323,7 @@ inline auto scan_impl(void*               temporary_storage,
                 {
                     lookback_scan_kernel<Determinism,
                                          Exclusive,
+                                         UseInitialValue,
                                          config,
                                          InputIterator,
                                          OutputIterator,
@@ -361,7 +342,9 @@ inline auto scan_impl(void*               temporary_storage,
                                                                            i != size_t(0),
                                                                            number_of_launch > 1);
                 });
-            ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR("lookback_scan_kernel", current_size, start)
+            ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR("lookback_scan_kernel",
+                                                        current_size,
+                                                        start);
 
             // Swap the last_elements
             if(number_of_launch > 1)
@@ -372,7 +355,8 @@ inline auto scan_impl(void*               temporary_storage,
                                                         ::rocprim::identity<AccType>(),
                                                         stream,
                                                         debug_synchronous);
-                if(error != hipSuccess) return error;
+                if(error != hipSuccess)
+                    return error;
             }
         }
     }
@@ -384,10 +368,11 @@ inline auto scan_impl(void*               temporary_storage,
             std::cout << "block_size " << block_size << '\n';
             std::cout << "number of blocks " << number_of_blocks << '\n';
             std::cout << "items_per_block " << items_per_block << '\n';
-            start = std::chrono::high_resolution_clock::now();
+            start = std::chrono::steady_clock::now();
         }
 
         single_scan_kernel<Exclusive, // flag for exclusive scan operation
+                           UseInitialValue,
                            config,
                            InputIterator,
                            OutputIterator,
@@ -400,10 +385,7 @@ inline auto scan_impl(void*               temporary_storage,
     return hipSuccess;
 }
 
-#undef ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR
-#undef ROCPRIM_DETAIL_HIP_SYNC
-
-} // end of detail namespace
+} // namespace detail
 
 /// \brief Parallel inclusive scan primitive for device level.
 ///
@@ -425,31 +407,31 @@ inline auto scan_impl(void*               temporary_storage,
 /// * By default, the input type is used for accumulation. A custom type
 /// can be specified using the \p AccType type parameter, see the example below.
 ///
-/// \tparam Config - [optional] Configuration of the primitive, must be `default_config` or `scan_config`.
-/// \tparam InputIterator - random-access iterator type of the input range. Must meet the
+/// \tparam Config [optional] Configuration of the primitive, must be `default_config` or `scan_config`.
+/// \tparam InputIterator random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam OutputIterator - random-access iterator type of the output range. Must meet the
+/// \tparam OutputIterator random-access iterator type of the output range. Must meet the
 /// requirements of a C++ OutputIterator concept. It can be a simple pointer type.
-/// \tparam BinaryFunction - type of binary function used for scan. Default type
+/// \tparam BinaryFunction type of binary function used for scan. Default type
 /// is \p rocprim::plus<T>, where \p T is a \p value_type of \p InputIterator.
-/// \tparam AccType - accumulator type used to propagate the scanned values. Default type
-/// is value type of the input iterator.
+/// \tparam AccType accumulator type used to propagate the scanned values. The default is the type that
+/// is returned by a function of type BinaryFunction when it is passed an InputIterator value.
 ///
-/// \param [in] temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param [in] temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// \p storage_size and function returns without performing the scan operation.
-/// \param [in,out] storage_size - reference to a size (in bytes) of \p temporary_storage.
-/// \param [in] input - iterator to the first element in the range to scan.
-/// \param [out] output - iterator to the first element in the output range. It can be
+/// \param [in,out] storage_size reference to a size (in bytes) of \p temporary_storage.
+/// \param [in] input iterator to the first element in the range to scan.
+/// \param [out] output iterator to the first element in the output range. It can be
 /// same as \p input.
-/// \param [in] size - number of element in the input range.
-/// \param [in] scan_op - binary operation function object that will be used for scan.
+/// \param [in] size number of element in the input range.
+/// \param [in] scan_op binary operation function object that will be used for scan.
 /// The signature of the function should be equivalent to the following:
 /// <tt>T f(const T &a, const T &b);</tt>. The signature does not need to have
 /// <tt>const &</tt>, but function object must not modify the objects passed to it.
 /// Default is BinaryFunction().
-/// \param [in] stream - [optional] HIP stream object. Default is \p 0 (default stream).
-/// \param [in] debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param [in] stream [optional] HIP stream object. Default is \p 0 (default stream).
+/// \param [in] debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors. Default value is \p false.
 ///
 /// \returns \p hipSuccess (\p 0) after successful scan; otherwise a HIP runtime error of
@@ -524,7 +506,9 @@ template<class Config = default_config,
          class OutputIterator,
          class BinaryFunction
          = ::rocprim::plus<typename std::iterator_traits<InputIterator>::value_type>,
-         class AccType = typename std::iterator_traits<InputIterator>::value_type>
+         class AccType
+         = ::rocprim::accumulator_t<BinaryFunction,
+                                    typename std::iterator_traits<InputIterator>::value_type>>
 inline hipError_t inclusive_scan(void*             temporary_storage,
                                  size_t&           storage_size,
                                  InputIterator     input,
@@ -534,9 +518,173 @@ inline hipError_t inclusive_scan(void*             temporary_storage,
                                  const hipStream_t stream            = 0,
                                  bool              debug_synchronous = false)
 {
+    // AccType may be const or a reference. Get the non-const, non-reference type.
+    // This is necessary because we may need to assign to instances of this type or create pointers to it.
+    using safe_acc_type = typename std::remove_const<typename std::remove_reference<AccType>::type>::type;
+
     // input_type() is a dummy initial value (not used)
     return detail::scan_impl<detail::lookback_scan_determinism::default_determinism,
                              false,
+                             false,
+                             Config,
+                             InputIterator,
+                             OutputIterator,
+                             safe_acc_type,
+                             BinaryFunction,
+                             safe_acc_type>(temporary_storage,
+                                            storage_size,
+                                            input,
+                                            output,
+                                            safe_acc_type{},
+                                            size,
+                                            scan_op,
+                                            stream,
+                                            debug_synchronous);
+}
+
+/// \brief Seeded parallel inclusive scan primitive for device level.
+///
+/// inclusive_scan function performs a device-wide inclusive prefix scan operation
+/// using binary \p scan_op operator.
+///
+/// \par Overview
+/// * Supports non-commutative scan operators. However, a scan operator should be
+/// associative.
+/// * When used with non-associative functions (e.g. floating point arithmetic operations):
+///    - the results may be non-deterministic and/or vary in precision,
+///    - and bit-wise reproducibility is not guaranteed, that is, results from multiple runs
+///      using the same input values on the same device may not be bit-wise identical.
+///   If deterministic behavior is required, Use \link deterministic_inclusive_scan()
+///   rocprim::deterministic_inclusive_scan \endlink instead.
+/// * Returns the required size of \p temporary_storage in \p storage_size
+/// if \p temporary_storage in a null pointer.
+/// * Ranges specified by \p input and \p output must have at least \p size elements.
+/// * By default, the input type is used for accumulation. A custom type
+/// can be specified using the \p AccType type parameter, see the example below.
+///
+/// \tparam Config [optional] Configuration of the primitive, must be `default_config` or `scan_config`.
+/// \tparam InputIterator random-access iterator type of the input range. Must meet the
+/// requirements of a C++ InputIterator concept. It can be a simple pointer type.
+/// \tparam OutputIterator random-access iterator type of the output range. Must meet the
+/// requirements of a C++ OutputIterator concept. It can be a simple pointer type.
+/// \tparam InitValueType type of the initial value.
+/// \tparam BinaryFunction type of binary function used for scan. Default type
+/// is \p rocprim::plus<T>, where \p T is a \p value_type of \p InputIterator.
+/// \tparam AccType accumulator type used to propagate the scanned values. Default type
+/// is value type of the input iterator.
+///
+/// \param [in] temporary_storage pointer to a device-accessible temporary storage. When
+/// a null pointer is passed, the required allocation size (in bytes) is written to
+/// \p storage_size and function returns without performing the scan operation.
+/// \param [in,out] storage_size reference to a size (in bytes) of \p temporary_storage.
+/// \param [in] input iterator to the first element in the range to scan.
+/// \param [out] output iterator to the first element in the output range. It can be
+/// same as \p input.
+/// \param [in] initial_value initial value to start the scan.
+/// A rocpim::future_value may be passed to use a value that will be later computed.
+/// \param [in] size number of element in the input range.
+/// \param [in] scan_op binary operation function object that will be used for scan.
+/// The signature of the function should be equivalent to the following:
+/// <tt>T f(const T &a, const T &b);</tt>. The signature does not need to have
+/// <tt>const &</tt>, but function object must not modify the objects passed to it.
+/// Default is BinaryFunction().
+/// \param [in] stream [optional] HIP stream object. Default is \p 0 (default stream).
+/// \param [in] debug_synchronous [optional] If true, synchronization after every kernel
+/// launch is forced in order to check for errors. Default value is \p false.
+///
+/// \returns \p hipSuccess (\p 0) after successful scan; otherwise a HIP runtime error of
+/// type \p hipError_t.
+///
+/// \par Example
+/// \parblock
+/// In this example a device-level inclusive sum operation is performed on an array of
+/// integer values (<tt>short</tt>s are scanned into <tt>int</tt>s).
+///
+/// \code{.cpp}
+/// #include <rocprim/rocprim.hpp>
+///
+/// // Prepare input and output (declare pointers, allocate device memory etc.)
+/// size_t input_size;    // e.g., 8
+/// short * input;        // e.g., [1, 2, 3, 4, 5, 6, 7, 8]
+/// int * output;         // empty array of 8 elements
+/// int initial_value;    // e.g. 10
+///
+/// size_t temporary_storage_size_bytes;
+/// void * temporary_storage_ptr = nullptr;
+/// // Get required size of the temporary storage
+/// rocprim::inclusive_scan(
+///     temporary_storage_ptr, temporary_storage_size_bytes,
+///     input, output, initial_value, input_size, rocprim::plus<int>()
+/// );
+///
+/// // allocate temporary storage
+/// hipMalloc(&temporary_storage_ptr, temporary_storage_size_bytes);
+///
+/// // perform scan
+/// rocprim::inclusive_scan(
+///     temporary_storage_ptr, temporary_storage_size_bytes,
+///     input, output, initial_value, input_size, rocprim::plus<int>()
+/// );
+/// // output: [11, 13, 16, 20, 25, 31, 38, 46]
+/// \endcode
+///
+/// The same example as above, but now a custom accumulator type is specified.
+///
+/// \code{.cpp}
+/// #include <rocprim/rocprim.hpp>
+///
+/// size_t input_size;
+/// short * input;
+/// int * output;
+/// int initial_value;
+///
+/// size_t temporary_storage_size_bytes;
+/// void * temporary_storage_ptr = nullptr;
+///
+/// rocprim::inclusive_scan(
+///     temporary_storage_ptr, temporary_storage_size_bytes,
+///     input, output, initial_value, input_size, rocprim::plus<int>()
+/// );
+///
+/// hipMalloc(&temporary_storage_ptr, temporary_storage_size_bytes);
+///
+/// // Use type parameter to set custom accumulator type
+/// rocprim::inclusive_scan<rocprim::default_config,
+///                         short*,
+///                         int*,
+///                         rocprim::plus<int>,
+///                         int>(temporary_storage_ptr,
+///                              temporary_storage_size_bytes,
+///                              input_iterator,
+///                              output,
+///                              initial_value
+///                              input_size,
+///                              rocprim::plus<int>());
+/// \endcode
+/// \endparblock
+template<class Config = default_config,
+         class InputIterator,
+         class OutputIterator,
+         class InitValueType,
+         class BinaryFunction
+         = ::rocprim::plus<typename std::iterator_traits<InputIterator>::value_type>,
+         class AccType
+         = ::rocprim::accumulator_t<BinaryFunction,
+                                    typename std::iterator_traits<InputIterator>::value_type>>
+inline hipError_t inclusive_scan(void*               temporary_storage,
+                                 size_t&             storage_size,
+                                 InputIterator       input,
+                                 OutputIterator      output,
+                                 const InitValueType initial_value,
+                                 const size_t        size,
+                                 BinaryFunction      scan_op           = BinaryFunction(),
+                                 const hipStream_t   stream            = 0,
+                                 bool                debug_synchronous = false)
+{
+    // input_type() is a dummy initial value (not used)
+    return detail::scan_impl<detail::lookback_scan_determinism::default_determinism,
+                             false,
+                             true,
                              Config,
                              InputIterator,
                              OutputIterator,
@@ -546,7 +694,7 @@ inline hipError_t inclusive_scan(void*             temporary_storage,
                                       storage_size,
                                       input,
                                       output,
-                                      AccType{},
+                                      initial_value,
                                       size,
                                       scan_op,
                                       stream,
@@ -565,7 +713,9 @@ template<class Config = default_config,
          class OutputIterator,
          class BinaryFunction
          = ::rocprim::plus<typename std::iterator_traits<InputIterator>::value_type>,
-         class AccType = typename std::iterator_traits<InputIterator>::value_type>
+         class AccType
+         = ::rocprim::accumulator_t<BinaryFunction,
+                                    typename std::iterator_traits<InputIterator>::value_type>>
 inline hipError_t deterministic_inclusive_scan(void*             temporary_storage,
                                                size_t&           storage_size,
                                                InputIterator     input,
@@ -575,8 +725,58 @@ inline hipError_t deterministic_inclusive_scan(void*             temporary_stora
                                                const hipStream_t stream  = 0,
                                                bool              debug_synchronous = false)
 {
+    // AccType may be const or a reference. Get the non-const, non-reference type.
+    // This is necessary because we may need to assign to instances of this type or create pointers to it.
+    using safe_acc_type = typename std::remove_const<typename std::remove_reference<AccType>::type>::type;
+
     return detail::scan_impl<detail::lookback_scan_determinism::deterministic,
                              false,
+                             false,
+                             Config,
+                             InputIterator,
+                             OutputIterator,
+                             safe_acc_type,
+                             BinaryFunction,
+                             safe_acc_type>(temporary_storage,
+                                            storage_size,
+                                            input,
+                                            output,
+                                            safe_acc_type{},
+                                            size,
+                                            scan_op,
+                                            stream,
+                                            debug_synchronous);
+}
+
+/// \brief Seeded bitwise-reproducible parallel inclusive scan primitive for device level.
+///
+/// This function behaves the same as <tt>inclusive_scan()</tt>, except that unlike
+/// <tt>inclusive_scan()</tt>, it provides run-to-run deterministic behavior for
+/// non-associative scan operators like floating point arithmetic operations.
+/// Refer to the documentation for \link inclusive_scan() rocprim::inclusive_scan \endlink
+/// for a detailed description of this function.
+template<class Config = default_config,
+         class InputIterator,
+         class OutputIterator,
+         class InitValueType,
+         class BinaryFunction
+         = ::rocprim::plus<typename std::iterator_traits<InputIterator>::value_type>,
+         class AccType
+         = ::rocprim::accumulator_t<BinaryFunction,
+                                    typename std::iterator_traits<InputIterator>::value_type>>
+inline hipError_t deterministic_inclusive_scan(void*             temporary_storage,
+                                               size_t&           storage_size,
+                                               InputIterator     input,
+                                               OutputIterator    output,
+                                               InitValueType     initial_value,
+                                               const size_t      size,
+                                               BinaryFunction    scan_op = BinaryFunction(),
+                                               const hipStream_t stream  = 0,
+                                               bool              debug_synchronous = false)
+{
+    return detail::scan_impl<detail::lookback_scan_determinism::deterministic,
+                             false,
+                             true,
                              Config,
                              InputIterator,
                              OutputIterator,
@@ -586,7 +786,7 @@ inline hipError_t deterministic_inclusive_scan(void*             temporary_stora
                                       storage_size,
                                       input,
                                       output,
-                                      AccType{},
+                                      initial_value,
                                       size,
                                       scan_op,
                                       stream,
@@ -611,34 +811,35 @@ inline hipError_t deterministic_inclusive_scan(void*             temporary_stora
 /// if \p temporary_storage in a null pointer.
 /// * Ranges specified by \p input and \p output must have at least \p size elements.
 ///
-/// \tparam Config - [optional] Configuration of the primitive, must be `default_config` or `scan_config`.
-/// \tparam InputIterator - random-access iterator type of the input range. Must meet the
+/// \tparam Config [optional] Configuration of the primitive, must be `default_config` or `scan_config`.
+/// \tparam InputIterator random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam OutputIterator - random-access iterator type of the output range. Must meet the
+/// \tparam OutputIterator random-access iterator type of the output range. Must meet the
 /// requirements of a C++ OutputIterator concept. It can be a simple pointer type.
-/// \tparam InitValueType - type of the initial value.
-/// \tparam BinaryFunction - type of binary function used for scan. Default type
+/// \tparam InitValueType type of the initial value.
+/// \tparam BinaryFunction type of binary function used for scan. Default type
 /// is \p rocprim::plus<T>, where \p T is a \p value_type of \p InputIterator.
-/// \tparam AccType - accumulator type used to propagate the scanned values. Default type
-/// is 'InitValueType', unless it's 'rocprim::future_value'. Then it will be the wrapped input type.
+/// \tparam AccType accumulator type used to propagate the scanned values. The default is the type that
+/// is returned by a function of type BinaryFunction when it is passed a value of type \p InitValueType,
+/// unless it's 'rocprim::future_value'. Then it will be the wrapped input type.
 ///
-/// \param [in] temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param [in] temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// \p storage_size and function returns without performing the scan operation.
-/// \param [in,out] storage_size - reference to a size (in bytes) of \p temporary_storage.
-/// \param [in] input - iterator to the first element in the range to scan.
-/// \param [out] output - iterator to the first element in the output range. It can be
+/// \param [in,out] storage_size reference to a size (in bytes) of \p temporary_storage.
+/// \param [in] input iterator to the first element in the range to scan.
+/// \param [out] output iterator to the first element in the output range. It can be
 /// same as \p input.
-/// \param [in] initial_value - initial value to start the scan.
+/// \param [in] initial_value initial value to start the scan.
 /// A rocpim::future_value may be passed to use a value that will be later computed.
-/// \param [in] size - number of element in the input range.
-/// \param [in] scan_op - binary operation function object that will be used for scan.
+/// \param [in] size number of element in the input range.
+/// \param [in] scan_op binary operation function object that will be used for scan.
 /// The signature of the function should be equivalent to the following:
 /// <tt>T f(const T &a, const T &b);</tt>. The signature does not need to have
 /// <tt>const &</tt>, but function object must not modify the objects passed to it.
 /// The default value is \p BinaryFunction().
-/// \param [in] stream - [optional] HIP stream object. The default is \p 0 (default stream).
-/// \param [in] debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param [in] stream [optional] HIP stream object. The default is \p 0 (default stream).
+/// \param [in] debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors. The default value is \p false.
 ///
 /// \returns \p hipSuccess (\p 0) after successful scan; otherwise a HIP runtime error of
@@ -690,7 +891,8 @@ template<class Config = default_config,
          class InitValueType,
          class BinaryFunction
          = ::rocprim::plus<typename std::iterator_traits<InputIterator>::value_type>,
-         class AccType = detail::input_type_t<InitValueType>>
+         class AccType
+         = ::rocprim::accumulator_t<BinaryFunction, rocprim::detail::input_type_t<InitValueType>>>
 inline hipError_t exclusive_scan(void*               temporary_storage,
                                  size_t&             storage_size,
                                  InputIterator       input,
@@ -701,22 +903,27 @@ inline hipError_t exclusive_scan(void*               temporary_storage,
                                  const hipStream_t   stream            = 0,
                                  bool                debug_synchronous = false)
 {
+    // AccType may be const or a reference. Get the non-const, non-reference type.
+    // This is necessary because we may need to assign to instances of this type or create pointers to it.
+    using safe_acc_type = typename std::remove_const<typename std::remove_reference<AccType>::type>::type;
+
     return detail::scan_impl<detail::lookback_scan_determinism::default_determinism,
+                             true,
                              true,
                              Config,
                              InputIterator,
                              OutputIterator,
                              InitValueType,
                              BinaryFunction,
-                             AccType>(temporary_storage,
-                                      storage_size,
-                                      input,
-                                      output,
-                                      initial_value,
-                                      size,
-                                      scan_op,
-                                      stream,
-                                      debug_synchronous);
+                             safe_acc_type>(temporary_storage,
+                                            storage_size,
+                                            input,
+                                            output,
+                                            initial_value,
+                                            size,
+                                            scan_op,
+                                            stream,
+                                            debug_synchronous);
 }
 
 /// \brief Bitwise-reproducible parallel exclusive scan primitive for device level.
@@ -732,7 +939,8 @@ template<class Config = default_config,
          class InitValueType,
          class BinaryFunction
          = ::rocprim::plus<typename std::iterator_traits<InputIterator>::value_type>,
-         class AccType = detail::input_type_t<InitValueType>>
+         class AccType
+         = ::rocprim::accumulator_t<BinaryFunction, rocprim::detail::input_type_t<InitValueType>>>
 inline hipError_t deterministic_exclusive_scan(void*               temporary_storage,
                                                size_t&             storage_size,
                                                InputIterator       input,
@@ -743,22 +951,27 @@ inline hipError_t deterministic_exclusive_scan(void*               temporary_sto
                                                const hipStream_t   stream  = 0,
                                                bool                debug_synchronous = false)
 {
+    // AccType may be const or a reference. Get the non-const, non-reference type.
+    // This is necessary because we may need to assign to instances of this type or create pointers to it.
+    using safe_acc_type = typename std::remove_const<typename std::remove_reference<AccType>::type>::type;
+
     return detail::scan_impl<detail::lookback_scan_determinism::deterministic,
+                             true,
                              true,
                              Config,
                              InputIterator,
                              OutputIterator,
                              InitValueType,
                              BinaryFunction,
-                             AccType>(temporary_storage,
-                                      storage_size,
-                                      input,
-                                      output,
-                                      initial_value,
-                                      size,
-                                      scan_op,
-                                      stream,
-                                      debug_synchronous);
+                             safe_acc_type>(temporary_storage,
+                                            storage_size,
+                                            input,
+                                            output,
+                                            initial_value,
+                                            size,
+                                            scan_op,
+                                            stream,
+                                            debug_synchronous);
 }
 
 /// @}
