@@ -1,4 +1,4 @@
-// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -25,6 +25,7 @@
 
 #include "../config.hpp"
 #include "../detail/various.hpp"
+#include "../intrinsics/arch.hpp"
 
 BEGIN_ROCPRIM_NAMESPACE
 
@@ -33,28 +34,6 @@ BEGIN_ROCPRIM_NAMESPACE
 
 // Sizes
 
-/// \brief [DEPRECATED] Returns a number of threads in a hardware warp.
-///
-/// It is constant for a device.
-/// This function is not supported for gfx1030 and newer architectures and will be removed in a future release.
-/// Please use the new host_warp_size() and device_warp_size() functions.
-[[deprecated("Use host_warp_size() or device_warp_size() "
-             "instead.")]] ROCPRIM_HOST_DEVICE inline constexpr unsigned int
-    warp_size()
-{
-    return ROCPRIM_WAVEFRONT_SIZE;
-}
-
-/// \brief Returns a number of threads in a hardware warp for the actual target.
-/// At device side this constant is available at compile time.
-///
-/// It is constant for a device.
-ROCPRIM_DEVICE ROCPRIM_INLINE
-constexpr unsigned int device_warp_size()
-{
-    return ROCPRIM_WAVEFRONT_SIZE;
-}
-
 /// \brief Returns flat size of a multidimensional block (tile).
 ROCPRIM_DEVICE ROCPRIM_INLINE
 unsigned int flat_block_size()
@@ -75,12 +54,7 @@ unsigned int flat_tile_size()
 ROCPRIM_DEVICE ROCPRIM_INLINE
 unsigned int lane_id()
 {
-#ifndef __HIP_CPU_RT__
     return ::__lane_id();
-#else
-    using namespace hip::detail;
-    return id(Fiber::this_fiber()) % device_warp_size();
-#endif
 }
 
 /// \brief Returns flat (linear, 1D) thread identifier in a multidimensional block (tile).
@@ -133,15 +107,15 @@ unsigned int flat_tile_thread_id()
 ROCPRIM_DEVICE ROCPRIM_INLINE
 unsigned int warp_id()
 {
-    return flat_block_thread_id()/device_warp_size();
+    return flat_block_thread_id() / arch::wavefront::size();
 }
 
 /// \brief Returns warp id in a block (tile), given the flat (linear, 1D) thread identifier in a multidimensional tile (block).
-/// \param flat_id - the flat id that should be used to compute the warp id.
+/// \param flat_id the flat id that should be used to compute the warp id.
 ROCPRIM_DEVICE ROCPRIM_INLINE
 unsigned int warp_id(unsigned int flat_id)
 {
-    return flat_id/device_warp_size();
+    return flat_id / arch::wavefront::size();
 }
 
 #ifndef DOXYGEN_SHOULD_SKIP_THIS
@@ -151,7 +125,7 @@ template<unsigned int BlockSizeX, unsigned int BlockSizeY, unsigned int BlockSiz
 ROCPRIM_DEVICE ROCPRIM_INLINE
 unsigned int warp_id()
 {
-    return flat_block_thread_id<BlockSizeX, BlockSizeY, BlockSizeZ>()/device_warp_size();
+    return flat_block_thread_id<BlockSizeX, BlockSizeY, BlockSizeZ>() / arch::wavefront::size();
 }
 #endif // DOXYGEN_SHOULD_SKIP_THIS
 
@@ -304,13 +278,6 @@ namespace detail
         return lane_id()%LogicalWarpSize;
     }
 
-    template<>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    unsigned int logical_lane_id<device_warp_size()>()
-    {
-        return lane_id();
-    }
-
     // Return id of "logical warp" in a block
     template<unsigned int LogicalWarpSize>
     ROCPRIM_DEVICE ROCPRIM_INLINE
@@ -319,13 +286,6 @@ namespace detail
         return flat_block_thread_id()/LogicalWarpSize;
     }
 
-    template<>
-    ROCPRIM_DEVICE ROCPRIM_INLINE
-    unsigned int logical_warp_id<device_warp_size()>()
-    {
-        return warp_id();
-    }
-
     ROCPRIM_DEVICE ROCPRIM_INLINE
     void memory_fence_system()
     {
