@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 
-# Copyright (c) 2022-2024 Advanced Micro Devices, Inc. All rights reserved.
+# Copyright (c) 2022-2025 Advanced Micro Devices, Inc. All rights reserved.
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -41,7 +41,7 @@ from collections import defaultdict
 from typing import Dict, List, Callable, Optional, Tuple
 from jinja2 import Environment, PackageLoader, select_autoescape
 
-TARGET_ARCHITECTURES = ['gfx803', 'gfx900', 'gfx906', 'gfx908', 'gfx90a', 'gfx1030', 'gfx1100', 'gfx1102']
+TARGET_ARCHITECTURES = ['gfx803', 'gfx900', 'gfx906', 'gfx908', 'gfx90a', 'gfx942', 'gfx1030', 'gfx1100', 'gfx1102']
 # C++ typename used for optional types
 EMPTY_TYPENAME = "empty_type"
 
@@ -295,10 +295,10 @@ class BenchmarksOfArchitecture:
             empty_fallback = FallbackCase(None, EMPTY_TYPENAME, 0, 0, False)
 
             # If a type is optional, also generate the fallbacks where the type is empty.
-            fallback_entries_0: List[FallbackCase] = self.fallback_entries
+            fallback_entries_0: List[FallbackCase] = self.fallback_entries.copy()
             if config_selection_types[0].is_optional:
                 fallback_entries_0.append(empty_fallback)
-            fallback_entries_1: List[FallbackCase] = self.fallback_entries
+            fallback_entries_1: List[FallbackCase] = self.fallback_entries.copy()
             if config_selection_types[1].is_optional:
                 fallback_entries_1.append(empty_fallback)
 
@@ -461,6 +461,14 @@ class AlgorithmDeviceReduce(Algorithm):
     def __init__(self, fallback_entries):
         Algorithm.__init__(self, fallback_entries)
 
+class AlgorithmDeviceSegmentedReduce(Algorithm):
+    algorithm_name = "device_segmented_reduce"
+    config_selection_params = [
+        SelectionType(name="key_type", is_optional=False, select_on_size_only=False)]
+    cpp_configuration_template_name = "segmented_reduce_config_template"
+    def __init__(self, fallback_entries):
+        Algorithm.__init__(self, fallback_entries)
+
 class AlgorithmDeviceScan(Algorithm):
     algorithm_name = "device_scan"
     cpp_configuration_template_name = "scan_config_template"
@@ -523,6 +531,14 @@ class AlgorithmDeviceAdjacentDifferenceInplace(Algorithm):
     def __init__(self, fallback_entries):
         Algorithm.__init__(self, fallback_entries)
 
+class AlgorithmDeviceAdjacentFind(Algorithm):
+    algorithm_name = "device_adjacent_find"
+    cpp_configuration_template_name = "adjacent_find_config_template"
+    config_selection_params = [
+        SelectionType(name="input_type", is_optional=False, select_on_size_only=False)]
+    def __init__(self, fallback_entries):
+        Algorithm.__init__(self, fallback_entries)
+
 class AlgorithmDeviceSegmentedRadixSort(Algorithm):
     algorithm_name = "device_segmented_radix_sort"
     cpp_configuration_template_name = "segmented_radix_sort_config_template"
@@ -540,6 +556,14 @@ class AlgorithmDeviceTransform(Algorithm):
     def __init__(self, fallback_entries):
         Algorithm.__init__(self, fallback_entries)
 
+class AlgorithmDeviceTransformPointer(Algorithm):
+    algorithm_name = "device_transform_pointer"
+    cpp_configuration_template_name = "transform_pointer_config_template"
+    config_selection_params = [
+        SelectionType(name="value_type", is_optional=False, select_on_size_only=False)]
+    def __init__(self, fallback_entries):
+        Algorithm.__init__(self, fallback_entries)
+
 class AlgorithmDevicePartitionTwoWayPredicate(Algorithm):
     algorithm_name = "device_partition_two_way_predicate"
     cpp_configuration_template_name = "partition_two_way_predicate_config_template"
@@ -580,6 +604,14 @@ class AlgorithmDevicePartitionThreeWay(Algorithm):
     def __init__(self, fallback_entries):
         Algorithm.__init__(self, fallback_entries)
 
+class AlgorithmDeviceSearchN(Algorithm):
+    algorithm_name = "device_search_n"
+    cpp_configuration_template_name = "search_n_config_template"
+    config_selection_params = [
+        SelectionType(name="data_type", is_optional=False, select_on_size_only=False)]
+    def __init__(self, fallback_entries):
+        Algorithm.__init__(self, fallback_entries)
+
 class AlgorithmDeviceSelectFlag(Algorithm):
     algorithm_name = "device_select_flag"
     cpp_configuration_template_name = "select_flag_config_template"
@@ -596,6 +628,15 @@ class AlgorithmDeviceSelectPredicate(Algorithm):
     def __init__(self, fallback_entries):
         Algorithm.__init__(self, fallback_entries)
 
+class AlgorithmDeviceSelectPredicatedFlag(Algorithm):
+    algorithm_name = "device_select_predicated_flag"
+    cpp_configuration_template_name = "select_predicated_flag_config_template"
+    config_selection_params = [
+        SelectionType(name="data_type", is_optional=False, select_on_size_only=False),
+        SelectionType(name="flag_type", is_optional=False, select_on_size_only=True)]
+    def __init__(self, fallback_entries):
+        Algorithm.__init__(self, fallback_entries)
+
 class AlgorithmDeviceSelectUnique(Algorithm):
     algorithm_name = "device_select_unique"
     cpp_configuration_template_name = "select_unique_config_template"
@@ -622,6 +663,35 @@ class AlgorithmDeviceReduceByKey(Algorithm):
     def __init__(self, fallback_entries):
         Algorithm.__init__(self, fallback_entries)
 
+class AlgorithmDeviceFindFirstOf(Algorithm):
+    algorithm_name = "device_find_first_of"
+    cpp_configuration_template_name = "find_first_of_config_template"
+    config_selection_params = [
+        SelectionType(name="value_type", is_optional=False, select_on_size_only=True)]
+    def __init__(self, fallback_entries):
+        Algorithm.__init__(self, fallback_entries)
+
+class AlgorithmDeviceRunLengthEncode(Algorithm):
+    algorithm_name = 'device_run_length_encode'
+    cpp_configuration_template_name = 'run_length_encode_config_template'
+    config_selection_params = [SelectionType(name='key_type', is_optional=False, select_on_size_only=False)]
+    def __init__(self, fallback_entries):
+        Algorithm.__init__(self, fallback_entries)
+
+class AlgorithmDeviceRunLengthEncodeNonTrivial(Algorithm):
+    algorithm_name = 'device_run_length_encode_non_trivial'
+    cpp_configuration_template_name = 'run_length_encode_non_trivial_runs_config_template'
+    config_selection_params = [SelectionType(name='key_type', is_optional=False, select_on_size_only=False)]
+    def __init__(self, fallback_entries):
+        Algorithm.__init__(self, fallback_entries)
+
+class AlgorithmDeviceMerge(Algorithm):
+    algorithm_name = "device_merge"
+    cpp_configuration_template_name = "merge_config_template"
+    config_selection_params = [
+        SelectionType(name="key_type", is_optional=False, select_on_size_only=False),
+        SelectionType(name="value_type", is_optional=True, select_on_size_only=True)]
+
 def filt_algo_regex(e: FallbackCase, algorithm_name):
     if e.algo_regex:
         return re.match(e.algo_regex, algorithm_name) is not None
@@ -641,6 +711,8 @@ def create_algorithm(algorithm_name: str, fallback_entries: List[FallbackCase]):
         return AlgorithmDeviceRadixSortOnesweep(fallback_entries)
     elif algorithm_name == 'device_reduce':
         return AlgorithmDeviceReduce(fallback_entries)
+    elif algorithm_name == 'device_segmented_reduce':
+        return AlgorithmDeviceSegmentedReduce(fallback_entries)
     elif algorithm_name == 'device_scan':
         return AlgorithmDeviceScan(fallback_entries)
     elif algorithm_name == 'device_scan_by_key':
@@ -655,10 +727,14 @@ def create_algorithm(algorithm_name: str, fallback_entries: List[FallbackCase]):
         return AlgorithmDeviceAdjacentDifference(fallback_entries)
     elif algorithm_name == 'device_adjacent_difference_inplace':
         return AlgorithmDeviceAdjacentDifferenceInplace(fallback_entries)
+    elif algorithm_name == 'device_adjacent_find':
+        return AlgorithmDeviceAdjacentFind(fallback_entries)
     elif algorithm_name == 'device_segmented_radix_sort':
         return AlgorithmDeviceSegmentedRadixSort(fallback_entries)
     elif algorithm_name == 'device_transform':
         return AlgorithmDeviceTransform(fallback_entries)
+    elif algorithm_name == 'device_transform_pointer':
+        return AlgorithmDeviceTransformPointer(fallback_entries)
     elif algorithm_name == 'device_partition_two_way_predicate':
         return AlgorithmDevicePartitionTwoWayPredicate(fallback_entries)
     elif algorithm_name == 'device_partition_two_way_flag':
@@ -669,16 +745,28 @@ def create_algorithm(algorithm_name: str, fallback_entries: List[FallbackCase]):
         return AlgorithmDevicePartitionPredicate(fallback_entries)
     elif algorithm_name == 'device_partition_three_way':
         return AlgorithmDevicePartitionThreeWay(fallback_entries)
+    elif algorithm_name == 'device_search_n':
+        return AlgorithmDeviceSearchN(fallback_entries)
     elif algorithm_name == 'device_select_flag':
         return AlgorithmDeviceSelectFlag(fallback_entries)
     elif algorithm_name == 'device_select_predicate':
         return AlgorithmDeviceSelectPredicate(fallback_entries)
+    elif algorithm_name == 'device_select_predicated_flag':
+        return AlgorithmDeviceSelectPredicatedFlag(fallback_entries)
     elif algorithm_name == 'device_select_unique':
         return AlgorithmDeviceSelectUnique(fallback_entries)
     elif algorithm_name == 'device_select_unique_by_key':
         return AlgorithmDeviceSelectUniqueByKey(fallback_entries)
     elif algorithm_name == 'device_reduce_by_key':
         return AlgorithmDeviceReduceByKey(fallback_entries)
+    elif algorithm_name == 'device_find_first_of':
+        return AlgorithmDeviceFindFirstOf(fallback_entries)
+    elif algorithm_name == 'device_run_length_encode':
+        return AlgorithmDeviceRunLengthEncode(fallback_entries)
+    elif algorithm_name == 'device_run_length_encode_non_trivial':
+        return AlgorithmDeviceRunLengthEncodeNonTrivial(fallback_entries)
+    elif algorithm_name == 'device_merge':
+        return AlgorithmDeviceMerge(fallback_entries)
     else:
         raise(NotSupportedError(f'Algorithm "{algorithm_name}" is not supported (yet)'))
 
@@ -796,6 +884,7 @@ def main():
     parser.add_argument("-p", "--out_basedir", type=str, help="Base dir for the output files, for each algorithm a new file will be created in this directory", required=True)
     parser.add_argument("-c", "--fallback_configuration", type=argparse.FileType('r'), default=os.path.join(current_dir, "fallback_config.json"), help="Configuration for fallbacks for not tested datatypes")
     args = parser.parse_args()
+    #import pdb; pdb.set_trace()
 
     benchmark_manager = BenchmarkDataManager(args.fallback_configuration)
 
