@@ -1,6 +1,6 @@
 // MIT License
 //
-// Copyright (c) 2020-2024 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2020-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -20,8 +20,8 @@
 // OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 // SOFTWARE.
 
-#ifndef ROCPRIM_COMMON_TEST_HEADER
-#define ROCPRIM_COMMON_TEST_HEADER
+#ifndef ROCPRIM_COMMON_TEST_HEADER_HPP_
+#define ROCPRIM_COMMON_TEST_HEADER_HPP_
 
 #include <algorithm>
 #include <cmath>
@@ -36,36 +36,34 @@
 #include <utility>
 #include <vector>
 
+#include "../common/utils.hpp"
+
 // Google Test
 #include <gtest/gtest.h>
 
 // HIP API
 #include <hip/hip_runtime.h>
 #include <hip/hip_vector_types.h>
-#ifndef __HIP_CPU_RT__
-#include <hip/hip_ext.h>
-#endif
 
-// GoogleTest-compatible HIP_CHECK macro. FAIL is called to log the Google Test trace.
-// The lambda is invoked immediately as assertions that generate a fatal failure can
-// only be used in void-returning functions.
-#ifndef HIP_CHECK
-    #define HIP_CHECK(condition)                                                            \
+#define HIP_CHECK_MEMORY(condition)                                                         \
+    {                                                                                       \
+        hipError_t error = condition;                                                       \
+        if(error == hipErrorOutOfMemory)                                                    \
         {                                                                                   \
-            hipError_t error = condition;                                                   \
-            if(error != hipSuccess)                                                         \
-            {                                                                               \
-                [error]()                                                                   \
-                { FAIL() << "HIP error " << error << ": " << hipGetErrorString(error); }(); \
-                exit(error);                                                                \
-            }                                                                               \
-        }
-#endif
+            (void) hipGetLastError();                                                       \
+            std::cout << "Out of memory. Skipping size = " << size << std::endl;            \
+            break;                                                                          \
+        }                                                                                   \
+        if(error != hipSuccess)                                                             \
+        {                                                                                   \
+            std::cout << "HIP error: " << hipGetErrorString(error) << " line: " << __LINE__ \
+                      << std::endl;                                                         \
+            exit(error);                                                                    \
+        }                                                                                   \
+    }
 
-#if(defined(__GNUC__) || defined(__clang__)) && (defined(__GLIBCXX__) || defined(_LIBCPP_VERSION))
+#ifndef ROCPRIM_HAS_INT128_SUPPORT
     #define ROCPRIM_HAS_INT128_SUPPORT 1
-#else
-    #define ROCPRIM_HAS_INT128_SUPPORT 0
 #endif
 
 #define INSTANTIATE_TYPED_TEST_EXPANDED_1(line, test_suite_name, ...)         \
@@ -84,9 +82,6 @@
 #define INSTANTIATE_TYPED_TEST(test_suite_name, ...) \
     INSTANTIATE_TYPED_TEST_EXPANDED(__LINE__, test_suite_name, __VA_ARGS__)
 
-// C++17 or newer
-#define CPP17 __cplusplus >= 201703L
-
 #include <cstdlib>
 #include <string>
 #include <cctype>
@@ -94,37 +89,10 @@
 namespace test_common_utils
 {
 
-inline char* __get_env(const char* name)
-{
-    char* env;
-#ifdef _MSC_VER
-    size_t  len;
-    errno_t err = _dupenv_s(&env, &len, name);
-    if(err)
-    {
-        return nullptr;
-    }
-#else
-    env = std::getenv(name);
-#endif
-    return env;
-}
-
-inline void clean_env(char* name)
-{
-#ifdef _MSC_VER
-    if(name != nullptr)
-    {
-        free(name);
-    }
-#endif
-    (void)name;
-}
-
 inline int obtain_device_from_ctest()
 {
     static const std::string rg0    = "CTEST_RESOURCE_GROUP_0";
-    char*                    env    = __get_env(rg0.c_str());
+    char*                    env    = common::__get_env(rg0.c_str());
     int                      device = 0;
     if(env != nullptr)
     {
@@ -135,40 +103,15 @@ inline int obtain_device_from_ctest()
             amdgpu_target.begin(),
             // Feeding std::toupper plainly results in implicitly truncating conversions between int and char triggering warnings.
             [](unsigned char c) { return static_cast<char>(std::toupper(c)); });
-        char*       env_reqs = __get_env((rg0 + "_" + amdgpu_target).c_str());
+        char*       env_reqs = common::__get_env((rg0 + "_" + amdgpu_target).c_str());
         std::string reqs(env_reqs);
         device = std::atoi(
             reqs.substr(reqs.find(':') + 1, reqs.find(',') - (reqs.find(':') + 1)).c_str());
-        clean_env(env_reqs);
+        common::clean_env(env_reqs);
     }
-    clean_env(env);
+    common::clean_env(env);
     return device;
 }
-
-inline bool use_hmm()
-{
-
-    char*      env = __get_env("ROCPRIM_USE_HMM");
-    const bool hmm = (env != nullptr) && (strcmp(env, "1") == 0);
-    clean_env(env);
-    return hmm;
 }
 
-// Helper for HMM allocations: HMM is requested through ROCPRIM_USE_HMM=1 environment variable
-template <class T>
-hipError_t hipMallocHelper(T** devPtr, size_t size)
-{
-    if (use_hmm())
-    {
-        return hipMallocManaged(reinterpret_cast<void**>(devPtr), size);
-    }
-    else
-    {
-        return hipMalloc(reinterpret_cast<void**>(devPtr), size);
-    }
-    return hipSuccess;
-}
-
-}
-
-#endif
+#endif // ROCPRIM_COMMON_TEST_HEADER_HPP_
