@@ -1,6 +1,6 @@
 // MIT License
 //
-// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -22,21 +22,31 @@
 
 // required test headers
 #include "../common_test_header.hpp"
-#include "test_utils_types.hpp"
+
+#include "../../common/utils.hpp"
+#include "../../common/utils_custom_type.hpp"
+#include "../../common/utils_device_ptr.hpp"
+
+#include "bounds_checking_iterator.hpp"
+#include "test_utils.hpp"
+#include "test_utils_assertions.hpp"
+#include "test_utils_data_generation.hpp"
+#include "test_utils_hipgraphs.hpp"
 
 // required rocprim headers
+#include <rocprim/device/config_types.hpp>
 #include <rocprim/device/device_merge.hpp>
 #include <rocprim/functional.hpp>
 #include <rocprim/iterator/counting_iterator.hpp>
 #include <rocprim/iterator/transform_iterator.hpp>
-
-#include <gtest/gtest.h>
-
-#include <hip/hip_runtime.h>
+#include <rocprim/types.hpp>
 
 #include <algorithm>
-#include <fstream>
+#include <cstddef>
 #include <numeric>
+#include <stdint.h>
+#include <tuple>
+#include <utility>
 #include <vector>
 
 using DefaultConfig = rocprim::default_config;
@@ -68,11 +78,14 @@ public:
     static constexpr bool use_graphs = Params::use_graphs;
 };
 
-using custom_int2 = test_utils::custom_test_type<int>;
-using custom_double2 = test_utils::custom_test_type<double>;
+using custom_int2    = common::custom_type<int, int, true>;
+using custom_double2 = common::custom_type<double, double, true>;
+using custom_large   = common::custom_huge_type<1024, long long>;
 
-typedef ::testing::Types<
+using RocprimDeviceMergeTestsParams = ::testing::Types<
     DeviceMergeParams<int, double>,
+    DeviceMergeParams<rocprim::bfloat16, half>,
+    DeviceMergeParams<short, half>,
     DeviceMergeParams<unsigned long, unsigned int, rocprim::greater<unsigned long>>,
     DeviceMergeParams<float, custom_double2>,
     DeviceMergeParams<int, float>,
@@ -83,8 +96,8 @@ typedef ::testing::Types<
     DeviceMergeParams<rocprim::bfloat16, rocprim::bfloat16, rocprim::less<rocprim::bfloat16>>,
     DeviceMergeParams<custom_double2, custom_int2, rocprim::greater<custom_double2>>,
     DeviceMergeParams<custom_int2, char>,
-    DeviceMergeParams<int, int, ::rocprim::less<int>, true>>
-    RocprimDeviceMergeTestsParams;
+    DeviceMergeParams<int, int, ::rocprim::less<int>, true>,
+    DeviceMergeParams<custom_large, custom_large>>;
 
 // size1, size2
 std::vector<std::tuple<size_t, size_t>> get_sizes()
@@ -129,11 +142,19 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKey)
 
     for(auto sizes : get_sizes())
     {
-        if ((std::get<0>(sizes) == 0 || std::get<1>(sizes) == 0) && test_common_utils::use_hmm())
+        if((std::get<0>(sizes) == 0 || std::get<1>(sizes) == 0) && common::use_hmm())
         {
             // hipMallocManaged() currently doesnt support zero byte allocation
             continue;
         }
+
+        if((std::get<0>(sizes) + std::get<1>(sizes) >= 100000
+            && sizeof(key_type) > sizeof(size_t) * 16))
+        {
+            // Huge types are slow
+            continue;
+        }
+
         SCOPED_TRACE(
             testing::Message() << "with sizes = {" <<
             std::get<0>(sizes) << ", " << std::get<1>(sizes) << "}"
@@ -145,14 +166,16 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKey)
         // compare function
         compare_op_type compare_op;
 
-        for (size_t seed_index = 0; seed_index < random_seeds_count + seed_size; seed_index++)
+        for(size_t seed_index = 0; seed_index < number_of_runs; seed_index++)
         {
             unsigned int seed_value = seed_index < random_seeds_count  ? rand() : seeds[seed_index - random_seeds_count];
             SCOPED_TRACE(testing::Message() << "with seed = " << seed_value);
 
             // Generate data
-            std::vector<key_type> keys_input1 = test_utils::get_random_data<key_type>(size1, 0, size1, seed_value);
-            std::vector<key_type> keys_input2 = test_utils::get_random_data<key_type>(size2, 0, size2, seed_value);
+            std::vector<key_type> keys_input1
+                = test_utils::get_random_data_wrapped<key_type>(size1, 0, size1, seed_value);
+            std::vector<key_type> keys_input2
+                = test_utils::get_random_data_wrapped<key_type>(size2, 0, size2, seed_value);
             std::sort(keys_input1.begin(), keys_input1.end(), compare_op);
             std::sort(keys_input2.begin(), keys_input2.end(), compare_op);
             std::vector<key_type> keys_output(size1 + size2, (key_type)0);
@@ -170,40 +193,33 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKey)
 
             test_utils::out_of_bounds_flag out_of_bounds;
 
-            key_type * d_keys_input1;
-            key_type * d_keys_input2;
-            key_type * d_keys_output;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_keys_input1, keys_input1.size() * sizeof(key_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_keys_input2, keys_input2.size() * sizeof(key_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_keys_output, keys_output.size() * sizeof(key_type)));
-            HIP_CHECK(
-                hipMemcpy(
-                    d_keys_input1, keys_input1.data(),
-                    keys_input1.size() * sizeof(key_type),
-                    hipMemcpyHostToDevice
-                )
-            );
-            HIP_CHECK(
-                hipMemcpy(
-                    d_keys_input2, keys_input2.data(),
-                    keys_input2.size() * sizeof(key_type),
-                    hipMemcpyHostToDevice
-                )
-            );
+            common::device_ptr<key_type> d_keys_input1;
+            common::device_ptr<key_type> d_keys_input2;
+            common::device_ptr<key_type> d_keys_output;
+
+            if(!d_keys_input1.resize_with_memory_check(keys_input1.size())
+               || !d_keys_input2.resize_with_memory_check(keys_input2.size())
+               || !d_keys_output.resize_with_memory_check(keys_output.size()))
+            {
+                std::cout << "Out of memory. Skipping test with sizes = {" << size1 << ", " << size2
+                          << "}" << std::endl;
+                break;
+            }
+
+            d_keys_input1.store(keys_input1);
+            d_keys_input2.store(keys_input2);
 
             test_utils::bounds_checking_iterator<key_type> d_keys_checking_output(
-                d_keys_output,
+                d_keys_output.get(),
                 out_of_bounds.device_pointer(),
                 size1 + size2);
 
-            // temp storage
-            size_t temp_storage_size_bytes;
-            void * d_temp_storage = nullptr;
             // Get size of d_temp_storage
-            HIP_CHECK(rocprim::merge<config>(d_temp_storage,
+            size_t temp_storage_size_bytes;
+            HIP_CHECK(rocprim::merge<config>(nullptr,
                                              temp_storage_size_bytes,
-                                             d_keys_input1,
-                                             d_keys_input2,
+                                             d_keys_input1.get(),
+                                             d_keys_input2.get(),
                                              d_keys_checking_output,
                                              keys_input1.size(),
                                              keys_input2.size(),
@@ -215,7 +231,14 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKey)
             ASSERT_GT(temp_storage_size_bytes, 0);
 
             // allocate temporary storage
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
+            common::device_ptr<void> d_temp_storage;
+
+            if(!d_temp_storage.resize_with_memory_check(temp_storage_size_bytes))
+            {
+                std::cout << "Out of memory. Skipping test with sizes = {" << size1 << ", " << size2
+                          << "}" << std::endl;
+                break;
+            }
 
             test_utils::GraphHelper gHelper;
             if(TestFixture::use_graphs)
@@ -224,10 +247,10 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKey)
             }
 
             // Run
-            HIP_CHECK(rocprim::merge<config>(d_temp_storage,
+            HIP_CHECK(rocprim::merge<config>(d_temp_storage.get(),
                                              temp_storage_size_bytes,
-                                             d_keys_input1,
-                                             d_keys_input2,
+                                             d_keys_input1.get(),
+                                             d_keys_input2.get(),
                                              d_keys_checking_output,
                                              keys_input1.size(),
                                              keys_input2.size(),
@@ -246,22 +269,11 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKey)
             ASSERT_FALSE(out_of_bounds.get());
 
             // Copy keys_output to host
-            HIP_CHECK(
-                hipMemcpy(
-                    keys_output.data(), d_keys_output,
-                    keys_output.size() * sizeof(key_type),
-                    hipMemcpyDeviceToHost
-                )
-            );
+            keys_output = d_keys_output.load();
 
             // Check if keys_output values are as expected
             ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(keys_output, expected));
 
-            HIP_CHECK(hipFree(d_keys_input1));
-            HIP_CHECK(hipFree(d_keys_input2));
-            HIP_CHECK(hipFree(d_keys_output));
-            HIP_CHECK(hipFree(d_temp_storage));
-
             if (TestFixture::use_graphs)
                 gHelper.cleanupGraphHelper();
         }
@@ -293,11 +305,19 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKeyValue)
 
     for(auto sizes : get_sizes())
     {
-        if ((std::get<0>(sizes) == 0 || std::get<1>(sizes) == 0) && test_common_utils::use_hmm())
+        if((std::get<0>(sizes) == 0 || std::get<1>(sizes) == 0) && common::use_hmm())
         {
             // hipMallocManaged() currently doesnt support zero byte allocation
             continue;
         }
+
+        if((std::get<0>(sizes) + std::get<1>(sizes) >= 100000
+            && sizeof(key_type) > sizeof(size_t) * 16))
+        {
+            // Huge types are slow
+            continue;
+        }
+
         SCOPED_TRACE(
             testing::Message() << "with sizes = {" <<
             std::get<0>(sizes) << ", " << std::get<1>(sizes) << "}"
@@ -309,14 +329,16 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKeyValue)
         // compare function
         compare_op_type compare_op;
 
-        for (size_t seed_index = 0; seed_index < random_seeds_count + seed_size; seed_index++)
+        for(size_t seed_index = 0; seed_index < number_of_runs; seed_index++)
         {
             unsigned int seed_value = seed_index < random_seeds_count  ? rand() : seeds[seed_index - random_seeds_count];
             SCOPED_TRACE(testing::Message() << "with seed = " << seed_value);
 
             // Generate data
-            std::vector<key_type> keys_input1 = test_utils::get_random_data<key_type>(size1, 0, size1, seed_value);
-            std::vector<key_type> keys_input2 = test_utils::get_random_data<key_type>(size2, 0, size2, seed_value);
+            std::vector<key_type> keys_input1
+                = test_utils::get_random_data_wrapped<key_type>(size1, 0, size1, seed_value);
+            std::vector<key_type> keys_input2
+                = test_utils::get_random_data_wrapped<key_type>(size2, 0, size2, seed_value);
             std::sort(keys_input1.begin(), keys_input1.end(), compare_op);
             std::sort(keys_input2.begin(), keys_input2.end(), compare_op);
             std::vector<value_type> values_input1(size1);
@@ -351,68 +373,48 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKeyValue)
 
             test_utils::out_of_bounds_flag out_of_bounds;
 
-            key_type * d_keys_input1;
-            key_type * d_keys_input2;
-            key_type * d_keys_output;
-            value_type * d_values_input1;
-            value_type * d_values_input2;
-            value_type * d_values_output;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_keys_input1, keys_input1.size() * sizeof(key_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_keys_input2, keys_input2.size() * sizeof(key_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_keys_output, keys_output.size() * sizeof(key_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_values_input1, values_input1.size() * sizeof(value_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_values_input2, values_input2.size() * sizeof(value_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_values_output, values_output.size() * sizeof(value_type)));
-            HIP_CHECK(
-                hipMemcpy(
-                    d_keys_input1, keys_input1.data(),
-                    keys_input1.size() * sizeof(key_type),
-                    hipMemcpyHostToDevice
-                )
-            );
-            HIP_CHECK(
-                hipMemcpy(
-                    d_keys_input2, keys_input2.data(),
-                    keys_input2.size() * sizeof(key_type),
-                    hipMemcpyHostToDevice
-                )
-            );
-            HIP_CHECK(
-                hipMemcpy(
-                    d_values_input1, values_input1.data(),
-                    values_input1.size() * sizeof(value_type),
-                    hipMemcpyHostToDevice
-                )
-            );
-            HIP_CHECK(
-                hipMemcpy(
-                    d_values_input2, values_input2.data(),
-                    values_input2.size() * sizeof(value_type),
-                    hipMemcpyHostToDevice
-                )
-            );
+            common::device_ptr<key_type>   d_keys_input1;
+            common::device_ptr<key_type>   d_keys_input2;
+            common::device_ptr<key_type>   d_keys_output;
+            common::device_ptr<value_type> d_values_input1;
+            common::device_ptr<value_type> d_values_input2;
+            common::device_ptr<value_type> d_values_output;
+
+            if(!d_keys_input1.resize_with_memory_check(keys_input1.size())
+               || !d_keys_input2.resize_with_memory_check(keys_input2.size())
+               || !d_keys_output.resize_with_memory_check(keys_output.size())
+               || !d_values_input1.resize_with_memory_check(values_input1.size())
+               || !d_values_input2.resize_with_memory_check(values_input1.size())
+               || !d_values_output.resize_with_memory_check(values_output.size()))
+            {
+                std::cout << "Out of memory. Skipping test with sizes = {" << size1 << ", " << size2
+                          << "}" << std::endl;
+                break;
+            }
+
+            d_keys_input1.store(keys_input1);
+            d_keys_input2.store(keys_input2);
+            d_values_input1.store(values_input1);
+            d_values_input2.store(values_input2);
 
             test_utils::bounds_checking_iterator<key_type> d_keys_checking_output(
-                d_keys_output,
+                d_keys_output.get(),
                 out_of_bounds.device_pointer(),
-                size1 + size2
-            );
+                size1 + size2);
             test_utils::bounds_checking_iterator<value_type> d_values_checking_output(
-                d_values_output,
+                d_values_output.get(),
                 out_of_bounds.device_pointer(),
                 size1 + size2);
 
-            // temp storage
-            size_t temp_storage_size_bytes;
-            void * d_temp_storage = nullptr;
             // Get size of d_temp_storage
-            HIP_CHECK(rocprim::merge<config>(d_temp_storage,
+            size_t temp_storage_size_bytes;
+            HIP_CHECK(rocprim::merge<config>(nullptr,
                                              temp_storage_size_bytes,
-                                             d_keys_input1,
-                                             d_keys_input2,
+                                             d_keys_input1.get(),
+                                             d_keys_input2.get(),
                                              d_keys_checking_output,
-                                             d_values_input1,
-                                             d_values_input2,
+                                             d_values_input1.get(),
+                                             d_values_input2.get(),
                                              d_values_checking_output,
                                              keys_input1.size(),
                                              keys_input2.size(),
@@ -424,7 +426,14 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKeyValue)
             ASSERT_GT(temp_storage_size_bytes, 0);
 
             // allocate temporary storage
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
+            common::device_ptr<void> d_temp_storage;
+
+            if(!d_temp_storage.resize_with_memory_check(temp_storage_size_bytes))
+            {
+                std::cout << "Out of memory. Skipping test with sizes = {" << size1 << ", " << size2
+                          << "}" << std::endl;
+                break;
+            }
 
             test_utils::GraphHelper gHelper;
             if(TestFixture::use_graphs)
@@ -433,13 +442,13 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKeyValue)
             }
 
             // Run
-            HIP_CHECK(rocprim::merge<config>(d_temp_storage,
+            HIP_CHECK(rocprim::merge<config>(d_temp_storage.get(),
                                              temp_storage_size_bytes,
-                                             d_keys_input1,
-                                             d_keys_input2,
+                                             d_keys_input1.get(),
+                                             d_keys_input2.get(),
                                              d_keys_checking_output,
-                                             d_values_input1,
-                                             d_values_input2,
+                                             d_values_input1.get(),
+                                             d_values_input2.get(),
                                              d_values_checking_output,
                                              keys_input1.size(),
                                              keys_input2.size(),
@@ -457,20 +466,8 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKeyValue)
 
             ASSERT_FALSE(out_of_bounds.get());
 
-            HIP_CHECK(
-                hipMemcpy(
-                    keys_output.data(), d_keys_output,
-                    keys_output.size() * sizeof(key_type),
-                    hipMemcpyDeviceToHost
-                )
-            );
-            HIP_CHECK(
-                hipMemcpy(
-                    values_output.data(), d_values_output,
-                    values_output.size() * sizeof(value_type),
-                    hipMemcpyDeviceToHost
-                )
-            );
+            keys_output   = d_keys_output.load();
+            values_output = d_values_output.load();
 
             // Check if keys_output values are as expected
             std::vector<key_type> expected_key(expected.size());
@@ -483,21 +480,15 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKeyValue)
             ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(keys_output, expected_key));
             ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(values_output, expected_value));
 
-            HIP_CHECK(hipFree(d_keys_input1));
-            HIP_CHECK(hipFree(d_keys_input2));
-            HIP_CHECK(hipFree(d_keys_output));
-            HIP_CHECK(hipFree(d_values_input1));
-            HIP_CHECK(hipFree(d_values_input2));
-            HIP_CHECK(hipFree(d_values_output));
-            HIP_CHECK(hipFree(d_temp_storage));
-
             if (TestFixture::use_graphs)
                 gHelper.cleanupGraphHelper();
         }
     }
 
     if (TestFixture::use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 template<bool UseGraphs = false, typename config = rocprim::default_config>
@@ -520,18 +511,8 @@ void testMergeMismatchedIteratorTypes()
     std::vector<int> expected_keys_output(2 * keys_input1.size());
     std::iota(expected_keys_output.begin(), expected_keys_output.end(), 0);
 
-    int* d_keys_input1 = nullptr;
-    int* d_keys_output = nullptr;
-    HIP_CHECK(test_common_utils::hipMallocHelper(&d_keys_input1,
-                                                 keys_input1.size() * sizeof(keys_input1[0])));
-    HIP_CHECK(
-        test_common_utils::hipMallocHelper(&d_keys_output,
-                                           expected_keys_output.size() * sizeof(keys_input1[0])));
-
-    HIP_CHECK(hipMemcpy(d_keys_input1,
-                        keys_input1.data(),
-                        keys_input1.size() * sizeof(keys_input1[0]),
-                        hipMemcpyHostToDevice));
+    common::device_ptr<int> d_keys_input1(keys_input1);
+    common::device_ptr<int> d_keys_output(expected_keys_output.size());
 
     const auto d_keys_input2
         = rocprim::make_transform_iterator(rocprim::make_counting_iterator(0),
@@ -549,9 +530,9 @@ void testMergeMismatchedIteratorTypes()
     size_t temp_storage_size_bytes = 0;
     HIP_CHECK(rocprim::merge<config>(nullptr,
                                      temp_storage_size_bytes,
-                                     d_keys_input1,
+                                     d_keys_input1.get(),
                                      d_keys_input2,
-                                     d_keys_output,
+                                     d_keys_output.get(),
                                      keys_input1.size(),
                                      keys_input1.size(),
                                      rocprim::less<int>{},
@@ -560,8 +541,7 @@ void testMergeMismatchedIteratorTypes()
 
     ASSERT_GT(temp_storage_size_bytes, 0);
 
-    void* d_temp_storage = nullptr;
-    HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
+    common::device_ptr<void> d_temp_storage(temp_storage_size_bytes);
 
     test_utils::GraphHelper gHelper;
     if(UseGraphs)
@@ -569,11 +549,11 @@ void testMergeMismatchedIteratorTypes()
         gHelper.startStreamCapture(stream);
     }
 
-    HIP_CHECK(rocprim::merge<config>(d_temp_storage,
+    HIP_CHECK(rocprim::merge<config>(d_temp_storage.get(),
                                      temp_storage_size_bytes,
-                                     d_keys_input1,
+                                     d_keys_input1.get(),
                                      d_keys_input2,
-                                     d_keys_output,
+                                     d_keys_output.get(),
                                      keys_input1.size(),
                                      keys_input1.size(),
                                      rocprim::less<int>{},
@@ -585,18 +565,10 @@ void testMergeMismatchedIteratorTypes()
         gHelper.createAndLaunchGraph(stream);
     }
 
-    std::vector<int> keys_output(expected_keys_output.size());
-    HIP_CHECK(hipMemcpy(keys_output.data(),
-                        d_keys_output,
-                        keys_output.size() * sizeof(keys_output[0]),
-                        hipMemcpyDeviceToHost));
+    const auto keys_output = d_keys_output.load();
 
     ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(keys_output, expected_keys_output));
 
-    HIP_CHECK(hipFree(d_temp_storage));
-    HIP_CHECK(hipFree(d_keys_output));
-    HIP_CHECK(hipFree(d_keys_input1));
-
     if (UseGraphs)
     {
         gHelper.cleanupGraphHelper();
