@@ -1,6 +1,6 @@
 // MIT License
 //
-// Copyright (c) 2024 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -20,41 +20,52 @@
 // OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 // SOFTWARE.
 
+#include "../common_test_header.hpp"
+
+#include "../../common/utils_custom_type.hpp"
+#include "../../common/utils_data_generation.hpp"
+#include "../../common/utils_device_ptr.hpp"
+
 // required test headers
 #include "indirect_iterator.hpp"
+#include "test_seed.hpp"
+#include "test_utils.hpp"
 #include "test_utils_assertions.hpp"
 #include "test_utils_custom_float_type.hpp"
 #include "test_utils_custom_test_types.hpp"
 #include "test_utils_data_generation.hpp"
-#include "test_utils_types.hpp"
-
-#include "../common_test_header.hpp"
+#include "test_utils_hipgraphs.hpp"
+#include "test_utils_sort_comparator.hpp"
 
 // required rocprim headers
-#include <algorithm>
+#include <rocprim/block/block_radix_rank.hpp>
+#include <rocprim/config.hpp>
 #include <rocprim/device/config_types.hpp>
 #include <rocprim/device/detail/device_config_helper.hpp>
 #include <rocprim/device/device_partial_sort.hpp>
+#include <rocprim/device/device_partial_sort_config.hpp>
 #include <rocprim/functional.hpp>
+#include <rocprim/types.hpp>
 
-#include <iostream>
-#include <iterator>
-#include <vector>
-
-#include <cassert>
+#include <algorithm>
 #include <cstddef>
+#include <stdint.h>
+#include <type_traits>
+#include <vector>
 
 // Params for tests
 template<class KeyType,
          class CompareFunction    = ::rocprim::less<KeyType>,
          class Config             = ::rocprim::default_config,
          bool UseGraphs           = false,
-         bool UseIndirectIterator = false>
+         bool UseIndirectIterator = false,
+         class Decomposer         = ::rocprim::identity_decomposer>
 struct DevicePartialSortParams
 {
     using key_type                              = KeyType;
     using compare_function                      = CompareFunction;
     using config                                = Config;
+    using decomposer                            = Decomposer;
     static constexpr bool use_graphs            = UseGraphs;
     static constexpr bool use_indirect_iterator = UseIndirectIterator;
 };
@@ -66,6 +77,7 @@ public:
     using key_type                              = typename Params::key_type;
     using compare_function                      = typename Params::compare_function;
     using config                                = typename Params::config;
+    using decomposer                            = typename Params::decomposer;
     const bool            debug_synchronous     = false;
     static constexpr bool use_graphs            = Params::use_graphs;
     static constexpr bool use_indirect_iterator = Params::use_indirect_iterator;
@@ -75,17 +87,17 @@ public:
 using RocprimDevicePartialSortTestsParams = ::testing::Types<
     DevicePartialSortParams<unsigned short>,
     DevicePartialSortParams<char>,
-    DevicePartialSortParams<int>,
-    DevicePartialSortParams<test_utils::custom_test_type<int>>,
+    DevicePartialSortParams<const int>,
+    DevicePartialSortParams<common::custom_type<int, int, true>>,
     DevicePartialSortParams<unsigned long>,
     DevicePartialSortParams<long long, ::rocprim::greater<long long>>,
-    DevicePartialSortParams<float>,
+    DevicePartialSortParams<const float>,
     DevicePartialSortParams<int8_t>,
     DevicePartialSortParams<uint8_t>,
     DevicePartialSortParams<rocprim::half>,
     DevicePartialSortParams<rocprim::bfloat16>,
     DevicePartialSortParams<double>,
-    DevicePartialSortParams<test_utils::custom_test_type<float>>,
+    DevicePartialSortParams<common::custom_type<float, float, true>>,
     DevicePartialSortParams<test_utils::custom_float_type>,
     DevicePartialSortParams<test_utils::custom_test_array_type<int, 4>>,
     DevicePartialSortParams<int, ::rocprim::less<int>, rocprim::default_config, false, true>,
@@ -94,7 +106,14 @@ using RocprimDevicePartialSortTestsParams = ::testing::Types<
         ::rocprim::less<int>,
         rocprim::partial_sort_config<
             rocprim::
-                nth_element_config<128, 4, 32, 16, rocprim::block_radix_rank_algorithm::basic>>>>;
+                nth_element_config<128, 4, 32, 16, rocprim::block_radix_rank_algorithm::basic>>>,
+    DevicePartialSortParams<
+        common::custom_type<int, int, true>,
+        ::rocprim::less<common::custom_type<int, int, true>>,
+        rocprim::default_config,
+        false,
+        false,
+        test_utils::custom_test_type_decomposer<common::custom_type<int, int, true>>>>;
 
 template<class InputVector, class OutputVector, class CompareFunction>
 void inline compare_partial_sort_cpp_14(InputVector     input,
@@ -121,7 +140,6 @@ void inline compare_partial_sort_cpp_14(InputVector     input,
     ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(sorted_output, sorted_input));
 }
 
-#if CPP17
 template<class InputVector, class OutputVector, class CompareFunction>
 void inline compare_partial_sort_cpp_17(InputVector     input,
                                         OutputVector    output,
@@ -150,7 +168,6 @@ void inline compare_partial_sort_cpp_17(InputVector     input,
     // Check if the values are the same
     ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(sorted_output, sorted_input));
 }
-#endif
 
 template<class InputVector, class OutputVector, class CompareFunction>
 void inline compare_partial_sort(InputVector     input,
@@ -159,12 +176,7 @@ void inline compare_partial_sort(InputVector     input,
                                  CompareFunction compare_op)
 {
     compare_partial_sort_cpp_14(input, output, middle, compare_op);
-#if CPP17
-    // this comparison is only compiled and executed if c++17 is available
     compare_partial_sort_cpp_17(input, output, middle, compare_op);
-#else
-    ROCPRIM_PRAGMA_MESSAGE("c++17 not available skips direct comparison with std::partial_sort");
-#endif
 }
 
 TYPED_TEST_SUITE(RocprimDevicePartialSortTests, RocprimDevicePartialSortTestsParams);
@@ -175,13 +187,14 @@ TYPED_TEST(RocprimDevicePartialSortTests, PartialSort)
     SCOPED_TRACE(testing::Message() << "with device_id = " << device_id);
     HIP_CHECK(hipSetDevice(device_id));
 
-    using key_type                              = typename TestFixture::key_type;
+    using key_type                              = std::remove_cv_t<typename TestFixture::key_type>;
     using compare_function                      = typename TestFixture::compare_function;
     using config                                = typename TestFixture::config;
+    using decomposer                            = typename TestFixture::decomposer;
     const bool            debug_synchronous     = TestFixture::debug_synchronous;
-    static constexpr bool use_indirect_iterator = TestFixture::use_indirect_iterator;
+    constexpr bool        use_indirect_iterator = TestFixture::use_indirect_iterator;
 
-    for(size_t seed_index = 0; seed_index < random_seeds_count + seed_size; ++seed_index)
+    for(size_t seed_index = 0; seed_index < number_of_runs; ++seed_index)
     {
         unsigned int seed_value
             = seed_index < random_seeds_count ? rand() : seeds[seed_index - random_seeds_count];
@@ -209,36 +222,24 @@ TYPED_TEST(RocprimDevicePartialSortTests, PartialSort)
                     HIP_CHECK(hipStreamCreateWithFlags(&stream, hipStreamNonBlocking));
                 }
 
-                std::vector<key_type> input;
-                if(rocprim::is_floating_point<key_type>::value)
-                {
-                    input = test_utils::get_random_data<key_type>(size, -1000, 1000, seed_value);
-                }
-                else
-                {
-                    input = test_utils::get_random_data<key_type>(
-                        size,
-                        test_utils::numeric_limits<key_type>::min(),
-                        test_utils::numeric_limits<key_type>::max(),
-                        seed_value);
-                }
-                key_type* d_input;
-                HIP_CHECK(test_common_utils::hipMallocHelper(&d_input, size * sizeof(key_type)));
-                HIP_CHECK(hipMemcpy(d_input,
-                                    input.data(),
-                                    size * sizeof(key_type),
-                                    hipMemcpyHostToDevice));
+                std::vector<key_type> input = test_utils::get_random_data_wrapped<key_type>(
+                    size,
+                    common::generate_limits<key_type>::min(),
+                    common::generate_limits<key_type>::max(),
+                    seed_value);
 
-                key_type* d_output;
-                d_output = d_input;
+                common::device_ptr<key_type> d_input(input);
+
+                common::device_ptr<key_type>& d_output = d_input;
 
                 const auto input_it
-                    = test_utils::wrap_in_indirect_iterator<use_indirect_iterator>(d_input);
+                    = test_utils::wrap_in_indirect_iterator<use_indirect_iterator>(d_input.get());
 
                 compare_function compare_op;
+                decomposer       decomposer_op;
 
                 // Allocate temporary storage
-                size_t temp_storage_size_bytes{};
+                size_t temp_storage_size_bytes;
                 HIP_CHECK(rocprim::partial_sort<config>(nullptr,
                                                         temp_storage_size_bytes,
                                                         input_it,
@@ -246,19 +247,19 @@ TYPED_TEST(RocprimDevicePartialSortTests, PartialSort)
                                                         size,
                                                         compare_op,
                                                         stream,
-                                                        debug_synchronous));
+                                                        debug_synchronous,
+                                                        decomposer_op));
 
                 ASSERT_GT(temp_storage_size_bytes, 0);
-                void* d_temp_storage{};
-                HIP_CHECK(
-                    test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
 
-                test_utils::GraphHelper gHelper;;
+                common::device_ptr<void> d_temp_storage(temp_storage_size_bytes);
+
+                test_utils::GraphHelper gHelper;
                 if(TestFixture::use_graphs)
                 {
                     gHelper.startStreamCapture(stream);
                 }
-                HIP_CHECK(rocprim::partial_sort<config>(d_temp_storage,
+                HIP_CHECK(rocprim::partial_sort<config>(d_temp_storage.get(),
                                                         temp_storage_size_bytes,
                                                         input_it,
                                                         middle,
@@ -274,17 +275,10 @@ TYPED_TEST(RocprimDevicePartialSortTests, PartialSort)
                     gHelper.createAndLaunchGraph(stream);
                 }
 
-                std::vector<key_type> output(size);
-                HIP_CHECK(hipMemcpy(output.data(),
-                                    d_output,
-                                    size * sizeof(key_type),
-                                    hipMemcpyDeviceToHost));
+                const auto output = d_output.load();
 
                 compare_partial_sort(input, output, middle, compare_op);
 
-                HIP_CHECK(hipFree(d_input));
-                HIP_CHECK(hipFree(d_temp_storage));
-
                 if(TestFixture::use_graphs)
                 {
                     gHelper.cleanupGraphHelper();
@@ -328,7 +322,6 @@ void inline compare_partial_sort_copy_cpp_14(InputVector     input,
     ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(output, expected_output));
 }
 
-#if CPP17
 template<class InputVector, class OutputVector, class CompareFunction>
 void inline compare_partial_sort_copy_cpp_17(InputVector     input,
                                              OutputVector    output,
@@ -354,7 +347,6 @@ void inline compare_partial_sort_copy_cpp_17(InputVector     input,
     // Check if the values are the same
     ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(sorted_output, output));
 }
-#endif
 
 template<class InputVector, class OutputVector, class CompareFunction>
 void inline compare_partial_sort_copy(InputVector     input,
@@ -364,13 +356,7 @@ void inline compare_partial_sort_copy(InputVector     input,
                                       CompareFunction compare_op)
 {
     compare_partial_sort_copy_cpp_14(input, output, orignal_output, middle, compare_op);
-#if CPP17
-    // this comparison is only compiled and executed if c++17 is available
     compare_partial_sort_copy_cpp_17(input, output, orignal_output, middle, compare_op);
-#else
-    ROCPRIM_PRAGMA_MESSAGE(
-        "c++17 not available skips direct comparison with std::partial_sort_copy");
-#endif
 }
 
 TYPED_TEST(RocprimDevicePartialSortTests, PartialSortCopy)
@@ -379,13 +365,15 @@ TYPED_TEST(RocprimDevicePartialSortTests, PartialSortCopy)
     SCOPED_TRACE(testing::Message() << "with device_id = " << device_id);
     HIP_CHECK(hipSetDevice(device_id));
 
-    using key_type                              = typename TestFixture::key_type;
+    using key_type                              = std::remove_cv_t<typename TestFixture::key_type>;
     using compare_function                      = typename TestFixture::compare_function;
     using config                                = typename TestFixture::config;
+    using decomposer                            = typename TestFixture::decomposer;
     const bool            debug_synchronous     = TestFixture::debug_synchronous;
-    static constexpr bool use_indirect_iterator = TestFixture::use_indirect_iterator;
+    constexpr bool        input_is_const        = std::is_const_v<typename TestFixture::key_type>;
+    constexpr bool        use_indirect_iterator = TestFixture::use_indirect_iterator;
 
-    for(size_t seed_index = 0; seed_index < random_seeds_count + seed_size; ++seed_index)
+    for(size_t seed_index = 0; seed_index < number_of_runs; ++seed_index)
     {
         unsigned int seed_value
             = seed_index < random_seeds_count ? rand() : seeds[seed_index - random_seeds_count];
@@ -413,47 +401,26 @@ TYPED_TEST(RocprimDevicePartialSortTests, PartialSortCopy)
                     HIP_CHECK(hipStreamCreateWithFlags(&stream, hipStreamNonBlocking));
                 }
 
-                std::vector<key_type> input;
-                std::vector<key_type> output_original;
-                if(rocprim::is_floating_point<key_type>::value)
-                {
-                    input = test_utils::get_random_data<key_type>(size, -1000, 1000, seed_value);
-                    output_original
-                        = test_utils::get_random_data<key_type>(size, -1000, 1000, seed_value + 1);
-                }
-                else
-                {
-                    input = test_utils::get_random_data<key_type>(
+                std::vector<key_type> input = test_utils::get_random_data_wrapped<key_type>(
+                    size,
+                    common::generate_limits<key_type>::min(),
+                    common::generate_limits<key_type>::max(),
+                    seed_value);
+                std::vector<key_type> output_original
+                    = test_utils::get_random_data_wrapped<key_type>(
                         size,
-                        test_utils::numeric_limits<key_type>::min(),
-                        test_utils::numeric_limits<key_type>::max(),
-                        seed_value);
-                    output_original = test_utils::get_random_data<key_type>(
-                        size,
-                        test_utils::numeric_limits<key_type>::min(),
-                        test_utils::numeric_limits<key_type>::max(),
+                        common::generate_limits<key_type>::min(),
+                        common::generate_limits<key_type>::max(),
                         seed_value + 1);
-                }
 
-                key_type* d_input;
-                HIP_CHECK(test_common_utils::hipMallocHelper(&d_input, size * sizeof(key_type)));
-                HIP_CHECK(hipMemcpy(d_input,
-                                    input.data(),
-                                    size * sizeof(key_type),
-                                    hipMemcpyHostToDevice));
+                common::device_ptr<key_type> d_input(input);
+                common::device_ptr<key_type> d_output(output_original);
 
-                key_type* d_output;
-
-                HIP_CHECK(test_common_utils::hipMallocHelper(&d_output, size * sizeof(key_type)));
-                HIP_CHECK(hipMemcpy(d_output,
-                                    output_original.data(),
-                                    size * sizeof(key_type),
-                                    hipMemcpyHostToDevice));
-
-                const auto input_it
-                    = test_utils::wrap_in_indirect_iterator<use_indirect_iterator>(d_input);
+                const auto input_it = test_utils::wrap_in_indirect_iterator<use_indirect_iterator>(
+                    test_utils::wrap_in_const<input_is_const>(d_input.get()));
 
                 compare_function compare_op;
+                decomposer       decomposer_op;
 
                 // Allocate temporary storage
                 size_t temp_storage_size_bytes{};
@@ -461,28 +428,28 @@ TYPED_TEST(RocprimDevicePartialSortTests, PartialSortCopy)
                 HIP_CHECK(rocprim::partial_sort_copy<config>(nullptr,
                                                              temp_storage_size_bytes,
                                                              input_it,
-                                                             d_output,
+                                                             d_output.get(),
                                                              middle,
                                                              size,
                                                              compare_op,
                                                              stream,
-                                                             debug_synchronous));
+                                                             debug_synchronous,
+                                                             decomposer_op));
 
                 ASSERT_GT(temp_storage_size_bytes, 0);
-                void* d_temp_storage{};
-                HIP_CHECK(
-                    test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
 
-                test_utils::GraphHelper gHelper;;
+                common::device_ptr<void> d_temp_storage(temp_storage_size_bytes);
+
+                test_utils::GraphHelper gHelper;
                 if(TestFixture::use_graphs)
                 {
                     gHelper.startStreamCapture(stream);
                 }
 
-                HIP_CHECK(rocprim::partial_sort_copy<config>(d_temp_storage,
+                HIP_CHECK(rocprim::partial_sort_copy<config>(d_temp_storage.get(),
                                                              temp_storage_size_bytes,
                                                              input_it,
-                                                             d_output,
+                                                             d_output.get(),
                                                              middle,
                                                              size,
                                                              compare_op,
@@ -496,18 +463,10 @@ TYPED_TEST(RocprimDevicePartialSortTests, PartialSortCopy)
                     gHelper.createAndLaunchGraph(stream);
                 }
 
-                std::vector<key_type> output(size);
-                HIP_CHECK(hipMemcpy(output.data(),
-                                    d_output,
-                                    size * sizeof(key_type),
-                                    hipMemcpyDeviceToHost));
+                const auto output = d_output.load();
 
                 compare_partial_sort_copy(input, output, output_original, middle, compare_op);
 
-                HIP_CHECK(hipFree(d_input));
-                HIP_CHECK(hipFree(d_output));
-                HIP_CHECK(hipFree(d_temp_storage));
-
                 if(TestFixture::use_graphs)
                 {
                     gHelper.cleanupGraphHelper();
