@@ -1,6 +1,6 @@
 // MIT License
 //
-// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -22,14 +22,39 @@
 
 #include "../common_test_header.hpp"
 
+#include "../../common/utils_custom_type.hpp"
+#include "../../common/utils_device_ptr.hpp"
+
+// required test headers
+#include "identity_iterator.hpp"
+#include "test_seed.hpp"
+#include "test_utils.hpp"
+#include "test_utils_assertions.hpp"
+#include "test_utils_custom_test_types.hpp"
+#include "test_utils_data_generation.hpp"
+#include "test_utils_hipgraphs.hpp"
+
 // required rocprim headers
+#include <rocprim/block/block_reduce.hpp>
+#include <rocprim/config.hpp>
+#include <rocprim/device/config_types.hpp>
+#include <rocprim/device/detail/device_config_helper.hpp>
 #include <rocprim/device/device_reduce.hpp>
 #include <rocprim/functional.hpp>
 #include <rocprim/iterator/constant_iterator.hpp>
 #include <rocprim/iterator/counting_iterator.hpp>
+#include <rocprim/thread/thread_operators.hpp>
+#include <rocprim/type_traits.hpp>
+#include <rocprim/types.hpp>
+#include <rocprim/types/key_value_pair.hpp>
 
-// required test headers
-#include "test_utils_types.hpp"
+#include <algorithm>
+#include <cstddef>
+#include <iostream>
+#include <limits>
+#include <stdint.h>
+#include <type_traits>
+#include <vector>
 
 using bra = ::rocprim::block_reduce_algorithm;
 
@@ -92,7 +117,7 @@ public:
 template<class Params>
 class RocprimDeviceReducePrecisionTests : public RocprimDeviceReduceTests<Params>{};
 
-typedef ::testing::Types<
+using RocprimDeviceReduceTestsParams = ::testing::Types<
     DeviceReduceParams<unsigned int>,
     DeviceReduceParams<long, long, true>,
     DeviceReduceParams<short, int>,
@@ -107,8 +132,10 @@ typedef ::testing::Types<
     DeviceReduceParams<uint8_t, uint8_t>,
     DeviceReduceParams<rocprim::half, rocprim::half>,
     DeviceReduceParams<rocprim::bfloat16, rocprim::bfloat16>,
-    DeviceReduceParams<test_utils::custom_test_type<float>, test_utils::custom_test_type<float>>,
-    DeviceReduceParams<test_utils::custom_test_type<int>, test_utils::custom_test_type<float>>,
+    DeviceReduceParams<common::custom_type<float, float, true>,
+                       common::custom_type<float, float, true>>,
+    DeviceReduceParams<common::custom_type<int, int, true>,
+                       common::custom_type<float, float, true>>,
     DeviceReduceParams<rocprim::half,
                        rocprim::half,
                        false,
@@ -130,21 +157,20 @@ typedef ::testing::Types<
                        bra::default_algorithm,
                        false,
                        true>,
-    DeviceReduceParams<test_utils::custom_test_type<double>,
-                       test_utils::custom_test_type<double>,
+    DeviceReduceParams<common::custom_type<double, double, true>,
+                       common::custom_type<double, double, true>,
                        false,
                        ROCPRIM_GRID_SIZE_LIMIT,
                        bra::default_algorithm,
                        false,
                        true>,
-    DeviceReduceParams<int, int, false, ROCPRIM_GRID_SIZE_LIMIT, bra::default_algorithm, true>>
-    RocprimDeviceReduceTestsParams;
+    DeviceReduceParams<int, int, false, ROCPRIM_GRID_SIZE_LIMIT, bra::default_algorithm, true>>;
 
-typedef ::testing::Types<DeviceReduceParams<double, double>,
-                         DeviceReduceParamsList(float, float, false, 2048),
-                         DeviceReduceParams<rocprim::half, rocprim::half>,
-                         DeviceReduceParams<rocprim::bfloat16, rocprim::bfloat16>>
-    RocprimDeviceReducePrecisionTestsParams;
+using RocprimDeviceReducePrecisionTestsParams
+    = ::testing::Types<DeviceReduceParams<double, double>,
+                       DeviceReduceParamsList(float, float, false, 2048),
+                       DeviceReduceParams<rocprim::half, rocprim::half>,
+                       DeviceReduceParams<rocprim::bfloat16, rocprim::bfloat16>>;
 
 TYPED_TEST_SUITE(RocprimDeviceReduceTests, RocprimDeviceReduceTestsParams);
 TYPED_TEST_SUITE(RocprimDeviceReducePrecisionTests, RocprimDeviceReducePrecisionTestsParams);
@@ -167,25 +193,23 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceEmptyInput)
         HIP_CHECK(hipStreamCreateWithFlags(&stream, hipStreamNonBlocking));
     }
 
-    U * d_output;
-    HIP_CHECK(test_common_utils::hipMallocHelper(&d_output, sizeof(U)));
+    common::device_ptr<U> d_output(1);
 
     const U initial_value = U(1234);
 
-    size_t temp_storage_size_bytes;
     // Get size of d_temp_storage
-    HIP_CHECK(
-        rocprim::reduce<Config>(
-            nullptr, temp_storage_size_bytes,
-            rocprim::make_constant_iterator<T>(T(345)),
-            d_output,
-            initial_value,
-            0, rocprim::minimum<U>(), stream, debug_synchronous
-        )
-    );
+    size_t temp_storage_size_bytes;
+    HIP_CHECK(rocprim::reduce<Config>(nullptr,
+                                      temp_storage_size_bytes,
+                                      rocprim::make_constant_iterator<T>(T(345)),
+                                      d_output.get(),
+                                      initial_value,
+                                      0,
+                                      rocprim::minimum<U>(),
+                                      stream,
+                                      debug_synchronous));
 
-    void * d_temp_storage = nullptr;
-    HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
+    common::device_ptr<void> d_temp_storage(temp_storage_size_bytes);
 
     test_utils::GraphHelper gHelper;
     if(TestFixture::use_graphs)
@@ -194,17 +218,16 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceEmptyInput)
     }
 
     // Run
-    HIP_CHECK(
-        rocprim::reduce<Config>(
-            d_temp_storage, temp_storage_size_bytes,
-            rocprim::make_constant_iterator<T>(T(345)),
-            d_output,
-            initial_value,
-            0, rocprim::minimum<U>(), stream, debug_synchronous
-        )
-    );
+    HIP_CHECK(rocprim::reduce<Config>(d_temp_storage.get(),
+                                      temp_storage_size_bytes,
+                                      rocprim::make_constant_iterator<T>(T(345)),
+                                      d_output.get(),
+                                      initial_value,
+                                      0,
+                                      rocprim::minimum<U>(),
+                                      stream,
+                                      debug_synchronous));
 
-    
     if(TestFixture::use_graphs)
     {
         gHelper.createAndLaunchGraph(stream);
@@ -212,18 +235,9 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceEmptyInput)
 
     HIP_CHECK(hipDeviceSynchronize());
 
-    U output;
-    HIP_CHECK(
-        hipMemcpy(
-            &output, d_output,
-            sizeof(U),
-            hipMemcpyDeviceToHost
-        )
-    );
-    ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(output, initial_value));
+    const auto output = d_output.load()[0];
 
-    hipFree(d_output);
-    hipFree(d_temp_storage);
+    ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(output, initial_value));
 
     if (TestFixture::use_graphs)
     {
@@ -245,7 +259,7 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceSum)
     static constexpr bool use_identity_iterator = TestFixture::use_identity_iterator;
     using Config = size_limit_config_t<TestFixture::size_limit>;
 
-    for (size_t seed_index = 0; seed_index < random_seeds_count + seed_size; seed_index++)
+    for(size_t seed_index = 0; seed_index < number_of_runs; seed_index++)
     {
         unsigned int seed_value = seed_index < random_seeds_count  ? rand() : seeds[seed_index - random_seeds_count];
         SCOPED_TRACE(testing::Message() << "with seed = " << seed_value);
@@ -268,41 +282,25 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceSum)
                 HIP_CHECK(hipStreamCreateWithFlags(&stream, hipStreamNonBlocking));
             }
 
-
             SCOPED_TRACE(testing::Message() << "with size = " << size);
 
             // Generate data
             std::vector<T> input = test_utils::get_random_data<T>(size, 0, 100, seed_value);
-            std::vector<U> output(1, U(0));
 
-            T * d_input;
-            U * d_output;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_input, input.size() * sizeof(T)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_output, output.size() * sizeof(U)));
-            HIP_CHECK(
-                hipMemcpy(
-                    d_input, input.data(),
-                    input.size() * sizeof(T),
-                    hipMemcpyHostToDevice
-                )
-            );
-            HIP_CHECK(hipDeviceSynchronize());
+            common::device_ptr<T> d_input(input);
+            common::device_ptr<U> d_output(1);
 
             // Calculate expected results on host
-            U expected = test_utils::host_reduce(input.begin(), input.end(), rocprim::plus<U>());
-            // fix for custom_test_type case with size == 0
-            if(size == 0)
-                expected = U();
+            std::vector<U> expected
+                = test_utils::host_reduce(input.begin(), input.end(), rocprim::plus<U>());
 
-            // temp storage
-            size_t temp_storage_size_bytes;
-            void * d_temp_storage = nullptr;
             // Get size of d_temp_storage
+            size_t temp_storage_size_bytes;
             HIP_CHECK(rocprim::reduce<Config>(
-                d_temp_storage,
+                nullptr,
                 temp_storage_size_bytes,
-                d_input,
-                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
+                d_input.get(),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output.get()),
                 input.size(),
                 rocprim::plus<U>(),
                 stream,
@@ -312,8 +310,7 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceSum)
             ASSERT_GT(temp_storage_size_bytes, 0);
 
             // allocate temporary storage
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
-            HIP_CHECK(hipDeviceSynchronize());
+            common::device_ptr<void> d_temp_storage(temp_storage_size_bytes);
 
             test_utils::GraphHelper gHelper;
             if(TestFixture::use_graphs)
@@ -322,16 +319,16 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceSum)
             }
 
             // Run
-            HIP_CHECK(
-                rocprim::reduce<Config>(
-                    d_temp_storage, temp_storage_size_bytes,
-                    d_input,
-                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
-                    input.size(), rocprim::plus<U>(), stream, debug_synchronous
-                )
-            );
+            HIP_CHECK(rocprim::reduce<Config>(
+                d_temp_storage.get(),
+                temp_storage_size_bytes,
+                d_input.get(),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output.get()),
+                input.size(),
+                rocprim::plus<U>(),
+                stream,
+                debug_synchronous));
 
-            
             if(TestFixture::use_graphs)
             {
                 gHelper.createAndLaunchGraph(stream);
@@ -341,22 +338,24 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceSum)
             HIP_CHECK(hipDeviceSynchronize());
 
             // Copy output to host
-            HIP_CHECK(
-                hipMemcpy(
-                    output.data(), d_output,
-                    output.size() * sizeof(U),
-                    hipMemcpyDeviceToHost
-                )
-            );
-            HIP_CHECK(hipDeviceSynchronize());
+            const auto output = d_output.load();
 
             // Check if output values are as expected
-            ASSERT_NO_FATAL_FAILURE(
-                test_utils::assert_near(output[0], expected, test_utils::precision<U> * size));
-
-            hipFree(d_input);
-            hipFree(d_output);
-            hipFree(d_temp_storage);
+            if(size == 0)
+            {
+                ASSERT_NO_FATAL_FAILURE(
+                    test_utils::assert_near(output[0], U{}, test_utils::precision<U> * 1));
+            }
+            else
+            {
+                for(size_t i = 0; i < output.size(); ++i)
+                {
+                    ASSERT_NO_FATAL_FAILURE(
+                        test_utils::assert_near(output[i],
+                                                expected[i],
+                                                test_utils::precision<U> * (size - 1 - i)));
+                }
+            }
 
             if (TestFixture::use_graphs)
             {
@@ -367,26 +366,6 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceSum)
     }
 }
 
-template<
-    class Key,
-    class Value
->
-struct arg_min
-{
-    ROCPRIM_HOST_DEVICE inline
-    rocprim::key_value_pair<Key, Value>
-    operator()(const rocprim::key_value_pair<Key, Value>& a,
-               const rocprim::key_value_pair<Key, Value>& b) const
-    {
-        rocprim::less<Value>     less_v;
-        rocprim::less<Key>       less_k;
-        rocprim::equal_to<Value> eq_v;
-        return (less_v(b.value, a.value) || (eq_v(a.value, b.value) && less_k(b.key, a.key))) ? b
-                                                                                              : a;
-    }
-};
-
-
 TYPED_TEST(RocprimDeviceReduceTests, ReduceArgMinimum)
 {
     int device_id = test_common_utils::obtain_device_from_ctest();
@@ -399,7 +378,7 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceArgMinimum)
     static constexpr bool use_identity_iterator = TestFixture::use_identity_iterator;
     using Config = size_limit_config_t<TestFixture::size_limit>;
 
-    for (size_t seed_index = 0; seed_index < random_seeds_count + seed_size; seed_index++)
+    for(size_t seed_index = 0; seed_index < number_of_runs; seed_index++)
     {
         unsigned int seed_value = seed_index < random_seeds_count  ? rand() : seeds[seed_index - random_seeds_count];
         SCOPED_TRACE(testing::Message() << "with seed = " << seed_value);
@@ -422,23 +401,12 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceArgMinimum)
                 input[i].key = (int)i;
                 input[i].value = test_utils::get_random_data<T>(1, 1, 100, seed_value)[0];
             }
-            std::vector<key_value> output(1);
 
-            key_value * d_input;
-            key_value * d_output;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_input, input.size() * sizeof(key_value)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_output, output.size() * sizeof(key_value)));
-            HIP_CHECK(
-                hipMemcpy(
-                    d_input, input.data(),
-                    input.size() * sizeof(key_value),
-                    hipMemcpyHostToDevice
-                )
-            );
-            HIP_CHECK(hipDeviceSynchronize());
+            common::device_ptr<key_value> d_input(input);
+            common::device_ptr<key_value> d_output(1);
 
-            arg_min<int, T> reduce_op;
-            const key_value max(std::numeric_limits<int>::max(), test_utils::numeric_limits<T>::max());
+            rocprim::arg_min reduce_op;
+            const key_value max(std::numeric_limits<int>::max(), rocprim::numeric_limits<T>::max());
 
             // Calculate expected results on host
             key_value expected = max;
@@ -447,25 +415,24 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceArgMinimum)
                 expected = reduce_op(expected, input[i]);
             }
 
-            // temp storage
-            size_t temp_storage_size_bytes;
-            void * d_temp_storage = nullptr;
             // Get size of d_temp_storage
-            HIP_CHECK(
-                rocprim::reduce<Config>(
-                    d_temp_storage, temp_storage_size_bytes,
-                    d_input,
-                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
-                    max, input.size(), reduce_op, stream, debug_synchronous
-                )
-            );
+            size_t temp_storage_size_bytes;
+            HIP_CHECK(rocprim::reduce<Config>(
+                nullptr,
+                temp_storage_size_bytes,
+                d_input.get(),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output.get()),
+                max,
+                input.size(),
+                reduce_op,
+                stream,
+                debug_synchronous));
 
             // temp_storage_size_bytes must be >0
             ASSERT_GT(temp_storage_size_bytes, 0);
 
             // allocate temporary storage
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
-            HIP_CHECK(hipDeviceSynchronize());
+            common::device_ptr<void> d_temp_storage(temp_storage_size_bytes);
 
             test_utils::GraphHelper gHelper;
             if(TestFixture::use_graphs)
@@ -474,16 +441,17 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceArgMinimum)
             }
 
             // Run
-            HIP_CHECK(
-                rocprim::reduce<Config>(
-                    d_temp_storage, temp_storage_size_bytes,
-                    d_input,
-                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
-                    max, input.size(), reduce_op, stream, debug_synchronous
-                )
-            );
+            HIP_CHECK(rocprim::reduce<Config>(
+                d_temp_storage.get(),
+                temp_storage_size_bytes,
+                d_input.get(),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output.get()),
+                max,
+                input.size(),
+                reduce_op,
+                stream,
+                debug_synchronous));
 
-            
             if(TestFixture::use_graphs)
             {
                 gHelper.createAndLaunchGraph(stream);
@@ -493,23 +461,12 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceArgMinimum)
             HIP_CHECK(hipDeviceSynchronize());
 
             // Copy output to host
-            HIP_CHECK(
-                hipMemcpy(
-                    output.data(), d_output,
-                    output.size() * sizeof(key_value),
-                    hipMemcpyDeviceToHost
-                )
-            );
-            HIP_CHECK(hipDeviceSynchronize());
+            const auto output = d_output.load();
 
             // Check if output values are as expected
             test_utils::assert_eq(output[0].key, expected.key);
             test_utils::assert_eq(output[0].value, expected.value);
 
-            hipFree(d_input);
-            hipFree(d_output);
-            hipFree(d_temp_storage);
-
             if (TestFixture::use_graphs)
             {
                 gHelper.cleanupGraphHelper();
@@ -537,7 +494,7 @@ void testLargeIndices()
         HIP_CHECK(hipStreamCreateWithFlags(&stream, hipStreamNonBlocking));
     }
 
-    for(size_t seed_index = 0; seed_index < random_seeds_count + seed_size; seed_index++)
+    for(size_t seed_index = 0; seed_index < number_of_runs; seed_index++)
     {
         unsigned int seed_value
             = seed_index < random_seeds_count ? rand() : seeds[seed_index - random_seeds_count];
@@ -549,25 +506,21 @@ void testLargeIndices()
 
             const Iterator input {0};
 
-            T* d_output = nullptr;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_output, sizeof(T)));
+            common::device_ptr<T> d_output(1);
 
-            // temp storage
-            size_t temp_storage_size_bytes = 0;
-            void*  d_temp_storage          = nullptr;
             // Get size of d_temp_storage
+            size_t temp_storage_size_bytes;
             HIP_CHECK(rocprim::reduce(nullptr,
                                       temp_storage_size_bytes,
                                       input,
-                                      d_output,
+                                      d_output.get(),
                                       size,
                                       rocprim::plus<T>{},
                                       stream,
                                       debug_synchronous));
 
             // allocate temporary storage
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
-            HIP_CHECK(hipDeviceSynchronize());
+            common::device_ptr<void> d_temp_storage(temp_storage_size_bytes);
 
             test_utils::GraphHelper gHelper;
             if(use_graphs)
@@ -576,16 +529,15 @@ void testLargeIndices()
             }
 
             // Run
-            HIP_CHECK(rocprim::reduce(d_temp_storage,
+            HIP_CHECK(rocprim::reduce(d_temp_storage.get(),
                                       temp_storage_size_bytes,
                                       input,
-                                      d_output,
+                                      d_output.get(),
                                       size,
-                                      rocprim::plus<T> {},
+                                      rocprim::plus<T>{},
                                       stream,
                                       debug_synchronous));
 
-            
             if(use_graphs)
             {
                 gHelper.createAndLaunchGraph(stream, true, false);
@@ -595,9 +547,7 @@ void testLargeIndices()
             HIP_CHECK(hipDeviceSynchronize());
 
             // Copy output to host
-            T output = 0;
-            HIP_CHECK(hipMemcpy(&output, d_output, sizeof(T), hipMemcpyDeviceToHost));
-            HIP_CHECK(hipDeviceSynchronize());
+            const auto output = d_output.load()[0];
 
             // Sum of numbers 0 to n - 1 is n(n - 1) / 2, note that this is correct even in case of overflow
             // The division is not integer division but either n or n - 1 has to be even.
@@ -605,9 +555,6 @@ void testLargeIndices()
 
             ASSERT_EQ(output, expected_output);
 
-            hipFree(d_temp_storage);
-            hipFree(d_output);
-
             if(use_graphs)
             {
                 gHelper.cleanupGraphHelper();
@@ -649,7 +596,7 @@ TYPED_TEST(RocprimDeviceReducePrecisionTests, ReduceSumInputEqualExponentFunctio
         //if(size == 0)
         //    continue;
         // as all numbers here are the same and have only 1 significant bit in matnissa the error is like this
-        const float precision = std::max(0.0, test_utils::precision<U> / 2.0 * size - 1.0);
+        const float precision = std::max(0.0, test_utils::precision<U> / 2.0 * (size - 1) - 1.0);
         if(precision > 0.5)
         {
             std::cout << "Test is skipped from size " << size
@@ -672,33 +619,26 @@ TYPED_TEST(RocprimDeviceReducePrecisionTests, ReduceSumInputEqualExponentFunctio
         T lowest = static_cast<T>(
             -1.0
             * static_cast<double>(
-                test_utils::numeric_limits<
+                rocprim::numeric_limits<
                     T>::min())); // smallest (closest to zero) normal (negative) non-zero number
 
         // Generate data
         std::vector<T> input(size, lowest);
-        std::vector<U> output(1, (U)0);
 
-        T* d_input;
-        U* d_output;
-        HIP_CHECK(test_common_utils::hipMallocHelper(&d_input, input.size() * sizeof(T)));
-        HIP_CHECK(test_common_utils::hipMallocHelper(&d_output, output.size() * sizeof(U)));
-        HIP_CHECK(
-            hipMemcpy(d_input, input.data(), input.size() * sizeof(T), hipMemcpyHostToDevice));
-        HIP_CHECK(hipDeviceSynchronize());
+        common::device_ptr<T> d_input(input);
+        common::device_ptr<U> d_output(1);
 
         // Calculate expected results on host mathematically (instead of using reduce on host)
-        U expected = static_cast<U>(static_cast<double>(size) * static_cast<double>(lowest));
+        std::vector<U> expected
+            = test_utils::host_reduce(input.begin(), input.end(), rocprim::plus<U>());
 
-        // temp storage
-        size_t temp_storage_size_bytes;
-        void*  d_temp_storage = nullptr;
         // Get size of d_temp_storage
+        size_t temp_storage_size_bytes;
         HIP_CHECK(rocprim::reduce<Config>(
-            d_temp_storage,
+            nullptr,
             temp_storage_size_bytes,
-            d_input,
-            test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
+            d_input.get(),
+            test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output.get()),
             input.size(),
             rocprim::plus<U>(),
             stream,
@@ -708,8 +648,7 @@ TYPED_TEST(RocprimDeviceReducePrecisionTests, ReduceSumInputEqualExponentFunctio
         ASSERT_GT(temp_storage_size_bytes, 0);
 
         // allocate temporary storage
-        HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
-        HIP_CHECK(hipDeviceSynchronize());
+        common::device_ptr<void> d_temp_storage(temp_storage_size_bytes);
 
         test_utils::GraphHelper gHelper;
         if(TestFixture::use_graphs)
@@ -719,16 +658,15 @@ TYPED_TEST(RocprimDeviceReducePrecisionTests, ReduceSumInputEqualExponentFunctio
 
         // Run
         HIP_CHECK(rocprim::reduce<Config>(
-            d_temp_storage,
+            d_temp_storage.get(),
             temp_storage_size_bytes,
-            d_input,
-            test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
+            d_input.get(),
+            test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output.get()),
             input.size(),
             rocprim::plus<U>(),
             stream,
             debug_synchronous));
 
-        
         if(TestFixture::use_graphs)
         {
             gHelper.createAndLaunchGraph(stream);
@@ -738,17 +676,25 @@ TYPED_TEST(RocprimDeviceReducePrecisionTests, ReduceSumInputEqualExponentFunctio
         HIP_CHECK(hipDeviceSynchronize());
 
         // Copy output to host
-        HIP_CHECK(
-            hipMemcpy(output.data(), d_output, output.size() * sizeof(U), hipMemcpyDeviceToHost));
-        HIP_CHECK(hipDeviceSynchronize());
+        const auto output = d_output.load();
 
         // Check if output values are as expected
-        ASSERT_NO_FATAL_FAILURE(test_utils::assert_near(output[0], expected, precision));
-
-        hipFree(d_input);
-        hipFree(d_output);
-        hipFree(d_temp_storage);
-
+        //ASSERT_NO_FATAL_FAILURE(test_utils::assert_near(output[0], expected, precision));
+        if(size == 0)
+        {
+            ASSERT_NO_FATAL_FAILURE(test_utils::assert_near(output[0], U{}, precision));
+        }
+        else
+        {
+            for(size_t i = 0; i < output.size(); ++i)
+            {
+                ASSERT_NO_FATAL_FAILURE(test_utils::assert_near(
+                    output[i],
+                    expected[i],
+                    test_utils::precision<U>
+                        * std::max(0.0, test_utils::precision<U> / 2.0 * (size - 1 - i) - 1.0)));
+            }
+        }
         if (TestFixture::use_graphs)
         {
             gHelper.cleanupGraphHelper();
@@ -771,7 +717,7 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceMinimum)
     static constexpr bool use_identity_iterator = TestFixture::use_identity_iterator;
     using Config = size_limit_config_t<TestFixture::size_limit>;
 
-    for (size_t seed_index = 0; seed_index < random_seeds_count + seed_size; seed_index++)
+    for(size_t seed_index = 0; seed_index < number_of_runs; seed_index++)
     {
         unsigned int seed_value = seed_index < random_seeds_count  ? rand() : seeds[seed_index - random_seeds_count];
         SCOPED_TRACE(testing::Message() << "with seed = " << seed_value);
@@ -789,26 +735,15 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceMinimum)
 
             // Generate data
             std::vector<T> input = test_utils::get_random_data<T>(size, 1, 100, seed_value);
-            std::vector<U> output(1, U(0));
 
-            T * d_input;
-            U * d_output;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_input, input.size() * sizeof(T)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_output, output.size() * sizeof(U)));
-            HIP_CHECK(
-                hipMemcpy(
-                    d_input, input.data(),
-                    input.size() * sizeof(T),
-                    hipMemcpyHostToDevice
-                )
-            );
-            HIP_CHECK(hipDeviceSynchronize());
+            common::device_ptr<T> d_input(input);
+            common::device_ptr<U> d_output(1);
 
             // reduce function
             binary_op_type min_op;
 
             // Calculate expected results on host
-            U expected = U(test_utils::numeric_limits<U>::max());
+            U expected = U(rocprim::numeric_limits<U>::max());
             for(unsigned int i = 0; i < input.size(); i++)
             {
                 expected = min_op(expected, input[i]);
@@ -816,22 +751,22 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceMinimum)
 
             // Get size of d_temp_storage
             size_t temp_storage_size_bytes;
-            void * d_temp_storage = nullptr;
-            HIP_CHECK(
-                rocprim::reduce<Config>(
-                    d_temp_storage, temp_storage_size_bytes,
-                    d_input,
-                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
-                    test_utils::numeric_limits<U>::max(), input.size(), rocprim::minimum<U>(), stream, TestFixture::debug_synchronous
-                )
-            );
+            HIP_CHECK(rocprim::reduce<Config>(
+                nullptr,
+                temp_storage_size_bytes,
+                d_input.get(),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output.get()),
+                rocprim::numeric_limits<U>::max(),
+                input.size(),
+                rocprim::minimum<U>(),
+                stream,
+                TestFixture::debug_synchronous));
 
             // temp_storage_size_bytes must be >0
             ASSERT_GT(temp_storage_size_bytes, 0);
 
             // allocate temporary storage
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
-            HIP_CHECK(hipDeviceSynchronize());
+            common::device_ptr<void> d_temp_storage(temp_storage_size_bytes);
 
             test_utils::GraphHelper gHelper;
             if(TestFixture::use_graphs)
@@ -840,25 +775,25 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceMinimum)
             }
 
             // Run
-            HIP_CHECK(
-                rocprim::reduce<Config>(
-                    d_temp_storage, temp_storage_size_bytes,
-                    d_input,
-                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
-                    test_utils::numeric_limits<U>::max(), input.size(), rocprim::minimum<U>(), stream, TestFixture::debug_synchronous
-                )
-            );
+            HIP_CHECK(rocprim::reduce<Config>(
+                d_temp_storage.get(),
+                temp_storage_size_bytes,
+                d_input.get(),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output.get()),
+                rocprim::numeric_limits<U>::max(),
+                input.size(),
+                rocprim::minimum<U>(),
+                stream,
+                TestFixture::debug_synchronous));
 
             // Copy output to host
-            HIP_CHECK(
-                hipMemcpyAsync(
-                    output.data(), d_output,
-                    output.size() * sizeof(U),
-                    hipMemcpyDeviceToHost, stream
-                )
-            );
+            std::vector<U> output(1, U(0));
+            HIP_CHECK(hipMemcpyAsync(output.data(),
+                                     d_output.get(),
+                                     output.size() * sizeof(U),
+                                     hipMemcpyDeviceToHost,
+                                     stream));
 
-            
             if(TestFixture::use_graphs)
             {
                 gHelper.createAndLaunchGraph(stream, true, false);
@@ -874,10 +809,6 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceMinimum)
                     ? 0
                     : std::max(test_utils::precision<T>, test_utils::precision<U>)));
 
-            hipFree(d_input);
-            hipFree(d_output);
-            hipFree(d_temp_storage);
-
             if (TestFixture::use_graphs)
             {
                 gHelper.cleanupGraphHelper();
