@@ -1,6 +1,6 @@
 // MIT License
 //
-// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -22,12 +22,31 @@
 
 #include "../common_test_header.hpp"
 
-// required rocprim headers
-#include <rocprim/device/device_run_length_encode.hpp>
+#include "../../common/utils_custom_type.hpp"
+#include "../../common/utils_data_generation.hpp"
+#include "../../common/utils_device_ptr.hpp"
 
 // required test headers
-#include "rocprim/types.hpp"
-#include "test_utils_types.hpp"
+#include "identity_iterator.hpp"
+#include "test_seed.hpp"
+#include "test_utils.hpp"
+#include "test_utils_assertions.hpp"
+#include "test_utils_custom_test_types.hpp"
+#include "test_utils_data_generation.hpp"
+#include "test_utils_hipgraphs.hpp"
+
+// required rocprim headers
+#include <rocprim/device/config_types.hpp>
+#include <rocprim/device/detail/device_config_helper.hpp>
+#include <rocprim/device/device_run_length_encode.hpp>
+#include <rocprim/device/device_run_length_encode_config.hpp>
+#include <rocprim/types.hpp>
+
+#include <algorithm>
+#include <cstddef>
+#include <random>
+#include <stdint.h>
+#include <vector>
 
 template<class Key,
          class Count,
@@ -35,12 +54,14 @@ template<class Key,
          unsigned int MaxSegmentLength,
          // Tests output iterator with void value_type (OutputIterator concept)
          bool UseIdentityIterator = false,
-         class Config             = rocprim::default_config>
+         class Config             = rocprim::default_config,
+         class NonTrivialConfig   = rocprim::default_config>
 struct params
 {
     using key_type = Key;
     using count_type = Count;
     using config                                        = Config;
+    using non_trivial_config                            = NonTrivialConfig;
     static constexpr unsigned int min_segment_length = MinSegmentLength;
     static constexpr unsigned int max_segment_length = MaxSegmentLength;
     static constexpr bool use_identity_iterator = UseIdentityIterator;
@@ -52,40 +73,47 @@ public:
     using params = Params;
 };
 
-using custom_int2 = test_utils::custom_test_type<int>;
-using custom_double2 = test_utils::custom_test_type<double>;
+using custom_int2    = common::custom_type<int, int, true>;
+using custom_double2 = common::custom_type<double, double, true>;
 
-typedef ::testing::Types<
-    params<int, int, 1, 1, true>,
-    params<double, int, 3, 5>,
-    params<float, int, 1, 10>,
+using Params = ::testing::Types<
+    // Tests with default configuration
+    params<int8_t, int8_t, 100, 2000>,
+    params<uint8_t, uint8_t, 100, 2000>,
+    params<int8_t, int8_t, 1000, 5000>,
+    params<uint8_t, uint8_t, 1000, 5000>,
+    params<int, unsigned int, 1000, 5000>,
+    params<unsigned int, size_t, 2048, 2048>,
+    params<unsigned int, unsigned int, 1000, 50000>,
     params<unsigned long long, size_t, 1, 30>,
-    params<custom_int2, unsigned int, 20, 100>,
     params<float, unsigned long long, 100, 400>,
+    params<float, int, 1, 10>,
+    params<double, int, 3, 5>,
+    params<double, int, 100, 2000>,
+    params<int, rocprim::half, 100, 2000>,
+    // half should be supported, but is missing some key operators.
+    // we should uncomment these, as soon as these are implemented and the tests compile and work as intended.
+    //params<rocprim::half, int, 100, 2000>,
+    params<int, rocprim::bfloat16, 1000, 5000>,
+    params<rocprim::bfloat16, int, 1000, 5000>,
+    // Tests for custom types
+    params<custom_int2, unsigned int, 20, 100>,
+    params<custom_double2, custom_int2, 10, 30000, true>,
+    params<unsigned long long, custom_double2, 100000, 100000>,
+    // Tests for supported config structs
     params<unsigned int,
            unsigned int,
            200,
            600,
            false,
+           // RLE config
            rocprim::run_length_encode_config<rocprim::reduce_by_key_config<128, 5>,
-                                             rocprim::select_config<64, 3>>>,
-    params<double, int, 100, 2000>,
-    params<custom_double2, custom_int2, 10, 30000, true>,
-    params<int, unsigned int, 1000, 5000>,
-    params<int8_t, int8_t, 100, 2000>,
-    params<uint8_t, uint8_t, 100, 2000>,
-    params<int, rocprim::half, 100, 2000>,
-    // half should be supported, but is missing some key operators.
-    // we should uncomment these, as soon as these are implemented and the tests compile and work as intended.
-    //params<rocprim::half, int, 100, 2000>,
-    params<int8_t, int8_t, 1000, 5000>,
-    params<uint8_t, uint8_t, 1000, 5000>,
-    params<int, rocprim::bfloat16, 1000, 5000>,
-    params<rocprim::bfloat16, int, 1000, 5000>,
-    params<unsigned int, size_t, 2048, 2048>,
-    params<unsigned int, unsigned int, 1000, 50000>,
-    params<unsigned long long, custom_double2, 100000, 100000>>
-    Params;
+                                             rocprim::select_config<64, 3>>,
+           // RLE non-trivial config
+           rocprim::run_length_encode_config<rocprim::reduce_by_key_config<256, 15>,
+                                             rocprim::select_config<256, 13>>>,
+    // Tests for when output's value_type is void
+    params<int, int, 1, 1, true>>;
 
 TYPED_TEST_SUITE(RocprimDeviceRunLengthEncode, Params);
 
@@ -117,9 +145,10 @@ TYPED_TEST(RocprimDeviceRunLengthEncode, Encode)
 
     const unsigned int seed = 123;
     std::default_random_engine gen(seed);
-    std::vector<key_type> random_keys = test_utils::get_random_data<key_type>(64, -100, 100, seed);
+    std::vector<key_type>      random_keys
+        = test_utils::get_random_data_wrapped<key_type>(64, -100, 100, seed);
 
-    for (size_t seed_index = 0; seed_index < random_seeds_count + seed_size; seed_index++)
+    for(size_t seed_index = 0; seed_index < number_of_runs; seed_index++)
     {
         unsigned int seed_value = seed_index < random_seeds_count  ? rand() : seeds[seed_index - random_seeds_count];
         SCOPED_TRACE(testing::Message() << "with seed = " << seed_value);
@@ -130,17 +159,20 @@ TYPED_TEST(RocprimDeviceRunLengthEncode, Encode)
 
             hipStream_t stream = 0; // default
 
+            // Default stream does not support hipGraph stream capture, so create one
+            HIP_CHECK(hipStreamCreateWithFlags(&stream, hipStreamNonBlocking));
+
             // Generate data and calculate expected results
             std::vector<key_type> unique_expected;
             std::vector<count_type> counts_expected;
             size_t runs_count_expected = 0;
 
-            std::vector<key_type> input(size);
-            std::uniform_int_distribution<size_t> key_count_dis(
+            std::vector<key_type>                    input(size);
+            common::uniform_int_distribution<size_t> key_count_dis(
                 TestFixture::params::min_segment_length,
-                TestFixture::params::max_segment_length
-            );
-            std::vector<count_type> values_input = test_utils::get_random_data<count_type>(size, 0, 100, seed_value);
+                TestFixture::params::max_segment_length);
+            std::vector<count_type> values_input
+                = test_utils::get_random_data_wrapped<count_type>(size, 0, 100, seed_value);
 
             size_t offset = 0;
             key_type current_key = get_random_value_no_duplicate(key_type(0), random_keys, size);
@@ -164,83 +196,55 @@ TYPED_TEST(RocprimDeviceRunLengthEncode, Encode)
                 offset += key_count;
             }
 
-            key_type * d_input;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_input, size * sizeof(key_type)));
-            HIP_CHECK(
-                hipMemcpy(
-                    d_input, input.data(),
-                    size * sizeof(key_type),
-                    hipMemcpyHostToDevice
-                )
-            );
+            common::device_ptr<key_type> d_input(input);
 
-            key_type * d_unique_output;
-            count_type * d_counts_output;
-            count_type * d_runs_count_output;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_unique_output, runs_count_expected * sizeof(key_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_counts_output, runs_count_expected * sizeof(count_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_runs_count_output, sizeof(count_type)));
+            common::device_ptr<key_type>   d_unique_output(runs_count_expected);
+            common::device_ptr<count_type> d_counts_output(runs_count_expected);
+            common::device_ptr<count_type> d_runs_count_output(1);
 
             size_t temporary_storage_bytes = 0;
 
             HIP_CHECK(rocprim::run_length_encode<config>(
                 nullptr,
                 temporary_storage_bytes,
-                d_input,
+                d_input.get(),
                 size,
-                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_unique_output),
-                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_counts_output),
-                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_runs_count_output),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_unique_output.get()),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_counts_output.get()),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(
+                    d_runs_count_output.get()),
                 stream,
                 debug_synchronous));
 
             ASSERT_GT(temporary_storage_bytes, 0U);
 
-            void * d_temporary_storage;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_temporary_storage, temporary_storage_bytes));
+            common::device_ptr<void> d_temporary_storage(temporary_storage_bytes);
+
+            test_utils::GraphHelper gHelper;
+            gHelper.startStreamCapture(stream);
 
             HIP_CHECK(rocprim::run_length_encode<config>(
-                d_temporary_storage,
+                d_temporary_storage.get(),
                 temporary_storage_bytes,
-                d_input,
+                d_input.get(),
                 size,
-                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_unique_output),
-                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_counts_output),
-                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_runs_count_output),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_unique_output.get()),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_counts_output.get()),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(
+                    d_runs_count_output.get()),
                 stream,
                 debug_synchronous));
 
-            HIP_CHECK(hipFree(d_temporary_storage));
+            gHelper.createAndLaunchGraph(stream);
 
-            std::vector<key_type> unique_output(runs_count_expected);
-            std::vector<count_type> counts_output(runs_count_expected);
-            std::vector<count_type> runs_count_output(1);
-            HIP_CHECK(
-                hipMemcpy(
-                    unique_output.data(), d_unique_output,
-                    runs_count_expected * sizeof(key_type),
-                    hipMemcpyDeviceToHost
-                )
-            );
-            HIP_CHECK(
-                hipMemcpy(
-                    counts_output.data(), d_counts_output,
-                    runs_count_expected * sizeof(count_type),
-                    hipMemcpyDeviceToHost
-                )
-            );
-            HIP_CHECK(
-                hipMemcpy(
-                    runs_count_output.data(), d_runs_count_output,
-                    sizeof(count_type),
-                    hipMemcpyDeviceToHost
-                )
-            );
+            HIP_CHECK(hipDeviceSynchronize());
 
-            HIP_CHECK(hipFree(d_input));
-            HIP_CHECK(hipFree(d_unique_output));
-            HIP_CHECK(hipFree(d_counts_output));
-            HIP_CHECK(hipFree(d_runs_count_output));
+            const auto unique_output     = d_unique_output.load();
+            const auto counts_output     = d_counts_output.load();
+            const auto runs_count_output = d_runs_count_output.load();
+
+            gHelper.cleanupGraphHelper();
+            HIP_CHECK(hipStreamDestroy(stream));
 
             // Validating results
 
@@ -264,16 +268,18 @@ TYPED_TEST(RocprimDeviceRunLengthEncode, NonTrivialRuns)
     using key_type = typename TestFixture::params::key_type;
     using count_type = typename TestFixture::params::count_type;
     using offset_type = typename TestFixture::params::count_type;
-    using config      = typename TestFixture::params::config;
+    using config      = typename TestFixture::params::non_trivial_config;
 
     constexpr bool use_identity_iterator = TestFixture::params::use_identity_iterator;
+
     const bool debug_synchronous = false;
 
     const unsigned int seed = 123;
     std::default_random_engine gen(seed);
-    std::vector<key_type> random_keys = test_utils::get_random_data<key_type>(64, -100, 100, seed);
+    std::vector<key_type>      random_keys
+        = test_utils::get_random_data_wrapped<key_type>(64, -100, 100, seed);
 
-    for (size_t seed_index = 0; seed_index < random_seeds_count + seed_size; seed_index++)
+    for(size_t seed_index = 0; seed_index < number_of_runs; seed_index++)
     {
         unsigned int seed_value = seed_index < random_seeds_count  ? rand() : seeds[seed_index - random_seeds_count];
         SCOPED_TRACE(testing::Message() << "with seed = " << seed_value);
@@ -284,18 +290,19 @@ TYPED_TEST(RocprimDeviceRunLengthEncode, NonTrivialRuns)
 
             hipStream_t stream = 0; // default
 
+            // Default stream does not support hipGraph stream capture, so create one
+            HIP_CHECK(hipStreamCreateWithFlags(&stream, hipStreamNonBlocking));
+
             // Generate data and calculate expected results
             std::vector<offset_type> offsets_expected;
             std::vector<count_type> counts_expected;
             size_t runs_count_expected = 0;
 
-            std::vector<key_type> input(size);
-            std::uniform_int_distribution<size_t> key_count_dis(
+            std::vector<key_type>                    input(size);
+            common::uniform_int_distribution<size_t> key_count_dis(
                 TestFixture::params::min_segment_length,
-                TestFixture::params::max_segment_length
-            );
+                TestFixture::params::max_segment_length);
             std::bernoulli_distribution is_trivial_dis(0.1);
-            std::vector<count_type> values_input = test_utils::get_random_data<count_type>(size, 0, 100, seed_value);
 
             size_t offset = 0;
             key_type current_key = get_random_value_no_duplicate(key_type(0), random_keys, size);
@@ -331,96 +338,78 @@ TYPED_TEST(RocprimDeviceRunLengthEncode, NonTrivialRuns)
                 offset += key_count;
             }
 
-            key_type * d_input;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_input, size * sizeof(key_type)));
-            HIP_CHECK(
-                hipMemcpy(
-                    d_input, input.data(),
-                    size * sizeof(key_type),
-                    hipMemcpyHostToDevice
-                )
-            );
+            common::device_ptr<key_type> d_input(input);
 
-            offset_type * d_offsets_output;
-            count_type * d_counts_output;
-            count_type * d_runs_count_output;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_offsets_output, std::max<size_t>(1, runs_count_expected) * sizeof(offset_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_counts_output, std::max<size_t>(1, runs_count_expected) * sizeof(count_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_runs_count_output, sizeof(count_type)));
-
-            size_t temporary_storage_bytes = 0;
+            common::device_ptr<offset_type> d_offsets_output(
+                std::max<size_t>(1, runs_count_expected));
+            common::device_ptr<count_type> d_counts_output(
+                std::max<size_t>(1, runs_count_expected));
+            common::device_ptr<count_type> d_runs_count_output(1);
 
+            size_t temporary_storage_bytes;
             HIP_CHECK(rocprim::run_length_encode_non_trivial_runs<config>(
                 nullptr,
                 temporary_storage_bytes,
-                d_input,
+                d_input.get(),
                 size,
-                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_offsets_output),
-                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_counts_output),
-                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_runs_count_output),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(
+                    d_offsets_output.get()),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_counts_output.get()),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(
+                    d_runs_count_output.get()),
                 stream,
                 debug_synchronous));
 
             ASSERT_GT(temporary_storage_bytes, 0U);
 
-            void * d_temporary_storage;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_temporary_storage, temporary_storage_bytes));
+            common::device_ptr<void> d_temporary_storage(temporary_storage_bytes);
+
+            test_utils::GraphHelper gHelper;
+            gHelper.startStreamCapture(stream);
 
             HIP_CHECK(rocprim::run_length_encode_non_trivial_runs<config>(
-                d_temporary_storage,
+                d_temporary_storage.get(),
                 temporary_storage_bytes,
-                d_input,
+                d_input.get(),
                 size,
-                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_offsets_output),
-                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_counts_output),
-                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_runs_count_output),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(
+                    d_offsets_output.get()),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_counts_output.get()),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(
+                    d_runs_count_output.get()),
                 stream,
                 debug_synchronous));
 
-            HIP_CHECK(hipFree(d_temporary_storage));
+            gHelper.createAndLaunchGraph(stream);
+
+            HIP_CHECK(hipDeviceSynchronize());
+
+            std::vector<offset_type> offsets_output;
+            std::vector<count_type>  counts_output;
+            const auto               runs_count_output = d_runs_count_output.load();
 
-            std::vector<offset_type> offsets_output(runs_count_expected);
-            std::vector<count_type> counts_output(runs_count_expected);
-            std::vector<count_type> runs_count_output(1);
             if(runs_count_expected > 0)
             {
-                HIP_CHECK(
-                    hipMemcpy(
-                        offsets_output.data(), d_offsets_output,
-                        runs_count_expected * sizeof(offset_type),
-                        hipMemcpyDeviceToHost
-                    )
-                );
-                HIP_CHECK(
-                    hipMemcpy(
-                        counts_output.data(), d_counts_output,
-                        runs_count_expected * sizeof(count_type),
-                        hipMemcpyDeviceToHost
-                    )
-                );
+                offsets_output = d_offsets_output.load();
+                counts_output  = d_counts_output.load();
             }
-            HIP_CHECK(
-                hipMemcpy(
-                    runs_count_output.data(), d_runs_count_output,
-                    sizeof(count_type),
-                    hipMemcpyDeviceToHost
-                )
-            );
 
-            HIP_CHECK(hipFree(d_input));
-            HIP_CHECK(hipFree(d_offsets_output));
-            HIP_CHECK(hipFree(d_counts_output));
-            HIP_CHECK(hipFree(d_runs_count_output));
+            gHelper.cleanupGraphHelper();
+            HIP_CHECK(hipStreamDestroy(stream));
 
             // Validating results
 
             std::vector<count_type> runs_count_expected_2;
             runs_count_expected_2.push_back(static_cast<count_type>(runs_count_expected));
-            test_utils::assert_eq(runs_count_output, runs_count_expected_2, 1);
-
-            ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(offsets_output, offsets_expected, runs_count_expected));
-            ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(counts_output, counts_expected, runs_count_expected));
+            SCOPED_TRACE(testing::Message() << "runs_count_output");
+            ASSERT_NO_FATAL_FAILURE(
+                test_utils::assert_eq(runs_count_output, runs_count_expected_2, 1));
+            SCOPED_TRACE(testing::Message() << "offsets_output");
+            ASSERT_NO_FATAL_FAILURE(
+                test_utils::assert_eq(offsets_output, offsets_expected, runs_count_expected));
+            SCOPED_TRACE(testing::Message() << "counts_output");
+            ASSERT_NO_FATAL_FAILURE(
+                test_utils::assert_eq(counts_output, counts_expected, runs_count_expected));
         }
     }
-
 }
